/*==============================================================================

            ITR_MARVIN - the MSGC Front End Reconstruction Program 
            ------------------------------------------------------

      o   pedestal subtraction
      o   pedestal following
      o   raw noise following
      o   baseline subtraction 
      o   subtracted noise following
      o   cluster finding 

  Author: Michael Schmelling  1997/1998
  Adapted to the ITR:  Verena Buchholz 1999
                       Wolfgang Gradl  1999-?

  General Byte-Swap Policy as of 21-Jan-99 (RW):
  ----------------------------------------------

  On file as well as between sevt->data and sevt->next in memory
  the data are ALWAYS in 'export' format, that means in Sun/Hera-B
  format. On Linux machines byte swapping has to be done even when
  just reading from sevt (since it is not clear to marvin, who
  created the sevt - eg. just from reading in DSVD or by an earlier
  marvin).

  ITR: July 1999: new Clusterfinder
                  pipelinelength for different helix versions 
	Sep 1999: in silraw chip->flag = 2 for false pipelinelength
        Sep 1999: use channelstatus from DB -> set data=0
	Oct 1999: one dead channel in ITRclusterfinder
	Oct 1999: use chambertype from DB
	Oct 1999: target information no longer available
	Nox 1999: vs 3-5: in format 4 variable izero/flag used for end of 
	          cluster
                  this version is shortened 
		  structure cluster expanded
		  ped etc in database instead of file
	Nov 1999: vs 3-6: structure cluster expanded (but long version)
                  sparse data file re-included
		  chambers ... as hex-numbers
        Mar 2000  two more clusteralgorithms:  
	            2  using channel variance for seeding (TZ, a la VDS)
		    3  A. Lange's cluster finder
	          optional: dynamic threshold for KM's cluster finder
	Apr 2000  (wg)  introduced new baseline correction and modified
                   noise and pedestal treatments (M. Schmelling, Marvin3-3)
	Nov 2000  (wg)  minor cosmetic changes in ITR_marvin_param_init()
	Feb 2001  (wg)  bug fix for cluster size in format 5
     Feb 8, 2001  (wg)  catch reading of corrupted sparse-events to prevent
                        core dumps :-)
     Nov26, 2001  (wg)  Use database values for noise and pedestal as 
                        initial guess (in ITR_marvin_readSetup()).
     Jan28, 2002  (wg)  chip->status has slightly changed meaning:
                        > 0   channel live and ok
                        ==0   masked because not connected
                        < 0   masked by hot-channel finder (Yu. Gorbounov)
     Feb06, 2002  (wg)  allow loong pipeline of Helix 3.1a with 142 columns
     Apr05, 2002  (wg)  merge experimental Region-of-Interest cluster finder
     May10, 2002  (wg)  fix stupid omission of checks for longer pipeline.
                        Well, I mean, really.

 $Id: ITR_marvinlib.c,v 1.1.1.13 2002/06/06 07:51:35 gradl Exp $
=============================================================================*/


/* include files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>

#include "table.h"
#include "component.h"
#include "event.h"
#include "etb.h"
#include "esp.h"
#include "grr.h"
#include "screen.h"  /*!*/


#include "aframelib.h"
#include "iol.h"
#include "itrleda.h"
#include "ITR_silraw.h"
#include "ITR_marvin.h"

#include "CVStag.ITR_marvin"

/*===========================================================================
 * prototypes for functions which should really not be called from outside
 *--------------------------------------------------------------------------*/
int ITR_marvin_get_cluster_length(int js0, int j0, int *pj1, int *pjs1,
				  int jlast, float **signal,
				  float localmean, float c_level_1,
				  float c_level_2);

int ITR_marvin_processSparsifiedChip(ITR_Marvin_Chip_t * chip, int cell);
int ITR_marvin_getROIClustersChamber(ITR_Marvin_Chamber_t * chamber, int cell);
int ITR_marvin_getROIClusters(void *source, void *dest, int maxsize, 
			      ESP_machine into_fmt, SLB_SUBHDR *fed,
			      int first_channel, int size, int *nChan);

/*===========================================================================
 *  global variables and macros
 *--------------------------------------------------------------------------*/

#define NCHAN_CHIP        128	/* number of channels per chip            */
#define ITRNCHAN_CHAMBER  768	/* number of channels per chamber (max)    */
#define NOCREATE          0	/* flag used with ITR_marvin_findChamber/Chip */
#define CREATE            1	/* flag used with ITR_marvin_findChamber/Chip */


/* static pointer to ITR_marvin's and sparsified event */
static ITR_Marvin_Event_t *marvin_evt_ptr = 0;
static ITR_Marvin_Sevt_t *marvin_sevt_ptr = 0;

/* array for fast access to chips from chipid          */
#define ITR_MAXCHIPID         0x4FFF
static ITR_Marvin_Chip_t *chip_ptr[ITR_MAXCHIPID] = { NULL };

static int (*clusterfinder) (ITR_Marvin_Chamber_t *, int);
static int ITR_eventcount = 0;
static FILE *sparse_file = NULL;
static int const kHeadersize = sizeof(ITR_Marvin_Header_t);
static int const moniFormat = 4;	/* current monitoring format */

/* flag, if clustering was done at all in this event   */
static int clustering_done = -1;

/* --wg-- 
   we don't want to compute the mean height of the column numbers for each chip 
   for each event.  Hence this global variable.                                  */
static float meanGain = 1.;

/* make ITR_marvin_iol a global variable inside marvinlib */
static ITR_Marvin_IOL_t ITR_marvin_iol = {
    xcvsMODULE_ITR_marvin,
    IOL_NVERBS(ITR_Marvin_IOL_t, 0),
    {"version", xcvsVERSION_ITR_marvin, IOL_TYPE_STRING},
    {"use_ITR_marvin", 1, " ", IOL_TYPE_INT},
    {"pathsd", "none", IOL_TYPE_STRING},
    {"nevmon", 500, " ", IOL_TYPE_INT},
    {"seedcut", 9., " ", IOL_TYPE_FLOAT},
    {"chi2cut", 20., " ", IOL_TYPE_FLOAT},
    {"smoothcut", 100., " ", IOL_TYPE_FLOAT},
    {"flatcut", 100., " ", IOL_TYPE_FLOAT},
    {"sdformat", 5, " ", IOL_TYPE_INT},
    {"sdtype", 2, " ", IOL_TYPE_INT},
    {"baseline", 2, " ", IOL_TYPE_INT},
    {"nbaseline", 1, " ", IOL_TYPE_INT},
    {"ncellshortp", 1, " ", IOL_TYPE_INT},
    {"ncellshortr", 1, " ", IOL_TYPE_INT},
    {"ncellshorts", 1, " ", IOL_TYPE_INT},
    {"ncelllongp", 1, " ", IOL_TYPE_INT},
    {"ncelllongr", 1, " ", IOL_TYPE_INT},
    {"ncelllongs", 1, " ", IOL_TYPE_INT},
    {"ped update rate", 1, " ", IOL_TYPE_INT},
    {"rvar update rate", 1, " ", IOL_TYPE_INT},
    {"svar update rate", 1, " ", IOL_TYPE_INT},
    {"reportRate", 100, " ", IOL_TYPE_INT},
    {"dumpRaw", 0, " ", IOL_TYPE_INT},
    {"dumpClusters", 0, " ", IOL_TYPE_INT},
    {"debug mode", 0, " ", IOL_TYPE_INT},
    {"pulseHeight", 5, " ", IOL_TYPE_FLOAT},
    {"minTotalCharge", 0, " ", IOL_TYPE_FLOAT},
    {"maxCluSize", 10, " ", IOL_TYPE_FLOAT},
    {"relSTEIG", 1, " ", IOL_TYPE_FLOAT},
    {"minDiff", 4, " ", IOL_TYPE_FLOAT},
    {"LocalMeanSize", 3, " ", IOL_TYPE_FLOAT},
    {"MaxPositiveCh", 0.15, " ", IOL_TYPE_FLOAT},
    {"CutRMS", 3, " ", IOL_TYPE_FLOAT},
    {"diffrile", 4, " ", IOL_TYPE_FLOAT},
    {"clusterfinder", 2, " ", IOL_TYPE_INT},
    {"gaincorrection", 0, " ", IOL_TYPE_INT},
    {"gaincorr minHeight", 20., " ", IOL_TYPE_FLOAT},
    {"dynamic threshold", 0, " ", IOL_TYPE_INT},
    {"cluster ratio", 1.2, " ", IOL_TYPE_FLOAT},
    {"hotcut", 4., " ", IOL_TYPE_FLOAT},
    {"svar maxSignal", 25., " ", IOL_TYPE_FLOAT}
};



/*============================================================================
 *                          standard functions 
 *===========================================================================*/

int ITR_marvin_param_init()
/*-----------------------------------------------------------------------------
 * Initialize ITR_marvin_iol
 * =====================
 *
 * Author: Michael Schmelling / 21-May-1998
 *---------------------------------------------------------------------------*/
{
    char *routine = "ITR_marvin_param_init";
    static int marvinParamInitDone = 0;

    /* if we were already called, return quietly */
    if (marvinParamInitDone)
	return (GRR_OK);


    aframe_mess(routine, GRR_SINFO,
		"ITR_marvin version %s started.", xcvsVERSION_ITR_marvin);

    /* make ITR_marvin_iol known to IOL */
    iol_declare_structure((IOL_GENERIC *) & ITR_marvin_iol);

    marvinParamInitDone = 1;

    return GRR_OK;
}


/*
 *  Reduced to a dummy procedure ( only calling ITR_marvin_init() 
 *  but this is important! ) 
 *
 *          OBSOLETE !         
 */
int ITR_marvin_startrun()
{
    return ITR_marvin_init();
}


/* -----------------------------------------------------------------------------
 *
 *    Initialize marvin
 *    =================
 *
 *    Author: Rainer Wanke / 9-May-1999
 *
 * ---------------------------------------------------------------------------*/
int ITR_marvin_init()
{
    static ITR_Marvin_Event_t marvin_evt;
    static ITR_Marvin_Sevt_t marvin_sevt;

    int run;
    char filename[256];

    char *routine = "ITR_marvin_init";

/*---------------------------------------------------------------------------*/

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO, "marvin internal initialization");


    /* make ITR_marvin_evt and ITR_marvin_sevt known to the rest of the world */
    marvin_evt_ptr = &marvin_evt;
    marvin_sevt_ptr = &marvin_sevt;

    /* read chip, chamber, and bond map information */
    ITR_marvin_readSetup();

    /* only selection of clusterfinder */
    if (ITR_marvin_iol.clusterfinder.value == 1)
	clusterfinder = ITR_marvin_getITRClusters;
    else if (ITR_marvin_iol.clusterfinder.value == 2)
	clusterfinder = ITR_marvin_getNegativeClusters;
    else if (ITR_marvin_iol.clusterfinder.value == 3)
	clusterfinder = ITR_marvin_getSparseITRClusters;
    else if (ITR_marvin_iol.clusterfinder.value == 4)
	clusterfinder = ITR_marvin_getROIClustersChamber;
    else {
	aframe_mess(routine, GRR_SWARNING, "false clusterfinder!");
	return GRR_FISHY;	/*VB: das klappt?? */
    }


    /*  open sparse data file */
    if (strcmp(ITR_marvin_iol.pathsd.value, "none")) {
	ITR_silraw_get_major(&run);
	sprintf(filename, "%s/sd.%04d", ITR_marvin_iol.pathsd.value, run);

	sparse_file = fopen(filename, "wb");
	if (sparse_file)
	    aframe_mess(routine, GRR_SINFO,
			"opened sparse data file <%s>", filename);
	else
	    aframe_mess(routine, GRR_SWARNING,
			"open of sparse data file <%s> failed", filename);
    }

    return GRR_OK;
}


/*-----------------------------------------------------------------------------
 *
 * Event processing 
 * ================
 *
 * Author: Michael Schmelling / 21-May-1998
 *         last modified   RW /  9-May-1999
 *
 *         Modified: 8/feb/2001 (wg) 
 *             Definition of return codes:  
 *                GRR_FISHY     permanent error, can't continue
 *                GRR_CONTINUE  error which may be recoverable, call again at
 *                              next event
 *
 *---------------------------------------------------------------------------*/
int ITR_marvin_event()
{
    ITR_Silraw_Event_t *silraw;
    ITR_Marvin_Event_t *marvin_evt;
    ITR_Marvin_Chamber_t *chamber;
    ITR_Marvin_Chip_t *chip;
    ITR_Marvin_Sevt_t *sevt;

    /* keep track of event numbers */
    static int evseen = -1;

    /* local variables */
    int i, major, minor, ichamber, nchip, chipid, pipelen, nsparse;
    int cellshort, celllong;
    int ncl, dncl, format, sevt_len, sevt_free;
    char *start_ptr;
    int rc;
    int nevmon;

    char *routine = "ITR_marvin_event";

/*---------------------------------------------------------------------------*/

    if (ITR_marvin_iol.use_ITR_marvin.value == 0)
	return GRR_OK;

    /* access information and return in case of error */
    silraw = ITR_silraw_fetch_event();
    if (!silraw) {
	aframe_mess(routine, GRR_SINFO, "no ITR_Silraw_Event_t found");
	return GRR_FISHY;
    }
    marvin_evt = ITR_marvin_fetch_event();
    if (!marvin_evt) {
	aframe_mess(routine, GRR_SINFO, "no Marvin_Event_t found");
	return GRR_FISHY;
    }
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no ITR_Marvin_Sevt_t found");
	return GRR_FISHY;
    }

    /* reset all chips and chambers:
     * chip reset: chipstatus <= 0
     * chamber reset: first = max, last = 0, chamberstatus <= 0 */
    for (i = 0; i < marvin_evt->nchip; i++)
	if (marvin_evt->chip[i]->chipstatus > 0)
	    marvin_evt->chip[i]->chipstatus = 0;

    for (i = 0; i < marvin_evt->nchamber; i++) {
	marvin_evt->chamber[i]->first = ITRNCHAN_CHAMBER;
	marvin_evt->chamber[i]->last = 0;
	if (marvin_evt->chamber[i]->chamberstatus > 0)
	    marvin_evt->chamber[i]->chamberstatus = 0;
    }


    /* access basic event information */
    major = silraw->header.major;
    minor = silraw->header.minor;
    cellshort = silraw->header.cellshort;
    celllong = silraw->header.celllong;
    nchip = silraw->nchip;
    nsparse = silraw->nsparse;

    /* event count and raw data dumps ... if requested */
    if (!(++evseen % ITR_marvin_iol.reportRate.value))
	aframe_mess(routine, GRR_SINFO,
		    "processing event: major=%7d minor=%7d", major, minor);

    if (ITR_marvin_iol.dumpRaw.value)
	ITR_marvin_dumpRaw(ITR_marvin_iol.dumpRaw.value);

    /* 
     * different scenarios:
     *
     * 1) sevt->next == sevt->data : build header
     * 2) sevt->next  > sevt->data : nothing to be done
     *
     * in all cases additional cluster information from 
     * chips is later appended to the sevt-record
     *
     */

    ncl = 0;			/* total cluster count */
    sevt->format = ITR_marvin_iol.sdformat.value;	/* sparse data format  */

    if (sevt->format < 1 || sevt->format > 5) {
	aframe_mess(routine, GRR_SINFO, "invalid sd-format %d from IOL",
		    sevt->format);
	return GRR_FISHY;
    }

    /*
     *  realize the scenario described above 
     */

    if (nsparse > 0) {

	/* for empty sevt record copy the sevt header          */
	/* (since its just a copy, no byte-swapping is needed) */
	if (sevt->next == sevt->data) {

	    /* if corrupted sparsified block is read in, make sure that 
	     * no memory which does not belong to us is accessed and that 
	     * sevt->header->ncl = 0                      --- (wg, feb 8, 2001) */ 
	    if (silraw->sparse[0].length  < kHeadersize) {
		static int warnPrint = 1;
		int        zero = 0;
		if (warnPrint) {
		    aframe_mess(routine, GRR_SERROR, 
				"Corrupted sparsified data block: %d bytes",
				(unsigned short) silraw->sparse[i].length);
		    warnPrint = 0;
		}

		ESP_put_int( &zero, (int*) (sevt->data + kHeadersize - sizeof(int)), 1); 
		return GRR_FISHY;
	    }  
	    memcpy(sevt->next, silraw->sparse[0].data, kHeadersize);
	    sevt->next += kHeadersize;
	}
	
	/* then collect all the data records */
	for (i=0; i<nsparse; i++) {      
	    start_ptr  = silraw->sparse[i].data   + kHeadersize;
	    sevt_len   = (unsigned) silraw->sparse[i].length - kHeadersize;
	    sevt_free  = sevt->size - (sevt->next - sevt->data);
	  
	    if (silraw->sparse[i].length  < kHeadersize) {
		static int warnPrint = 1;
		if (warnPrint) {
		    aframe_mess(routine, GRR_SERROR, 
				"Corrupted sparsified data block: %d bytes",
				(unsigned short) silraw->sparse[i].length);
		    warnPrint = 0;
		}
		return GRR_FISHY;
	    }  

	    if ( sevt_len > sevt_free ) {
		aframe_mess(routine,GRR_SWARNING,"sevt buffer exhausted");  
		return GRR_CONTINUE;
	    }
	  
	    sevt->next = memcpy(sevt->next, start_ptr, sevt_len) + sevt_len;
	    ESP_get_int( (int*) start_ptr - 1, &dncl,   1); 
	    ESP_get_int( (int*) start_ptr - 2, &format, 1); 
	    ncl += dncl;
	    if ( format != ITR_marvin_iol.sdformat.value ) {
		aframe_mess(routine,GRR_SWARNING,
			    "preferences modified:  sd-format set from old=%d to new=%d"
			    ,
			    ITR_marvin_iol.sdformat.value,format);
		ITR_marvin_iol.sdformat.value = format; 
		return GRR_CONTINUE;
	    }
	}
      
	/* finally update the cluster count and flag "copy done" to silraw */
      
      ESP_put_int( &ncl, (int*) (sevt->data + kHeadersize - sizeof(int)), 1); 
      silraw->nsparse = 0;
    }
    else {
      
      /* initialize an sevt-header *//*VB: cell aufgeteilt... */
      if (sevt->next == sevt->data) {
	ESP_put_int(&major, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&minor, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&cellshort, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&celllong, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&(sevt->format), (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&ncl, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
      }
    }
    
    /*
     *  start processing of unsparsified data 
     *  =====================================
     */

    if (nchip > 0) {

	/* Loop over all ITR_silraw-chips to process them */
	if (ITR_marvin_iol.debug_mode.value)
	    aframe_mess(routine, GRR_SINFO,
			"Number of chips to be processed = %d", nchip);
	for (i = 0; i < nchip; i++) {

	    /* Don't do anything, if ITR_silraw-chip-flag is set or chipstatus < 0 */
	    /*VB silraw->chip[i].flag = 2 for false pipeline */
	    if (silraw->chip[i].flag >= 1)
		continue;
	    chipid = silraw->chip[i].chipid;
	    chip = ITR_marvin_findChip(chipid);
	    if (!chip)
		continue;
	    if (chip->chipstatus < 0)
		continue;

	    chip->digi = silraw->chip[i].data;
	    chip->hcolno = silraw->chip[i].cellnumheight;	/* ---wg--- */

	    if (ITR_marvin_iol.debug_mode.value)
		aframe_mess(routine, GRR_SINFO,
			    "processing chip=0x%04x chamber=%6d",
			    chip->chipid, chip->chamberid);

	    /*VB: */
	    pipelen = silraw->chip[i].pipelen;

	    if (pipelen != NCELLHELIXSHORT && pipelen != NCELLHELIXLONG && pipelen != NCELLHELIX31A) {
		if (ITR_marvin_iol.debug_mode.value)
		    aframe_mess(routine, GRR_SWARNING,
				"chip 0x%x:pipelinelength %d not valid",
				chipid, pipelen);
		continue;
	    }

	    if (!(silraw->chip[i].recid & 0x1)) {
		if (pipelen == NCELLHELIXSHORT) {
		    ITR_marvin_processChip(chip, cellshort);
		} else if (pipelen == NCELLHELIXLONG || pipelen == NCELLHELIX31A) {
		    ITR_marvin_processChip(chip, celllong);
		}
	    } else
		ITR_marvin_processSparsifiedChip(chip, pipelen);

	    /* --- wg --- */
	    /* ITR_marvin_setdeadChannels (chip); */

	}			/*end for(...nchip...)-loop */


	/* 
	 *  sparsification
	 */
	nevmon = ITR_marvin_iol.nevmon.value;
	ITR_eventcount++;

	/* loop over chambers for cluster finding */
	clustering_done = 0;
	for (ichamber = 0; ichamber < marvin_evt->nchamber; ichamber++) {
	    chamber = marvin_evt->chamber[ichamber];
	    if (!chamber) {
		aframe_mess(routine, GRR_SWARNING,
			    "Chamber #0x%x not found in marvin_evt!",
			    ichamber);
		continue;
	    }
	    if (ITR_marvin_iol.debug_mode.value)
		aframe_mess(routine, GRR_SINFO, "Found chamber 0x%x",
			    chamber->chamberid);

	    if (chamber->first < chamber->last
		&& chamber->chamberstatus > 0) {
		pipelen = chamber->chip[0]->pipelen;
		if (pipelen == NCELLHELIXSHORT) {
		    if (ITR_marvin_iol.dumpRaw.value)
			ITR_marvin_dumpChamber(chamber);

		    rc = ITR_marvin_calculate_threshold(chamber, nevmon, ITR_eventcount);

		    clusterfinder(chamber, cellshort);
		    clustering_done = 1;
		} else if (pipelen == NCELLHELIXLONG || pipelen == NCELLHELIX31A) {
		    if (ITR_marvin_iol.dumpRaw.value)
			ITR_marvin_dumpChamber(chamber);

		    rc = ITR_marvin_calculate_threshold(chamber, nevmon, ITR_eventcount);

		    clusterfinder(chamber, celllong);
		    clustering_done = 1;
		} else if (ITR_marvin_iol.debug_mode.value)
		    aframe_mess(routine, GRR_SWARNING,
				"chamber 0x%x: pipelinelength %d not valid",
				chamber->chamberid, pipelen);
	    }
	}			/* end for(..ichamber..)-loop */
    } /* end-if Nchip > 0 ? */
    else if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO, "# of chips = 0!");

    /* compute mean electronic gain */
    if ((ITR_marvin_iol.gaincorrection.value) &&
	(evseen % ITR_marvin_iol.pedUpdateRate.value == 0)) {
	meanGain = ITR_marvin_computeMeanColNo();
    }

    /* write cluster information to disk */
    if (sparse_file)
	ITR_marvin_writeSevt(sparse_file);


    /* dump clusters */
    if (ITR_marvin_iol.dumpClusters.value)
	ITR_marvin_dumpClusters();


    return GRR_OK;
}


/*-----------------------------------------------------------------------------
 *
 * End of run processing
 * =====================
 *
 * Author: Michael Schmelling / 21-May-1998
 *
 *---------------------------------------------------------------------------*/
int ITR_marvin_endrun()
{
    ITR_Marvin_Event_t *marvin_evt;
    int i;
    char *routine = "ITR_marvin_endrun";

    /* close the sparse data file */
    if (sparse_file)
	fclose(sparse_file);

    /* fetch ITR_marvin_evt and loop over chips */
    marvin_evt = ITR_marvin_fetch_event();
    if (!marvin_evt) {
	aframe_mess(routine, GRR_SINFO, "no ITR_Marvin_Event_t found");
	return GRR_FISHY;
    }

    for (i = 0; i < marvin_evt->nchip; i++)
	ITR_marvin_update_db(marvin_evt->chip[i]);

    /* ITR_leda_update_calib(1,1); moved to endprogram */

    return GRR_OK;
}



int ITR_marvin_endprogram()
/*-----------------------------------------------------------------------------
 * Final cleanup before exiting the program
 * ========================================
 *
 * Author: Michael Schmelling / 21-May-1998
 *---------------------------------------------------------------------------*/
{
    /* --wg-- this call really does not belong here,
       but in a specialized monitoring process (dmon-like)
       ITR_leda_update_calib(1,1); */

    return GRR_OK;
}


/*============================================================================
 *                          utility functions 
 *===========================================================================*/

/* ----------------------
 * fetch ITR_marvin event
 * ---------------------- */
ITR_Marvin_Event_t *ITR_marvin_fetch_event()
{
    char *routine = "ITR_marvin_fetch_event";

    if (!marvin_evt_ptr)
	aframe_mess(routine, GRR_SINFO, "No ITR_Marvin_Event_t found");

    return marvin_evt_ptr;
}


/* ------------------------
 * declare sparsified event
 * ------------------------ */
void ITR_marvin_declare_event(ITR_Marvin_Event_t * marvin_evt)
{
    marvin_evt_ptr = marvin_evt;
}


/* ----------------------
 * fetch sparsified event
 * ---------------------- */
ITR_Marvin_Sevt_t *ITR_marvin_fetch_sevt()
{
    char *routine = "ITR_marvin_fetch_sevt";

    if (!marvin_sevt_ptr)
	aframe_mess(routine, GRR_SINFO, "No ITR_Marvin_Sevt_t found");

    return marvin_sevt_ptr;
}


/* ------------------------
 * declare sparsified event
 * ------------------------ */
void ITR_marvin_declare_sevt(ITR_Marvin_Sevt_t * sevt)
{
    marvin_sevt_ptr = sevt;
}


/* ------------------------------------------------------------------------
 *
 *  Fill structures from chip, chamber, and bondmap tables stored with LEDA
 *  =======================================================================
 *
 *  Author: Rainer Wanke / 9-May-1999
 *
 * ------------------------------------------------------------------------ */
int ITR_marvin_readSetup()
{
    ITR_Marvin_Event_t *marvin_evt;
    ITR_Marvin_Chamber_t *chamber;
    ITR_Marvin_Chip_t *chip;

    static int readSetup_called = 0;

    int ncells;

    int ichip, icell, i, k;
    int ichan;

    LedaTableIterator(ITR_Chamber, chamberIt);
    char *routine = "ITR_marvin_readSetup";

/*---------------------------------------------------------------------------*/

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO, "Accessing ITR chip and chamber");


    /* Fetch event and do initializations */
    marvin_evt = ITR_marvin_fetch_event();
    if (!marvin_evt)
	return GRR_FISHY;


    /* wg 24jul00:  if readSetup was called before, clean up space *alloc()ed 
                    previously 
    */
    
    if (readSetup_called) {
      ITR_marvin_cleanSetup (marvin_evt);
    }

    readSetup_called = 1;

    marvin_evt->nchamber = 0;
    marvin_evt->nchip = 0;

    /* Loop over leda chambers
     * ----------------------- */
    for (chamberIt.begin(); chamberIt.valid(); chamberIt.next()) {

	LedaTableRow(ITR_Chamber) * ledachamber = chamberIt.get();
	LedaRelationIterator(ledachamber->chip, chipIt);

	/* add chamber to ITR_marvin_evt and initialize fields */
	chamber = malloc(sizeof(ITR_Marvin_Chamber_t));
	if (!chamber) {
	    aframe_mess(routine, GRR_SERROR,
			"Could not allocate space for chamber %d",
			ledachamber->id);
	    continue;
	}
	chamber->chamberid = ledachamber->id;
	chamber->chamberstatus = 0;
	chamber->nchan = 0;
	chamber->nchip = 0;
	chamber->first = ITRNCHAN_CHAMBER;
	chamber->last = 0;
	chamber->chambertype = ledachamber->type;


	/*WARNING:  (VB)
	   At this point all chambers from leda are known but in the next loop
	   over the chips therea are only chambers known which have chips in leda!
	   Because of this a loop over the chambers in marvin and in leda can deliver  
	   different chambers (for example in itrrhp).
	   So you must check the chamberIDs ... */


	/* loop over chips of that chamber
	 * ------------------------------- */
	for (chipIt.begin(); chipIt.valid(); chipIt.next()) {
	    LedaTableRow(ITR_Chip) * ledachip = chipIt.get();
	    LedaTableRow(ITR_Chanstat) * ledachanstat =
		LedaRelation(ledachip->chanstat);
	    LedaTableRow(ITR_Ped) * ledaped =
		LedaRelation(ledachip->pedestals);
	    LedaTableRow(ITR_Noise) * ledanoise =
		LedaRelation(ledachip->noise);

	    /* add chip to ITR_marvin_evt and initialize fields */
	    chip = malloc(sizeof(ITR_Marvin_Chip_t));
	    if (!chip) {
		aframe_mess(routine, GRR_SERROR,
			    "Could not allocate space for chip 0x%x",
			    ledachip->id);
		continue;
	    }
	    chip->chipid = ledachip->id;
	    chip->chamberid = ledachamber->id;
	    chip->chamber = (ITR_Marvin_Chamber_t *) chamber;
	    chip->chipstatus = (ledachip->status == 1) ? 0 : -1;
	    chip->pipelen = ledachip->pipelen;
	    chip->nchan = NCHIPCHAN;


	     /*VB*/ if (chip->pipelen == NCELLHELIXSHORT) {
		chip->ncellp = ITR_marvin_iol.ncellshortp.value;
		chip->ncellr = ITR_marvin_iol.ncellshortr.value;
		chip->ncells = ITR_marvin_iol.ncellshorts.value;
	    } else if (chip->pipelen == NCELLHELIXLONG || chip->pipelen == NCELLHELIX31A) {
		chip->ncellp = ITR_marvin_iol.ncelllongp.value;
		chip->ncellr = ITR_marvin_iol.ncelllongr.value;
		chip->ncells = ITR_marvin_iol.ncelllongs.value;
	    } else {
		aframe_mess(routine, GRR_SINFO,
			    "false pipelinelength %d for chip %d",
			    chip->pipelen, chip->chipid);
		continue;	/*VB:  chip markieren als falsch?? */
	    }

	    /* --wg--:
	       hopefully the relation iterator runs the same way the database was filled ... */
	    chip->first = chamber->nchip * NCHIPCHAN;
	    chip->last = (chamber->nchip + 1) * NCHIPCHAN - 1;

	    /*VB */
	    /*aframe_mess(routine, GRR_SINFO,"Chip:%d \t Pipeline %d\n ",
	       chip->chipid,chip->pipelen); */

	    /* allocate initialized space for data arrays, etc. */
	    chip->strip = calloc(chip->nchan, sizeof(int));
	    chip->data = calloc(chip->nchan, sizeof(float));
	    chip->status = calloc(chip->nchan, sizeof(int));
	    if (!chip->strip || !chip->data || !chip->status) {
		aframe_mess(routine, GRR_SERROR,
			    "Could not allocate space for data arrays of chip 0x%x",
			    chip->chipid);
		continue;
	    }

	    for (ichan = 0; ichan < chip->nchan; ichan++) {
		chip->status[ichan] = ledachanstat->status[ichan];
		chip->strip[ichan] = chamber->nchip * NCHIPCHAN + ichan;
	    }

	    /* allocate initialized space for pedestals and noise */
	    chip->peds = calloc(chip->ncellp * chip->nchan, sizeof(float));
	    chip->rvar = calloc(chip->ncellr * chip->nchan, sizeof(float));
	    chip->svar = calloc(chip->ncells * chip->nchan, sizeof(float));
	    /* aargh (---wg---)!  
	       if you change structure definitions in the header, 
	       please check all code for consistency!  Compiler warnings like
	       "assignment makes integer from pointer without a cast"
	       should never be ignored ! 

	       chip->nevtp = calloc( chip->ncellp, sizeof(int) );
	       chip->nevtr = calloc( chip->ncellr, sizeof(int) );
	       chip->nevts = calloc( chip->ncells, sizeof(int) );
	     */
	    if (!chip->peds || !chip->rvar || !chip->svar
		/*|| !chip->nevtp || !chip->nevtr || !chip->nevts */ ) {
		aframe_mess(routine, GRR_SERROR,
			    "Could not allocate space for peds/noise of chip 0x%x",
			    chip->chipid);
		continue;
	    }


	    /* --------------------------------------------------------------- */
	    /* initialization for new processing                               */
	    /* tz 03/11/99 pedestals and noise from DB                         */
            /* wg 06/nov/01  if database values are available, use them!       */

	    for (i = 0; i < chip->nchan * chip->ncellp; i++) {
		chip->peds[i] = (ledaped)   ? ledaped->pedestal[i] : 0;
		chip->svar[i] = (ledanoise) ? ledanoise->subvar[i] : 250;
		chip->rvar[i] = (ledanoise) ? ledanoise->rawvar[i] : 250;
 
		/* catch infinites stored in database */
		if (!finite (chip->svar[i])) chip->svar[i] = 0.;
		
		chip->psum[i] = chip->peds[i];
                chip->svsum[i]= chip->svar[i];
                chip->svinv[i]= (chip->svar[i] > 1e-7) ? 1./chip->svar[i] : 0.;
	    }


	    for (i = 0; i < NCHAN; i++) {
		/* (chip->svinv)[i] = 0.; */
		/* (chip->svsum)[i] = 100.; */
		(chip->chan)[i + 2 * NCHAN] = 0.;
		/* (chip->pval)[i] = 0; */
		/* (chip->psum)[i] = 0; */
	    }
	    chip->nevts = chip->nevtp = 0;
	    chip->nevas = chip->nevap = 1;

	    for (k = 1, i = 0; i < NCHAN; i++)
		if (chip->status[i] <= 0)
		    k++;
	    chip->istat[0] = k;
	    if (k <= 128) {
		for (k = 1, i = 0; i < NCHAN; i++)
		    if (chip->status[i] <= 0)
			chip->istat[k++] = i;
	    }

	    /* --------------------------------------------------------------- */


	    /* update counts for the chip and its chamber */
	    marvin_evt->chip[marvin_evt->nchip] = chip;
	    marvin_evt->nchip++;

	    chip_ptr[chip->chipid] = chip;
	    chamber->chip[chamber->nchip] = chip;
	    chamber->nchan += chip->nchan;
	    chamber->nchip++;
	    if (chip->first < chamber->first)
		chamber->first = chip->first;
	    if (chip->last > chamber->last)
		chamber->last = chip->last;

	}			/* end loop over chips */

	/* report the number of channels per chamber */
	if (ITR_marvin_iol.debug_mode.value)
	    aframe_mess(routine, GRR_SINFO,
			"chamber=0x%x type=%d channels=%d from %d to %d",
			chamber->chamberid, chamber->chambertype,
			chamber->nchan, chamber->first, chamber->last);
	if (chamber->nchan > ITRNCHAN_CHAMBER) {
	    aframe_mess(routine, GRR_SWARNING,
			"chamber 0x%x channels %d !!", chamber->chamberid,
			chamber->nchan);
	    return GRR_FISHY;
	}

	/*VB wg segmentation fault */
	/*warum geht 0 nicht???    */
	if (chamber->nchan < 1) {
	    if (ITR_marvin_iol.debug_mode.value)
		aframe_mess(routine, GRR_SWARNING,
			    "chamber 0x%x channels %d !!",
			    chamber->chamberid, chamber->nchan);
	    continue;
	}

	/* if nchan < last, set nchan to last chan - 1 */
	if (chamber->nchan < chamber->last - 1) {
	    chamber->nchan = chamber->last - 1;
	    aframe_mess(routine, GRR_SINFO,
			"nchan of chamber %d raised to %d",
			chamber->chamberid, chamber->nchan);
	}

	/* Now set up pointer arrays to chamber data */
	chamber->data = calloc(chamber->nchan, sizeof(float *));
	/*VB???? !hardcodiert chip0!!! */
	ncells = chamber->chip[0]->ncells;
	chamber->dvar = calloc(ncells * chamber->nchan, sizeof(float *));


	if (!chamber->data || !chamber->dvar) {
	    aframe_mess(routine, GRR_SERROR,
			"Could not allocate space for data pointers of chamber 0x%x",
			chamber->chamberid);
	    continue;
	}

	/* loop over chips of this chamber and fill geo-pointers */
	for (ichip = 0; ichip < chamber->nchip; ichip++) {
	    for (i = 0; i < chamber->chip[ichip]->nchan; i++) {
		int strip = chamber->chip[ichip]->strip[i];
		chamber->data[strip] = chamber->chip[ichip]->data + i;

		for (icell = 0; icell < chamber->chip[ichip]->ncells;
		     icell++)
		    chamber->dvar[strip + icell * chamber->nchan]
			= chamber->chip[ichip]->svar +
			icell * chamber->chip[ichip]->nchan + i;
	    }
	}

	/* update counts for the chamber */
	marvin_evt->chamber[marvin_evt->nchamber] = chamber;
	marvin_evt->nchamber++;

    }				/* end loop chambers */

    return GRR_OK;
}



/*-----------------------------------------------------------------------------
 *
 * Return the pointer to a specific chip
 * =====================================
 *
 * Author: Rainer Wanke / 9-May-1998
 *
 *---------------------------------------------------------------------------*/
ITR_Marvin_Chip_t *ITR_marvin_findChip(int chipid)
{
    char *routine = "ITR_marvin_findChip";

    if (chipid >= ITR_MAXCHIPID) {
	aframe_mess(routine, GRR_SWARNING, "Chip ID 0x%x too large!",
		    chipid);
	return 0;
    }

    return chip_ptr[chipid];
}


/*-----------------------------------------------------------------------------
 *
 * Return the pointer to a specific chamber
 * =====================================
 *
 * Author: Rainer Wanke / 9-May-1998
 *
 *---------------------------------------------------------------------------*/
ITR_Marvin_Chamber_t *ITR_marvin_findChamber(int chamberid)
{
    ITR_Marvin_Event_t *marvin_evt;

    int i;
    char *routine = "ITR_marvin_findChamber";

    /* fetch event */
    marvin_evt = ITR_marvin_fetch_event();
    if (!marvin_evt) {
	aframe_mess(routine, GRR_SWARNING, "No ITR_marvin_evt found!");
	return 0;
    }

    /* loop over chips to find it */
    for (i = 0; i < marvin_evt->nchamber; i++)
	if (marvin_evt->chamber[i]->chamberid == chamberid)
	    return marvin_evt->chamber[i];

    /* chip id not found */
    aframe_mess(routine, GRR_SWARNING, "No chamber 0x%x found!",
		chamberid);
    return 0;
}


/*-----------------------------------------------------------------------------
 *
 * Full processing for one readout chip
 * ====================================
 *   o pedestal subtraction
 *   o pedestal following
 *   o baseline subtraction
 *   o subtracted noise following
 *
   use a new pedestal and noise following scheme:

   a) get sums of the required quantity for 1,2,4,.... 2^n events
   b) whenever a power of two is reached, evaluate the quantity
   c) when nrset is reached, start accumulation anew until nrset
   d) average subsequent sums
   
   the variances stored for comparison with the signal are stored in 
   units of the SEEDCUT 

   The data array will hold d*abs(d) which can be compared to the 
   variances directly 

   Outliers are treated such, that pedestal and noise values are
   assumed to be equal to the current estimates

 * Author: Michael Schmelling 
 *---------------------------------------------------------------------------*/

int ITR_marvin_processChip(ITR_Marvin_Chip_t * chip, int cell)
{
    ITR_Marvin_Chamber_t *chamber;

    unsigned char *digi = chip->digi;
    float *pval  = chip->pval;
    float *psum  = chip->psum;
    float *svsum = chip->svsum;
    float *svinv = chip->svinv;
    float *chi2  = chip->chan;
    float *data  = chip->data;
    float *svar  = chip->svar;
    int   *istat = chip->istat;

    float CUT = 10;
    float wt;
    static int outlier[NCHAN];
    int i, index, ndrop;

    union {
	float f;
	int i;
    } cut;
    union {
	float f;
	int i;
    } value;

    char *routine = "ITR_marvin_processChip";
    /*===================================================*/

    /* Is chamber existing and ok ? */
    chamber = (ITR_Marvin_Chamber_t *) chip->chamber;
    if (!chamber) {
	aframe_mess(routine, GRR_SWARNING,
		    "Unknown detector chamber=<0x%x>", chip->chamberid);
	return GRR_FISHY;
    }
    if (chamber->chamberstatus < 0)
	return GRR_OK;

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO, "processing chip=%d on side=%d",
		    chip->chipid, chip->chamberid);

    /* unpack digits, subtract pedestals */

    (chip->nevtp)++;
    for (i = 0; i < NCHAN; i++) {
	psum[i] += data[i] = (float) (digi[i]);
	data[i] -= pval[i];
    }

    /* baseline subtraction with outlier removal */

    ndrop = ITR_marvin_subtractSpline(istat, data, CUT, outlier);

    /* subtracted noise following */

    cut.f = ITR_marvin_iol.seedcut.value;

    (chip->nevts)++;
    for (i = 0; i < NCHAN; i++) {
	value.f = data[i];
	value.f *= value.f;
	svsum[i] += value.f;
	value.f *= svinv[i];
	if (value.i < cut.i) {
	    chi2[i] = 0;
	} else {
	    if ((*(int *) (data + i)) < 0)
		chi2[i] = -value.f;
	    else
		chi2[i] = value.f;
	}
    }

    /* set chi2 for dead channels */

    index = istat[0];
    if (index > 1) {
	value.f = 0;		/* value.f=chip->side->polarity */
	for (i = 1; i < index; i++)
	    chi2[istat[i]] = value.f;
    }

    /* subtract outlier signals from subtracted variance */

    for (i = 0; i < ndrop; i++) {
	index = outlier[i];
	value.f = data[index];
	value.f *= value.f;
	cut.f = 25. * svar[index];
	if (cut.i == 0) {
	    cut.f = 250;
	    if (value.i > cut.i)
		svsum[index] += 20. - value.f;
	} else if (value.i > cut.i)
	    svsum[index] += 0.08 * cut.f - value.f;
    }

    /* deal with pedestal following */

    if (chip->nevtp == chip->nevap) {
	wt = 1. / (float) (chip->nevtp);
	for (i = 0; i < NCHAN; i++) {
	    pval[i] = 0.5 * pval[i] + wt * psum[i];
	    psum[i] = 0;
	}
	if (chip->nevap > NEVMONP)
	    chip->nevtp = NEVMONP;
	else
	    chip->nevap *= 2;
    }

    /* deal with subtracted variance following */

    if (chip->nevts == chip->nevas) {
	wt = 1. / (float) (chip->nevts);
	for (i = 0; i < NCHAN; i++) {
	    svar[i] = 0.5 * svar[i] + wt * svsum[i];
	    svinv[i] = 1. / svar[i];
	    svsum[i] = 0;
	}
	if (chip->nevas > NEVMONV)
	    chip->nevts = NEVMONV;
	else
	    chip->nevas *= 2;
    }


    /* update the range indices side->first/last and the sidestatus */

    if (chip->first < chamber->first)
	chamber->first = chip->first;
    if (chip->last > chamber->last)
	chamber->last = chip->last;
    chamber->chamberstatus = 1;

    return GRR_OK;

}



/*-----------------------------------------------------------------------------
 *
 * minimized processing for one readout chip with sparsified data
 * ==============================================================
 *
 * Since sparsified data is already pedestal and baseline corrected,
 * we don't need to do this again here.
 *
 * Author: Wolfgang Gradl  5/apr/00
 *
 *---------------------------------------------------------------------------*/
int ITR_marvin_processSparsifiedChip(ITR_Marvin_Chip_t * chip, int cell)
{

    ITR_Marvin_Chamber_t *chamber;

    /* static int status[128]; */
    float *peds, *rvar, *svar;
    int nevtp, nevtr, nevts, nevmon;
    int i, nchan;

    char *routine = "ITR_marvin_processSparsifiedChip";


    /* check the obvious */
    if (!chip) {
	aframe_mess(routine, GRR_SWARNING, "No chip given!");
	return GRR_FISHY;
    }

    nchan = chip->nchan;
    if (nchan != NCHAN_CHIP) {
	aframe_mess(routine, GRR_SWARNING,
		    "invalid number of channels <%d> for chip <0x%x> => exit",
		    nchan, chip->chipid);
	return GRR_FISHY;
    }

    /* Is chamber existing and ok ? */
    chamber = (ITR_Marvin_Chamber_t *) chip->chamber;
    if (!chamber) {
	aframe_mess(routine, GRR_SWARNING,
		    "Unknown detector chamber=<0x%x>", chip->chamberid);
	return GRR_FISHY;
    }
    if (chamber->chamberstatus < 0)
	return GRR_OK;

    /* initialize local pointers */
    peds = chip->peds + (cell % chip->ncellp) * nchan;
    rvar = chip->rvar + (cell % chip->ncellr) * nchan;
    svar = chip->svar + (cell % chip->ncells) * nchan;

    nevtp = chip->nevtp;
    nevtr = chip->nevtr;
    nevts = chip->nevts;

    nevmon = ITR_marvin_iol.nevmon.value;

    /* pedestal following */
    if (ITR_marvin_iol.pedUpdateRate.value) {
	(nevtp)++;
	ITR_marvin_followPedestals(chip, peds, nevmon, nevtp);
    }

    /* raw noise following */
    if (ITR_marvin_iol.rvarUpdateRate.value) {
	(nevtr)++;
	ITR_marvin_followRawNoise(chip, rvar, nevmon, nevtr);
    }

    /* subtracted noise following */
    if (ITR_marvin_iol.svarUpdateRate.value) {
	(nevts)++;
	ITR_marvin_followSubtractedNoise(chip, svar, nevmon, nevts);
    }

    for (i = 0; i < chip->nchan; i++)
	chip->data[i] = (float) ((signed char) chip->digi[i]);

    /* detction of hot channels */
    ITR_marvin_detect_hotchannel(chip, rvar, nevmon, nevtr);

    /* update the range indices chamber->first/last and the chamberstatus */
    if (chip->first < chamber->first)
	chamber->first = chip->first;
    if (chip->last > chamber->last)
	chamber->last = chip->last;
    chamber->chamberstatus = 1;

    return GRR_OK;
}



/* pedestal subtraction ------------------------------------------------*/
/* VB: Sep99 if chanstat = 0 (hot/dead) set chip->data=0                */
int ITR_marvin_subtractPedestals(ITR_Marvin_Chip_t * chip, float *peds)
{
    int i;

    for (i = 0; i < chip->nchan; i++) {
	chip->data[i] = chip->digi[i] - peds[i];

	/* move this out of the loop */
	if (chip->status[i] <= 0) {
	    chip->data[i] = 0;
	    chip->peds[i] = 0;
	}
    }
    return GRR_OK;
}


/* pedestal following ------------------------------------------------*/
int ITR_marvin_followPedestals(ITR_Marvin_Chip_t * chip, float *peds,
			       int nevmon, int nevtp)
{
    int i;
    static float wt;

    if (!nevmon)
	return GRR_OK;

    if (nevtp < nevmon)
	wt = 1. / (float) nevtp;
    else if (nevtp % ITR_marvin_iol.pedUpdateRate.value)
	return GRR_OK;

    for (i = 0; i < chip->nchan; i++)
	peds[i] += wt * chip->data[i];

    return GRR_OK;
}


/* pedestal subtraction AND following ----------------------------------------*/
/* at the moment not used ; the functions followPed and subtractPed do the same*/
/* VB: Sep99 if chanstat = 0 (hot/dead) set chip->data=0 */
int ITR_marvin_doPedestals(ITR_Marvin_Chip_t * chip, float *peds,
			   int nevmon, int nevtp)
{
    int i;
    static float wt;

    if (!nevmon)
	return GRR_OK;

    if (nevtp < nevmon)
	wt = 1. / (float) nevtp;

    for (i = 0; i < chip->nchan; i++) {
	chip->data[i] = chip->digi[i] - peds[i];
	if (chip->status[i] <= 0) {
	    chip->data[i] = 0;
	    chip->peds[i] = 0;
	}
	peds[i] += wt * chip->data[i];
    }

    return GRR_OK;
}


/*  raw noise following -----------------------------------------------
 * 
 *  note: there is no outlier removal for the raw noise
 *        since that figure should fluctuate much for live
 *        channels and only very little for dead ones. Without 
 *        outlier removal that discrimination is more pronounced. */
int ITR_marvin_followRawNoise(ITR_Marvin_Chip_t * chip, float *rvar,
			      int nevmon, int nevtr)
{
    int i;
    static float wt;

    if (!nevmon)
	return GRR_OK;

    if (nevtr < nevmon)
	wt = 1. / (float) nevtr;
    else if (nevtr % ITR_marvin_iol.rvarUpdateRate.value)
	return GRR_OK;

    for (i = 0; i < chip->nchan; i++)
	rvar[i] += wt * (chip->data[i] * chip->data[i] - rvar[i]);

    return GRR_OK;
}


/*  subtracted noise following------------------------------------------
 *  
 *  note: here some kind of outlier removal is needed, since 
 *        this noise figure is used in the clustering step to
 *        find signals. Consider the scenario of S/N=20 and 
 *        5% signal probability. Without filtering the estimated
 *        noise variance in 1000 events would be 20.950, with 
 *        filtering it would be 0.950. Although filtering biases
 *        the noise to small values, the error is small compared
 *        to the bias one would have without filtering.          */
int ITR_marvin_followSubtractedNoise(ITR_Marvin_Chip_t * chip, float *svar,
				     int nevmon, int nevts)
{
    static float varcut = 9.;	/* variance cut for noise filtering    */
    static float maxSignal;	/* max. signal allowed to calculate noise */
    static float wt;

    int i;
    float dvar;

    maxSignal = ITR_marvin_iol.svar_maxSignal.value;

    if (!nevmon)
	return GRR_OK;

    if (nevts < nevmon) {
	wt = 1. / (float) nevts;
	for (i = 0; i < chip->nchan; i++)
	    if (chip->data[i] * chip->data[i] < maxSignal)
		svar[i] += wt * (chip->data[i] * chip->data[i] - svar[i]);
    }

    else if (nevts % ITR_marvin_iol.svarUpdateRate.value)
	return GRR_OK;

    else {
	for (i = 0; i < chip->nchan; i++) {
	    dvar = chip->data[i] * chip->data[i] - svar[i];
	    if (dvar < varcut * svar[i])
		svar[i] += wt * dvar;
	}
    }

    return GRR_OK;
}


/* Check the smoothness of corrected data:---------------------
 *   If average S/N is above smooth_cut, chip is rejected
 *
 * return value: 1 = chip rejected by smooth-cut
 *               0 = chip accepted                */
int ITR_marvin_smoothCut(ITR_Marvin_Chip_t * chip, float *svar, int nevmon,
			 int nevts)
{
    int i;
    int j = 0;
    float dvar = 0.;
    static float svarcut = 0.1;

    char *routine = "ITR_marvin_smoothCut";

    for (i = 0; i < chip->nchan; i++) {
	if (chip->status[i] > 0 && svar[i] > svarcut) {
	    j++;
	    if (chip->data[i] > 0)
		dvar += chip->data[i] / svar[i];
	    else
		dvar -= chip->data[i] / svar[i];
	}
    }
    if (j > 0)
	dvar = (float) dvar / j;

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO, "Chip %4x: 'Smoothness' = %f",
		    chip->chipid, dvar);

    if (dvar > ITR_marvin_iol.smooth_cut.value && nevts >= nevmon)
	return 1;

    return 0;
}


/* Check the flatness of corrected data:-------------------------------
 *   If average S/N for certain regions is looked at
 * Regions are:  5 -> 24, 54 -> 83, 103 -> 122
 *
 * return value: 1 = chip rejected by flat-cut
 *               0 = chip accepted                */
int ITR_marvin_flatCut(ITR_Marvin_Chip_t * chip, float *svar, int nevmon,
		       int nevts)
{
    int i;
    int j1 = 0, j2 = 0, j3 = 0;
    float dvar1 = 0., dvar2 = 0., dvar3 = 0.;

    char *routine = "ITR_marvin_flatCut";

    for (i = 5; i < 25; i++)
	if (chip->status[i] > 0 && svar[i] > 0.1) {
	    j1++;
	    dvar1 += chip->data[i] / svar[i];
	}
    if (j1 > 0)
	dvar1 /= (float) j1;
    for (i = 54; i < 74; i++)
	if (chip->status[i] > 0 && svar[i] > 0.1) {
	    j2++;
	    dvar2 += chip->data[i] / svar[i];
	}
    if (j2 > 0)
	dvar2 /= (float) j2;
    for (i = 103; i < 123; i++)
	if (chip->status[i] > 0 && svar[i] > 0.1) {
	    j3++;
	    dvar3 += chip->data[i] / svar[i];
	}
    if (j3 > 0)
	dvar3 /= (float) j3;

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "Chip %4x: 'Flatness' = %f, %f, %f", chip->chipid,
		    dvar1, dvar2, dvar3);

    if (dvar1 < 0)
	dvar1 = -dvar1;
    if (dvar2 < 0)
	dvar2 = -dvar2;
    if (dvar3 < 0)
	dvar3 = -dvar3;
    if ((dvar1 > ITR_marvin_iol.flat_cut.value
	 || dvar2 > ITR_marvin_iol.flat_cut.value
	 || dvar3 > ITR_marvin_iol.flat_cut.value) && nevts >= nevmon) {

	aframe_mess(routine, GRR_SWARNING,
		    "Cutting Chip %4x, chamber %d: Flatness = %f, %f, %f",
		    chip->chipid, chip->chamberid, dvar1, dvar2, dvar3);

	return 1;
    }

    return 0;
}


/* column number height following --------------------------------------
 *  
 * Author: Wolfgang Gradl
 *
 *---------------------------------------------------------------------*/
int ITR_marvin_followColumnNumber(ITR_Marvin_Chip_t * chip, int nevmon,
				  int nevtp)
{
    static float wt;
    float dph;

    if (!nevmon)
	return GRR_OK;

    if (nevtp < nevmon)
	wt = 1. / (float) nevtp;

    else if (nevtp % ITR_marvin_iol.pedUpdateRate.value)
	return GRR_OK;

    /* p_n =  p_(n-1)           +   dph        * wt */
    dph = chip->hcolno - chip->hcolno_mean;
    chip->hcolno_mean += wt * dph;

    return GRR_OK;
}


/*----------------------------------------------------------------------
 * 
 * compute mean height of column number for all chips
 * only chips with a mean column number height of at least 
 * ITR_marvin_iol.gc_MinHeight.value are taken into account
 *
 * RETURNS:  mean height of column number
 * 
 * Author:  Wolfgang Gradl
 *----------------------------------------------------------------------*/
float ITR_marvin_computeMeanColNo(void)
{
    int numChip = 0, cnt;
    static float gainsum = 0.0;
    ITR_Marvin_Chip_t *marvinChip;

    gainsum = 0.0;
    for (cnt = 0; cnt < ITR_MAXCHIPID; cnt++) {
	if ((marvinChip = ITR_marvin_findChip(cnt)) != NULL) {
	    if (marvinChip->hcolno_mean > ITR_marvin_iol.gc_MinHeight.value
		&& marvinChip->hcolno_mean < 120.) {
		numChip++;
		gainsum += marvinChip->hcolno_mean;
	    }
	}
    }

    if (numChip)
	return (gainsum / numChip);
    else
	return (1.0);
}




/* ------------------------------------------------------------
 *
 * Return 0, if marvin is still accumulating pedestals/noise
 *        1, if marvin did clusterfinding in this event
 *
 * ------------------------------------------------------------ */
int ITR_marvin_ClusteringDone()
{
    return clustering_done;
}


/*-----------------------------------------------------------------------------
 *
 * Pack cluster information into the sparse-event struct  
 * =====================================================
 *
 * Author: Michael Schmelling / 14-Oct-1998
 * VB: July 99: cog added and format 4 with localmean
 * VB: Nov  99: physical width but "length" (=end-start) for storage space
 * WG: Apr  00: re-introduced sparse format 3 for online-sparsification
 *---------------------------------------------------------------------------*/
int ITR_marvin_packSevt(int chamberid, int pwi, int start, int end,
			float **d, float **v,
			int width, float *cog, float *charge,
			float *localmean)
{
    ITR_Marvin_Sevt_t *sevt;

    int i, oldsize, newsize = 0, ncltot, length;
    static int oldminor = -1;
    int izero, ione;
    float variance, strip, f4sn;
    unsigned char bwidth, b4sn, bcharge;
    short schamber;
    unsigned short swidth;
    const int csize = sizeof(char);
    const int isize = sizeof(int);

    char *routine = "ITR_marvin_packSevt";

  /*-------------------------------------------------------------------------*/

    /* get sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no sparse event buffer");
	return GRR_FISHY;
    }

    length = end - start;

    /* report cluster */
    if (ITR_marvin_iol.debug_mode.value) {
	aframe_mess(routine, GRR_SINFO,
		    "Found cluster chamber 0%x, strips %d -> %d",
		    chamberid, start, end);
	for (i = 0; i < length; i++)
	    aframe_mess(routine, GRR_SINFO, " %4d: Digit = %f, Var = %f",
			i, *(*(d + i)), *(*(v + i)));
    }

    /* first check buffer size  */
    oldsize = (sevt->next) - (sevt->data);

    /*VB: format 4 added... */
    /*this format is: int id, int (physical)width, 
       int beginofphysclauster, int beginofwholecluster, int endofwholecluster, 
       float localmean, float cog, float charge, float data[i], float variance[i] */
    switch (sevt->format) {
    case 1:
      	newsize = oldsize + 4 * sizeof(int) + 2 * width * sizeof(float);
	break;
    case 2:
	newsize = oldsize + 3 * sizeof(int) + 3 * sizeof(float);
	break;
    case 3:
        newsize = oldsize + sizeof(float) + 3 * sizeof(unsigned char) +
	  sizeof(short int);
	break;
    case 4:
	newsize = oldsize + 5 * sizeof(int) + 3 * sizeof(float) +
	    2 * length * sizeof(float);
	break;
    case 5:
	newsize = oldsize + sizeof (float) + 2 * sizeof (unsigned char) + 
	    2 * sizeof (short int);
	break;
    }



    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "sevt size = %d, maxsize = %d, format = %d", oldsize,
		    sevt->size, sevt->format);


    /* report full sevt buffer only once in a event ! */
    if (newsize > sevt->size) {
	ITR_Silraw_Event_t *silraw = ITR_silraw_fetch_event();
	if (!silraw) {
	    aframe_mess(routine, GRR_SINFO,
			"no ITR_Silraw_Event_t found!");
	    return GRR_FISHY;
	}
	if (silraw->header.minor != oldminor) {
	    aframe_mess(routine, GRR_SWARNING,
			"---------------------------------------");
	    aframe_mess(routine, GRR_SWARNING,
			"!!! Cluster buffer full (Evt %d) !!!",
			silraw->header.minor);
	    aframe_mess(routine, GRR_SWARNING,
			"---------------------------------------");
	    oldminor = silraw->header.minor;
	}
	return GRR_OK;
    }


    /* increase cluster count */
    /*VB +5 statt 4 wg diff pipeline in Marvin_Header */
    ESP_get_int((int *) sevt->data + 5, &ncltot, 1);
    ncltot++;
    ESP_put_int(&ncltot, (int *) sevt->data + 5, 1);


    /* fill the sparse data record */
    if (sevt->format == 1) {	/* the original sparse data format */
	ESP_put_int(&chamberid, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&width, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&start, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&izero, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	for (i = 0; i < width; i++) {
	    ESP_put_int((int *) d[i], (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	}
	for (i = 0; i < width; i++) {
	    ESP_put_int((int *) v[i], (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	}
    }

    /* wg 19/apr/00
     * re-introduced sparse format 3 a la VDS:
     *   uchar  chamberid     // one byte chamber identifier
     *   uchar  charge        // cluster charge
     *   uchar  sn4           // 4 * signal/noise
     *   uchar  width         // cluster width
     *   float  cog           // center of gravity
     */

    else if (sevt->format == 3) {
	*charge = *cog = 0;
	strip = variance = 0;
	for (i = 0; i < length; i++) {
	    *charge += *(d[i]);
	    strip += *(d[i]) * i;
	    variance += *(v[i]);
	}
	*charge = fabs(*charge);
	strip = fabs(strip);

	if (*charge)
	    *cog = start + strip / (*charge);
	else
	    *cog = -1.;

	if (variance) {
	    f4sn = 1. / sqrt(variance);
	    f4sn *= 4. * (*charge);
	} else
	    f4sn = 0.;

	bcharge = *charge > 254. ? 255 : (int) *charge;
	bwidth  =  width  > 127  ? 127 : width;
	b4sn    =  f4sn   > 254. ? 255 : (int) f4sn;
	schamber = (short int) (0xFFFF & chamberid);

	ESP_put_short(&schamber, (short int *) sevt->next, 1);
	sevt->next += sizeof(short int);

	*(sevt->next) = bcharge;
	sevt->next += csize;
	*(sevt->next) = b4sn;
	sevt->next += csize;
	*(sevt->next) = bwidth;
	sevt->next += csize;

	ESP_put_int((int *) (cog), (int *) sevt->next, 1);
	sevt->next += isize;
    }

    else if (sevt->format == 5) {
	*charge = *cog = 0;
	strip = variance = 0;
	for (i = 0; i < length; i++) {
	    *charge += *(d[i]);
	    strip += *(d[i]) * i;
	    variance += *(v[i]);
	}
	*charge = fabs(*charge);
	strip = fabs(strip);

	if (*charge)
	    *cog = start + strip / (*charge);
	else
	    *cog = -1.;

	if (variance) {
	    f4sn = 1. / sqrt(variance);
	    f4sn *= 4. * (*charge);
	} else
	    f4sn = 0.;

	bcharge = *charge > 254. ? 255 : (int) *charge;
	bwidth  =  width  > 127  ? 127 : width;
	b4sn    =  f4sn   > 254. ? 255 : (int) f4sn;
	schamber = (short int) (0xFFFF & chamberid);
	swidth  =  ((start & 0x7FF) << 5) | (bwidth & 0x1F);

	ESP_put_short(&schamber, (short int *) sevt->next, 1);
	sevt->next += sizeof(short int);

	*(sevt->next) = bcharge;
	sevt->next += csize;
	*(sevt->next) = b4sn;
	sevt->next += csize;

	ESP_put_short (&swidth, (short int*) sevt->next, 1);
	sevt->next += sizeof (short int);

	ESP_put_int((int *) (cog), (int *) sevt->next, 1);
	sevt->next += isize;
    }


    else if (sevt->format == 4) {	/*VB: the ITR format */
	if (*charge == 0. && *cog == 0.) {
	    strip = variance = 0;
	    for (i = 0; i < length; i++) {
		*charge += *(d[i]);
		strip += *(d[i]) * i;
		variance += *(v[i]);
	    }
	    if (*charge)
		*cog = start + strip / (*charge);
	    else
		*cog = -1.;
	}

	ESP_put_int(&chamberid, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&width, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&pwi, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&start, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int(&end, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int((int *) localmean, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int((int *) cog, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	ESP_put_int((int *) charge, (int *) sevt->next, 1);
	sevt->next += sizeof(int);
	for (i = 0; i < length; i++) {
	    ESP_put_int((int *) d[i], (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	}
	for (i = 0; i < length; i++) {
	    ESP_put_int((int *) v[i], (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	}
    }

    else {			/* deal with the more compressed formats */
	*charge = strip = variance = 0;
	for (i = 0; i < length; i++) {
	    *charge += *(d[i]);
	    strip += *(d[i]) * i;
	    variance += *(v[i]);
	}

	/*VB: if cog not calculated in get_ITRCluster because of calling getNegativeClusters */
	if (*localmean == 0 && *cog == 0) {
	    if (*charge)
		*cog = start + strip / (*charge);
	    else
		*cog = -1.;
	}
	if (!(*charge))
	    *cog = -1.;


	if (sevt->format == 2) {
	    width = -width;	/*VB: in this format is width=end-start?? */
	    ESP_put_int(&chamberid, (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	    ESP_put_int(&ione, (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	    ESP_put_int((int *) (cog), (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	    ESP_put_int(&width, (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	    ESP_put_int((int *) (charge), (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	    ESP_put_int((int *) (&variance), (int *) sevt->next, 1);
	    sevt->next += sizeof(int);
	    width = -width;
	}

    }

    return GRR_OK;
}


int ITR_marvin_update_db(ITR_Marvin_Chip_t * chip)
/*-----------------------------------------------------------------------------
 * Update Pedestal and Noise values in LEDA structures
 * ===================================================
 * (insteda of outputChip)
 * Author: TZ
 *         last change     03/11/99
 *---------------------------------------------------------------------------*/
{
    int i;
    const char *routine = "ITR_marvin_update_db";
    LedaTableIterator(ITR_Chip, chipit);

    for (chipit.begin(); chipit.valid(); chipit.next()) {
	LedaTableRow(ITR_Chip) * ledachip = chipit.get();
	LedaTableRow(ITR_Ped) * ledaped;
	LedaTableRow(ITR_Noise) * ledanoise;
	if (ledachip->id == chip->chipid) {
	    ledaped = LedaRelation(ledachip->pedestals);
	    ledanoise = LedaRelation(ledachip->noise);

	    if (ledaped == NULL || ledanoise == NULL) {
		aframe_mess((char *) routine, GRR_SWARNING,
			    "no tables ITR_Noise or ITR_Ped found.  Skipping update");
		return GRR_FISHY;
	    }

	    for (i = 0; i < chip->nchan; i++) {
		ledaped->pedestal[i] = chip->pval[i];
		if (finite(chip->svar[i]))
		    ledanoise->subvar[i] = chip->svar[i];
		else 
		    ledanoise->subvar[i] = 0;
		ledanoise->rawvar[i] = chip->rvar[i];
	    }
	}
    }

    return GRR_OK;
}


void ITR_marvin_dumpChamber(ITR_Marvin_Chamber_t * chamber)
/*-----------------------------------------------------------------------------
 * Dump one Chamber
 * ================
 *
 * Author: Verena Buchholz / 3-Aug-1999
 *---------------------------------------------------------------------------*/
{
    int ichan;
    float **chamberdata;

    chamberdata = chamber->data;

    printf("\n\nDump MarvinChamber 0x%x\n", chamber->chamberid);
    printf("-----------------------\n");
    printf("Type: %d  Status: %d\n", chamber->chambertype,
	   chamber->chamberstatus);
    printf("Filled from strip %d to %d, No. of Channels %d\n",
	   chamber->first, chamber->last, chamber->nchan);
    printf("No. of Chips %d\n", chamber->nchip);
    for (ichan = chamber->first; ichan <= chamber->last; ichan++) {
	printf("%2.3f\t", *(chamberdata[ichan]));
    }
    printf("\n");
}


void ITR_marvin_dumpRaw(int level)
/*-----------------------------------------------------------------------------
 * Dump the current event from ITR_silraw
 * ==================================
 *
 * Author: Michael Schmelling / 21-May-1998
 *         last change     MS /  8-Sep-1998
 *         VB (different pipeline) / Aug-1999
 *---------------------------------------------------------------------------*/
{
    ITR_Silraw_Event_t *silraw;
    unsigned char *chipdata;
    ITR_Marvin_Chip_t *chip;

    int i, k, chipid;
    int output[NCHAN_CHIP];
    int run, event, nchan, nchip;
    int cellshort, celllong;

    char *routine = "ITR_marvin_dumpRaw";
/*---------------------------------------------------------------------------*/

    if (level == 0)
	return;

    silraw = ITR_silraw_fetch_event();
    if (!silraw) {
	aframe_mess(routine, GRR_SINFO,
		    "failed to get ITR_silraw event\n");
	return;
    }

    ITR_silraw_get_major(&run);
    ITR_silraw_get_minor(&event);
    ITR_silraw_get_cellshort(&cellshort);
    ITR_silraw_get_celllong(&celllong);
    ITR_silraw_get_nchan(&nchan);

    printf("run=%d  event=%d cellshort=%d celllong=%d nchan=%d \n",
	   run, event, cellshort, celllong, nchan);

    if (level == 1)
	return;

    ITR_silraw_get_nchip(&nchip);
    for (i = 0; i < nchip; i++) {
	chipid = silraw->chip[i].chipid;
	chipdata = silraw->chip[i].data;
	chip = ITR_marvin_findChip(chipid);
	if (chip) {
	    if (cellshort > NCELLHELIXSHORT) {
		aframe_mess(routine, GRR_SINFO,
			    "cellshort %d > short pipeline!", cellshort);
	    }
	    if (celllong > NCELLHELIXLONG) {
		aframe_mess(routine, GRR_SINFO,
			    "celllong %d > long pipeline!", celllong);
	    }
	    ITR_silraw_reformat_data(chipdata, nchan, output);
	    if (level > 1) {
		printf("chamber=0x%x chip#=%d chipid=0x%x",
		       chip->chamberid, i, chipid);
		for (k = 0; k < nchan; k++) {
		    if (k % 24 == 0)
			printf("\n");
		    printf("%4d", output[k]);
		}
		printf("\n");
	    }
	}
    }

    if (level > 1)
	printf("\n\n\n");

    return;
}



void ITR_marvin_dumpClusters()
/*-----------------------------------------------------------------------------
 * dump sparse-event information
 * =============================
 *
 * Author: Michael Schmelling / 21-May-1998
 *         last change     MS / 14-Oct-1998
 *         VB (different pipelines) / Aug-1999
 *---------------------------------------------------------------------------*/
{
    ITR_Marvin_Sevt_t *sevt;
    ITR_Marvin_Header_t *header;
    ITR_Marvin_Cluster_t *cluster;

    char *routine = "ITR_marvin_dumpClusters";

  /*-------------------------------------------------------------------------*/

    /* get sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no sparse event buffer");
	return;
    }

    header = ITR_marvin_sevtHeader(sevt->data);

    printf
	("major=%d minor=%5d cellshort=%5d celllong=%5d format=%2d ncl=%4d \n",
	 header->major, header->minor, header->cellshort, header->celllong,
	 header->format, header->ncl);

    cluster = ITR_marvin_sevtCluster(sevt->data, -1);
    while (cluster) {
	printf
	    ("chamber=0x%4x  charge=%8.2f  s/n=%6.2f width=%2d cog=%8.2f physical begin=%4d localmean=%2.2f\n",
	     cluster->chamber, cluster->charge, cluster->sn,
	     cluster->width, cluster->cog, cluster->pwi,
	     cluster->localmean);

	cluster = ITR_marvin_sevtCluster(NULL, -1);
    }

    return;
}


int ITR_marvin_subtractBaseline(int *status, float *y, int ny, int degree)
/*-----------------------------------------------------------------------------
 * polynomial baseline subtraction
 * ===============================
 * A polynomial based on all array elements with status!=0 is subtracted
 * from the data vector y[ny]. The degree of the polynomial is specified
 * by degree. The function supports fits for degree=0,1,2,3. For other 
 * values the code needs to be modified. A return value of 0 indicates
 * success, if the fit fails 1 is returned. In that case y[] is unchanged.
 * 
 * note: Without compiler optimization the code is about 10-15% slower than 
 *       equivalent code without subscripted variables. With optimization 
 *       (already at gcc -O1) the difference in performance drops to 2.5%.
 *
 * Author: Michael Schmelling / 22-Jul-1998
 *         last change     MS / 17-Sep-1998
 *---------------------------------------------------------------------------*/
{
    int i, j, k, n;
    float x, dx, yval;
    float sx[7], sy[4], a[4], b[4][4];

    static int nold = 0;
    static float *p1, *p2, *p3, *p4, *p5, *p6;
    static float sxall[7];

  /*-------------------------------------------------------------------------*/

    /* linear baseline subtraction ? */
    if (degree == 1)
	return ITR_marvin_subtractBaseline_lin(status, y, ny);

    if (degree == 0)
	return ITR_marvin_subtractBaseline_const(status, y, ny);

    if (degree < 0)
	return 0;

    if (degree > 3)
	n = 4;
    else
	n = degree + 1;

    /* 
     *  initialize cumulants and basis functions 
     */

    if (nold != ny) {
	if (nold == 0)
	    p1 = (float *) malloc(6 * ny * sizeof(float));
	else
	    p1 = (float *) realloc(p1, 6 * ny * sizeof(float));
	p2 = p1 + ny;
	p3 = p2 + ny;
	p4 = p3 + ny;
	p5 = p4 + ny;
	p6 = p5 + ny;
	x = -1.;
	dx = 2. / (float) ny;
	for (i = 0; i < 7; i++)
	    sxall[i] = 0.;
	for (i = 0; i < ny; i++) {
	    p1[i] = x;
	    p2[i] = x * p1[i];
	    p3[i] = x * p2[i];
	    p4[i] = x * p3[i];
	    p5[i] = x * p4[i];
	    p6[i] = x * p5[i];
	    sxall[0] += 1.;
	    sxall[1] += p1[i];
	    sxall[2] += p2[i];
	    sxall[3] += p3[i];
	    sxall[4] += p4[i];
	    sxall[5] += p5[i];
	    sxall[6] += p6[i];
	    x += dx;
	}
	nold = ny;
    }

    /* 
     *  get the actual cumulants 
     */

    for (j = 0; j < n; j++)
	sy[j] = 0.;
    for (j = 0; j < 2 * n - 1; j++)
	sx[j] = sxall[j];

    switch (n) {
    case 1:
	for (i = 0; i < ny; i++) {
	    if (status[i] > 0)
		sy[0] += y[i];
	    else
		sx[0] -= 1.;
	}
	break;
    case 2:
	for (i = 0; i < ny; i++) {
	    if (status[i] > 0) {
		sy[0] += y[i];
		sy[1] += y[i] * p1[i];
	    } else {
		sx[0] -= 1.;
		sx[1] -= p1[i];
		sx[2] -= p2[i];
	    }
	}
	break;
    case 3:
	for (i = 0; i < ny; i++) {
	    if (status[i] > 0) {
		yval = y[i];
		sy[0] += yval;
		sy[1] += yval * p1[i];
		sy[2] += yval * p2[i];
	    } else {
		sx[0] -= 1.;
		sx[1] -= p1[i];
		sx[2] -= p2[i];
		sx[3] -= p3[i];
		sx[4] -= p4[i];
	    }
	}
	break;
    case 4:
	for (i = 0; i < ny; i++) {
	    if (status[i] > 0) {
		yval = y[i];
		sy[0] += yval;
		sy[1] += yval * p1[i];
		sy[2] += yval * p2[i];
		sy[3] += yval * p3[i];
	    } else {
		sx[0] -= 1.;
		sx[1] -= p1[i];
		sx[2] -= p2[i];
		sx[3] -= p3[i];
		sx[4] -= p4[i];
		sx[5] -= p5[i];
		sx[6] -= p6[i];
	    }
	}
	break;
    default:
	return (1);
    }

    /* 
     *  then determine the fit parameters
     */

    /* 
     *  LU-decomposition of the cumulant matrix is performed
     *  by a straightforward implementation of Crout's method 
     *  no pivoting is performed, which should be ok for the low
     *  order polynomials considered in this application.
     *  The following piece of code also traps singular matrices. 
     */

    for (i = 0; i < n; i++) {
	for (j = 0; j < n; j++) {
	    b[i][j] = sx[i + j];
	    if (i > j) {
		for (k = 0; k < j; k++)
		    b[i][j] -= b[i][k] * b[k][j];
		b[i][j] /= b[j][j];
	    } else {
		for (k = 0; k < i; k++)
		    b[i][j] -= b[i][k] * b[k][j];
		if (i == j && b[i][j] == 0.)
		    return (1);
	    }
	}
    }

    /* solution via forward/backward substitution */

    for (i = 0; i < n; i++) {
	a[i] = sy[i];
	for (j = 0; j < i; j++)
	    a[i] -= b[i][j] * a[j];
    }
    for (i = n - 1; i >= 0; i--) {
	for (j = i + 1; j < n; j++)
	    a[i] -= b[i][j] * a[j];
	a[i] /= b[i][i];
    }

    /* 
     *  subtract the baseline 
     */

    switch (n) {
    case 1:
	for (i = 0; i < ny; i++)
	    y[i] -= a[0];
	break;
    case 2:
	for (i = 0; i < ny; i++)
	    y[i] -= (a[0] + a[1] * p1[i]);
	break;
    case 3:
	for (i = 0; i < ny; i++)
	    y[i] -= (a[0] + a[1] * p1[i] + a[2] * p2[i]);
	break;
    case 4:
	for (i = 0; i < ny; i++)
	    y[i] -= (a[0] + a[1] * p1[i] + a[2] * p2[i] + a[3] * p3[i]);
	break;
    default:
	return (1);
    }

    return (0);
}


int ITR_marvin_subtractBaseline_const(int *status, float *y, int ny)
/*-----------------------------------------------------------------------------
 *
 * constant baseline subtraction
 * ===============================
 *
 *---------------------------------------------------------------------------*/
{
    int i;
    float x, dx;
    float sx0, sy0, a0;

    static int nold = 0;
    static float *p1;
    static float sxall0;

  /*-------------------------------------------------------------------------*/

    /* 
     *  initialize cumulants and basis functions 
     */

    if (nold != ny) {
	if (nold == 0)
	    p1 = (float *) malloc(ny * sizeof(float));
	else
	    p1 = (float *) realloc(p1, ny * sizeof(float));
	x = -1.;
	dx = 2. / (float) ny;
	sxall0 = 0.;
	for (i = 0; i < ny; i++) {
	    p1[i] = x;
	    sxall0 += 1.;
	    x += dx;
	}
	nold = ny;
    }

    /* get the actual cumulants */

    sy0 = 0.;
    sx0 = sxall0;

    for (i = 0; i < ny; i++)
	if (status[i] > 0)
	    sy0 += y[i];
	else
	    sx0 -= 1.;

    /* then determine the fit parameters */

    /* 
     *  LU-decomposition of the cumulant matrix is performed
     *  by a straightforward implementation of Crout's method 
     *  no pivoting is performed, which should be ok for the low
     *  order polynomials considered in this application.
     *  The following piece of code also traps singular matrices. 
     */

    if (sx0 == 0.)
	return 1;

    /* solution via forward/backward substitution */

    a0 = sy0 / sx0;

    /* subtract the baseline */

    for (i = 0; i < ny; i++)
	y[i] -= a0;

    return 0;
}


int ITR_marvin_subtractBaseline_lin(int *status, float *y, int ny)
/*-----------------------------------------------------------------------------
 *
 * linear baseline subtraction
 * ===========================
 *
 * Just the n = 2 special case of the general algorithm above
 *
 *---------------------------------------------------------------------------*/
{
    int i;
    float x, dx;
    float sx0, sx1, sx2, sy0, sy1, a0, a1, b00, b01, b10, b11;

    static int nold = 0;
    static float *p1, *p2;
    static float sxall0, sxall1, sxall2;

  /*-------------------------------------------------------------------------*/

    /* 
     *  initialize cumulants and basis functions 
     */

    if (nold != ny) {
	if (!nold) {
	    p1 = (float *) malloc(ny * sizeof(float));
	    p2 = (float *) malloc(ny * sizeof(float));
	} else {
	    p1 = (float *) realloc(p2, ny * sizeof(float));
	    p2 = (float *) realloc(p2, ny * sizeof(float));
	}
	x = -1.;
	dx = 2. / (float) ny;
	sxall0 = 0.;
	sxall1 = 0.;
	sxall2 = 0.;
	for (i = 0; i < ny; i++) {
	    p1[i] = x;
	    p2[i] = x * p1[i];
	    sxall0 += 1.;
	    sxall1 += p1[i];
	    sxall2 += p2[i];
	    x += dx;
	}
	nold = ny;
    }

    /* 
     *  get the actual cumulants 
     */

    sy0 = 0.;
    sy1 = 0.;
    sx0 = sxall0;
    sx1 = sxall1;
    sx2 = sxall2;

    for (i = 0; i < ny; i++)
	if (status[i] > 0) {
	    sy0 += y[i];
	    sy1 += y[i] * p1[i];
	} else {
	    sx0 -= 1.;
	    sx1 -= p1[i];
	    sx2 -= p2[i];
	}

    /* then determine the fit parameters */

    /* 
     *  LU-decomposition of the cumulant matrix is performed
     *  by a straightforward implementation of Crout's method 
     *  no pivoting is performed, which should be ok for the low
     *  order polynomials considered in this application.
     *  The following piece of code also traps singular matrices. 
     */

    b00 = sx0;
    if (b00 == 0.)
	return 1;
    b01 = sx1;
    b10 = sx1 / b00;
    b11 = sx2 - b10 * b01;
    if (b11 == 0.)
	return 1;

    /* solution via forward/backward substitution */

    a0 = sy0;
    a1 = sy1;
    a1 -= b10 * a0;
    a1 /= b11;
    a0 -= b01 * a1;
    a0 /= b00;

    /* subtract the baseline */

    for (i = 0; i < ny; i++)
	y[i] -= a0 + a1 * p1[i];

    return 0;
}


ITR_Marvin_Cluster_t *ITR_marvin_sevtCluster(char *data, int chamber)
/*-----------------------------------------------------------------------------
 * General access function for sparsified event data 
 * =================================================
 * Given as input the pointer to the data-record of a sparse-event 
 * structure, the information of the first cluster is unpacked into a local 
 * ITR_Marvin_Cluster_t struct and the pointer to this structure returned to
 * the caller. Subsequent calls with NULL instead of *data then access the 
 * other clusters. When the list of clusters is exhausted NULL is returned.  
 * The "chamber" argument allows to select specific chambers, i.e. only clusters
 * on the selected chamber are returned. chamber==-1 accepts all chambers.  
 *
 * The function works transparently on all sparse data format. Note, 
 * however, that dependng on the format, not all fields of the cluster 
 * struct can be filled. (Clearly the sparse-data formats are defined
 * such that the most important data describing a hit are available.)  
 *
 * Author: Michael Schmelling / 14-Oct-1998 
 *         last change     MS / 26-Oct-1998
 * VB: selection for specific chambers not longer possible / Jul-1999
 * VB: new format 4 for ITR and changes for the new header / Aug-1999
 * VB: 8/11/99 cluster.width is now the physical width, length is the number 
 *              of strips for the whole cluster
 * VB: 21/11/99 now length and width are the same
 *---------------------------------------------------------------------------*/
{
    static ITR_Marvin_Cluster_t cluster;
    static int format, ncl, nextcl;
    static char *next;
    static float *digits = 0, *variances = 0;

    int i, length;
    unsigned char bwidth, bcharge, b4sn;
    short int schamber;
    unsigned short int swidth;
    float cog;

    ITR_Marvin_Header_t *header;

    char *routine = "ITR_marvin_sevtCluster";

/*---------------------------------------------------------------------------*/
    /*VB: if check fails set chamber to -1 */
    if (chamber != -1) {
	aframe_mess(routine, GRR_SINFO,
		    "chamber set to -1! this accepts all chambers ");
	chamber = -1;
    }

    /* 
     *  deal with the data!=0 initializing call 
     */
    if (data) {
	header = ITR_marvin_sevtHeader(data);
	format = header->format;
	ncl = header->ncl;
	next = data + kHeadersize;
	nextcl = 0;
    }



    /*
     *  loop to find the next cluster matching the filter condition
     */

    while (nextcl++ < ncl) {

	cluster.format = format;

	switch (format) {

	case 1:		/* the original variable length sd-format */

	    ESP_get_int((int *) next, &cluster.chamber, 1);
	    next += sizeof(int);
	    ESP_get_int((int *) next, &cluster.width, 1);
	    next += sizeof(int);
	    ESP_get_int((int *) next, &cluster.first, 1);
	    next += sizeof(int) * 2;
	    digits = realloc(digits, cluster.width * sizeof(int));
	    ESP_get_int((int *) next, (int *) digits, cluster.width);
	    cluster.digits = digits;
	    next += cluster.width * sizeof(float);
	    variances = realloc(variances, cluster.width * sizeof(int));
	    ESP_get_int((int *) next, (int *) variances, cluster.width);
	    cluster.variances = variances;
	    next += cluster.width * sizeof(float);
	    if (chamber == -1) {	/*VB: kann eigtl.weg */
		cluster.charge = 0;
		cluster.cog = 0;
		cluster.variance = 0;
		cluster.localmean = 0;	/*VB hinzugefuegt */
		for (i = 0; i < cluster.width; i++) {
		    cluster.charge += cluster.digits[i];
		    cluster.cog += cluster.digits[i] * (float) i;
		    cluster.variance += cluster.variances[i];
		}
		if (cluster.charge != 0) {
		    cluster.cog =
			(float) cluster.first +
			cluster.cog / cluster.charge;
		}
		if (cluster.variance > 0) {
		    cluster.sn = cluster.charge;
		    cluster.sn *=
			sqrt((float) cluster.width / cluster.variance);
		} else {
		    cluster.sn = 0;
		}
		return &cluster;
	    }
	    break;


	case 2:		/* the 192 bit/cluster format */

	    ESP_get_int((int *) next, &cluster.chamber, 1);
	    next += sizeof(int) * 2;
	    ESP_get_int((int *) next, (int *) (&cluster.cog), 1);
	    next += sizeof(float);
	    ESP_get_int((int *) next, &cluster.width, 1);
	    next += sizeof(int);
	    ESP_get_int((int *) next, (int *) (&cluster.charge), 1);
	    next += sizeof(float);
	    ESP_get_int((int *) next, (int *) (&cluster.variance), 1);
	    next += sizeof(float);
	    cluster.width = -cluster.width;
	    if (cluster.variance > 0) {
		cluster.sn =
		    sqrt((float) cluster.width / cluster.variance);
		cluster.sn *= cluster.charge;
	    } else {
		cluster.sn = 0.;
	    }
	    cluster.first = -1;
	    cluster.digits = 0;
	    cluster.variances = 0;
	    /*VB: fill rest :sinnvoll??????? */
	    cluster.localmean = 0;
	    if (chamber == -1)	/*VB:das kann eigtl weg... */
		return &cluster;
	    break;



	case 3:		/* the 80 bit/cluster format */
	    ESP_get_short((short int *) next, &schamber, 1);
	    next += sizeof(short int);

	    bcharge = *next;
	    next++;
	    b4sn = *next;
	    next++;
	    bwidth = *next;
	    next++;

	    ESP_get_int((int *) next, (int *) &cog, 1);
	    next += sizeof(int);

	    cluster.chamber = (int) (schamber & 0xFFFF);
	    cluster.charge = (int) bcharge;
	    cluster.sn = ((float) b4sn) / 4.;
	    cluster.width = (int) bwidth;
	    cluster.cog = cog;

	    return &cluster;
	    break;

	    /*VB; format 4 added */
	case 4:		/* the ITR variable length sd-format */

	    ESP_get_int((int *) next, &cluster.chamber, 1);
	    next += sizeof(int);

	    ESP_get_int((int *) next, &cluster.width, 1);
	    next += sizeof(int);

	    ESP_get_int((int *) next, &cluster.pwi, 1);
	    next += sizeof(int);

	    ESP_get_int((int *) next, &cluster.first, 1);
	    next += sizeof(int);

	    ESP_get_int((int *) next, &cluster.end, 1);
	    next += sizeof(int);	/*VB added insteda of izero/flag */

	    ESP_get_int((int *) next, (int *) &cluster.localmean, 1);
	    next += sizeof(float);

	    ESP_get_int((int *) next, (int *) &cluster.cog, 1);
	    next += sizeof(float);

	    ESP_get_int((int *) next, (int *) &cluster.charge, 1);
	    next += sizeof(float);

	    length = cluster.end - cluster.first;

	    digits = realloc(digits, length * sizeof(int));
	    ESP_get_int((int *) next, (int *) digits, length);
	    cluster.digits = digits;
	    next += length * sizeof(float);

	    variances = realloc(variances, length * sizeof(int));
	    ESP_get_int((int *) next, (int *) variances, length);
	    cluster.variances = variances;
	    next += length * sizeof(float);

	    /*VB wird charge auf dem gleichen Intervall wie variances berechnet??, 
	       sonst ist sn falsch... */
	    cluster.variance = 0;
	    for (i = 0; i < length; i++)
		cluster.variance += cluster.variances[i];
	    if (cluster.variance > 0) {
		cluster.sn = cluster.charge;
		cluster.sn *= sqrt(1 / cluster.variance);
	    } else {
		cluster.sn = 0;
	    }
	    return &cluster;
	    break;

	case 5:		/* the 88 bit/cluster format */
	    ESP_get_short((short int *) next, &schamber, 1);
	    next += sizeof(short int);

	    bcharge = *next;
	    next++;
	    b4sn = *next;
	    next++;
	    ESP_get_short((short int *) next, &swidth, 1);
	    next += sizeof(short int);

	    ESP_get_int((int *) next, (int *) &cog, 1);
	    next += sizeof(int);

	    cluster.chamber = (int) (schamber & 0xFFFF);
	    cluster.charge = (int) bcharge;
	    cluster.sn = ((float) b4sn) / 4.;
	    cluster.width = (int) (swidth & 0x1F);
	    cluster.first = (int) ((swidth >> 5) & 0x7FF);
	    cluster.cog = cog;


	    return &cluster;
	    break;


	default:		/* format not recognized */

	    aframe_mess(routine, GRR_SERROR,
			"Found unknown sd-format %d", format);
	    break;
	}
    }

    return 0;
}


ITR_Marvin_Header_t *ITR_marvin_sevtHeader(char *data)
/*-----------------------------------------------------------------------------
 * Return a pointer to the header struct for the current event
 * ===========================================================
 * Given as input the pointer to the data-record of a sparse-event 
 * struct, the header information is unpacked into a ITR_Marvin_Header_t 
 * structure and the pointer to this structure returned to the caller.
 *
 * Author: Michael Schmelling / 14-Oct-1998
 * VB: different pipelines    / Aug-1999
 *---------------------------------------------------------------------------*/
{
    static ITR_Marvin_Header_t header;
/*---------------------------------------------------------------------------*/
    /*VB: cellshort and celllong */
    ESP_get_int((int *) data, &header.major, 1);
    ESP_get_int((int *) data + 1, &header.minor, 1);
    ESP_get_int((int *) data + 2, &header.cellshort, 1);
    ESP_get_int((int *) data + 3, &header.celllong, 1);
    ESP_get_int((int *) data + 4, &header.format, 1);
    ESP_get_int((int *) data + 5, &header.ncl, 1);

    return &header;
}


int ITR_marvin_readSevt(ITR_Marvin_Sevt_t * sevt, FILE * sdfile)
/*-----------------------------------------------------------------------------
 * Fill a sparse-event structure from an sparse data file
 * ======================================================
 * Note: the relatively compact code for the different formats is 
 *       obtained by use of a set of macros defined in marvin.h,
 *       which also catch error conditions in a transparent way. 
 *
 * Modified by R. Wanke / 21-Jan-1998:
 *       NO swapping done any more
 *
 * Author: Michael Schmelling         / 13-Oct-1998
 *         VB: changes for new header / Jul-1999
 *         VB: new format 4           / Aug-1999
 *---------------------------------------------------------------------------*/
{
    int entrys;
    int i, ncl, format, width;
    char *routine = "ITR_marvin_readSevt";

/*---------------------------------------------------------------------------*/

    sevt->next = sevt->data;

    /* 
     *  read the event header 
     */
    /*VB  numbers changed for new header */
    entrys = 6;
    SEVT_IINPUT(sdfile, 6);

    /* 
     *  verify the byte-swap status 
     */
    ESP_get_int((int *) sevt->data + (entrys - 2), &format, 1);
    ESP_get_int((int *) sevt->data + (entrys - 1), &ncl, 1);
    if (format < 0 || format > 65535) {
	aframe_mess(routine, GRR_SERROR,
		    "sevt format = %d = 0x%x out of range!", format,
		    format);
	return SEVT_FILE_CORRUPT;
    }

    /*
     *  read the rest of the event according to the format given in the header 
     */

    for (i = 0; i < ncl; i++) {

	switch (format) {

	case 2:
	    SEVT_IINPUT(sdfile, 2);
	    ESP_get_int((int *) sevt->next - 1, &width, 1);	/*get the always 1 field */
	    if (width != 1)
		return SEVT_FILE_CORRUPT;
	    SEVT_FINPUT(sdfile, 1);
	    SEVT_IINPUT(sdfile, 1);
	    SEVT_FINPUT(sdfile, 2);
	    break;

	case 3:
	    aframe_mess(routine, GRR_SINFO, "invalid sd-format %d for ITR",
			format);
	    return GRR_FISHY;
	    break;

	case 1:		/*VB: this was default */
	    SEVT_IINPUT(sdfile, 4);
	    ESP_get_int((int *) sevt->next - 3, &width, 1);
	    SEVT_FINPUT(sdfile, 2 * width);
	    break;

	case 4:		/*von Zahlen hoch, richtig ????????? */
	    SEVT_IINPUT(sdfile, 5);
	    ESP_get_int((int *) sevt->next - 4, &width, 1);
	    SEVT_FINPUT(sdfile, 1);
	    SEVT_FINPUT(sdfile, 2 * width);
	    break;
	}
    }

    return 0;
}


int ITR_marvin_importSevt(ITR_Marvin_Sevt_t * sevt)
/*-----------------------------------------------------------------------------
 * Do an in-place-byte-wap for a sparse-event structure to the local format
 * ========================================================================
 * sevt is byte-swapped, depending of cell and format > or < 65536
 *
 * PLEASE NOTE:
 * This routine is obsolete, since sevt-structure should ALWAYS be in 'export'
 * format and its contents are automatically byte-swapped at rading time.
 *
 * Author: Rainer Wanke / 20-Jan-1999
 *         VB: different pipelines and new format 4 / Jul-1999
 *---------------------------------------------------------------------------*/
{				/*VB: cellshort and celllong */
    int *next = (int *) sevt->data;
    int i, cellshort, celllong, ncl, format, width, entrys = 6;
    char *routine = "ITR_marvin_importSevt";

/*---------------------------------------------------------------------------*/

    /* verify the byte-swap status */
    memcpy(&cellshort, sevt->data + 2 * sizeof(int), sizeof(int));
    memcpy(&celllong, sevt->data + 3 * sizeof(int), sizeof(int));
    memcpy(&format, sevt->data + 4 * sizeof(int), sizeof(int));
    /*VB: Grenzen noch einmal anschauen!!!! */
    if (cellshort >= 0 && cellshort < 256 &&
	celllong >= 0 && celllong < 256 && format >= 0 && format < 5)
	return GRR_OK;

    /* Check some sort of consistency */
    if (!(cellshort < 0 || cellshort >= 256 ||
	  celllong < 0 || celllong >= 256 || format < 0 || format >= 5)) {
	aframe_mess(routine, GRR_SERROR,
		    "Could not interprete: cellshort=0x%x, celllong=0x%x, format=0x%x",
		    cellshort, celllong, format);
	return GRR_FISHY;
    }

    /* swap header in place */
    for (i = 0; i < entrys; i++) {	/*VB: 5->6 ?? */
	*next = ESP_SWAP_INT(*next);
	next++;
    }
    /*VB:  +1 everytime */
    memcpy(&format, sevt->data + 4 * sizeof(int), sizeof(int));
    memcpy(&ncl, sevt->data + 5 * sizeof(int), sizeof(int));

    /* read the rest of the event and do byte-swapping in place */
    for (i = 0; i < ncl; i++) {

	switch (format) {

	case 2:
	    for (i = 0; i < entrys; i++) {
		*next = ESP_SWAP_INT(*next);
		next++;
	    }
	    break;

	case 3:

	    aframe_mess(routine, GRR_SINFO, "invalid sd-format %d for ITR",
			format);
	    return GRR_FISHY;
	    break;

	case 1:
	    for (i = 0; i < 4; i++) {
		*next = ESP_SWAP_INT(*next);
		next++;
	    }
	    memcpy(&width, (char *) (next - 3), sizeof(int));
	    for (i = 0; i < 2 * width; i++) {
		*next = ESP_SWAP_INT(*next);
		next++;
	    }
	    break;

	case 4:
	    for (i = 0; i < 5; i++) {
		*next = ESP_SWAP_INT(*next);
		next++;
	    }
	    memcpy(&width, (char *) (next - 4), sizeof(int));	/*VB float == int ?? */
	    for (i = 0; i < 2 * width; i++) {
		*next = ESP_SWAP_INT(*next);
		next++;
	    }
	    break;
	}
    }

    return 0;
}


/* --------------------------------------------------------------
 *
 *  write sparsified event as function of the iol parameter
 *
 *  marvin sdtype = 0: 'Old style' sd-file (just a dump of sevt) 
 *                = 1: DAQ format (EVTH + SLB block + EVTT )
 *                = 2: Same as '1' + DAQ event record 
 *  VB: target information not longer available
 *  Author:  Rainer Wanke / 14-May-1999
 *
 * -------------------------------------------------------------- */
int ITR_marvin_writeSevt(FILE * sdfile)
{
    int rc;

    ITR_Marvin_Sevt_t *sevt;
    ITR_Silraw_Event_t *silraw;

    /* structures for DAQ event data */
    EVENT_HDR *event_hdr;
    EVENT_TAIL *event_tail;
    SLB_HDR slb_hdr;
    ETB_fed etb_fed;

    /* space for output event */
    int size;
    static int evt_size = 0;
    static ETB_event *daq_evt = 0;
    char dummy_evt[sizeof(EVENT_HDR) + sizeof(EVENT_TAIL)];
    ETB_out_event out_evt;

    char *routine = "ITR_marvin_writeSevt";

    /* check for existance of sd-file */
    if (!sdfile || ITR_marvin_iol.sdtype.value < 0)
	return GRR_OK;

    /* fetch the sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no ITR_Marvin_Sevt_t found");
	return GRR_FISHY;
    }

    /* First the trivial 'old style' case */
    if (ITR_marvin_iol.sdtype.value == 0) {
	fwrite(sevt->data, sizeof(char), sevt->next - sevt->data,
	       sparse_file);
	return GRR_OK;
    }

    /* Now write the event in DAQ format
     * --------------------------------- */

    /* Dynamically allocate space for new event */
    size = sevt->next - sevt->data + sizeof(EVENT_HDR) + sizeof(EVENT_TAIL)
	+ sizeof(SLB_HDR) + sizeof(SLB_TAIL)
	+ sizeof(SLB_SUBHDR);
    if (ITR_marvin_iol.sdtype.value >= 2)
	size += sizeof(EVT_DATA_RECORD);
    /*  if (  ITR_marvin_iol.sdtype.value == 3 )
       size +=  sizeof(SLB_HDR)   + sizeof(SLB_TAIL)
       + sizeof(SLB_SUBHDR) + sizeof(TGTS-RAW); */
    if (!daq_evt) {
	daq_evt = (ETB_event *) malloc(size);
	evt_size = size;
    } else if (size > evt_size) {
	daq_evt = (ETB_event *) realloc(daq_evt, size);
	evt_size = size;
    }
    if (!daq_evt) {
	aframe_mess(routine, GRR_SERROR,
		    "Could not allocate space to write sparsified event!");
	return GRR_FISHY;
    }

    /* fill dummy event */
    event_hdr = (EVENT_HDR *) dummy_evt;
    event_tail = (EVENT_TAIL *) (dummy_evt + sizeof(EVENT_HDR));
    event_hdr->length = sizeof(EVENT_HDR) + sizeof(EVENT_TAIL);
    event_hdr->recid = RID_NORMAL_EVENT;
    event_hdr->format = ESP_LOCAL_FMT;
    event_hdr->sltevt = 0;
    event_hdr->compmask = COMP_M_ITR;
    event_hdr->evtmask = 0;
    event_tail->sltevt = event_hdr->sltevt;

    /* create new event structure */
    rc = ETB_new_output_event(event_hdr, &out_evt, daq_evt,
			      evt_size, ESP_big_endian);
    if (rc != ETB_ok) {
	aframe_mess(routine, GRR_SERROR,
		    "ETB_new_output_event: return code = %d", rc);
	return GRR_FISHY;
    }

    /* add DAQ event record */
    if (ITR_marvin_iol.sdtype.value >= 2) {
	silraw = ITR_silraw_fetch_event();
	if (!silraw)
	    aframe_mess(routine, GRR_SWARNING,
			"Could not fetch ITR_silraw event!");
	else if (silraw->daq) {
	    rc = ETB_add_daq(&out_evt, silraw->daq);
	    if (rc != ETB_ok) {
		aframe_mess(routine, GRR_SERROR,
			    "ETB_add_daq: return code = %d", rc);
		return GRR_FISHY;
	    }
	} else
	    aframe_mess(routine, GRR_SWARNING,
			"No DAQ block in ITR_silraw event!");
    }

    /* Create new SLB block header */
    slb_hdr.recid = RID_SPARSE_SLB0;
    slb_hdr.fltevt = 0;
    slb_hdr.compid = COMP_ID_ITR;
    slb_hdr.slbid = 0;
    slb_hdr.nodeid = 0;
    rc = ETB_add_slb(&out_evt, &slb_hdr);
    if (rc != ETB_ok) {
	aframe_mess(routine, GRR_SERROR,
		    "ETB_add_slb: return code = %d", rc);
	return GRR_FISHY;
    }

    /* Create 'FED block' to write sparsified data and add FED to event */
    etb_fed.data = sevt->data;
    etb_fed.slb = &slb_hdr;
    etb_fed.hdr.length = sevt->next - sevt->data + sizeof(SLB_SUBHDR);
    etb_fed.hdr.fedid = 0;
    etb_fed.hdr.recid = RID_NORMAL_SPARSE;
    etb_fed.hdr.bxid = 0;
    rc = ETB_add_fed(&out_evt, &etb_fed, 0);
    if (rc != ETB_ok) {
	aframe_mess(routine, GRR_SERROR,
		    "ETB_add_fed: return code = %d", rc);
	return GRR_FISHY;
    }


    /* write event to file */
    rc = ETB_write_event(sdfile, (ETB_event *) daq_evt);
    if (rc != ETB_ok) {
	aframe_mess(routine, GRR_SERROR,
		    "ETB_write_event: return code = %d", rc);
	return GRR_FISHY;
    }

    return GRR_OK;

}


int ITR_marvin_getITRClusters(ITR_Marvin_Chamber_t * chamber, int cell)
{

    /* ***********************************************************************
       //
       // Author: Katharina Mueller, Physikalisches Institut Heidelberg, 6/99
       //         VB: changed for HERA-B framework  
       //
       // Search for MSGC Pulses with special noise treatement
       // 
       // ITR_marvin_getITRClusters: usage with VDS-Tools 
       //
       // Pulse search: Warning: Pulse is negativ!
       //               local search: Start, Mid and End of Pulse are searched,
       //               and finally cuts applied. Pulse search starts with
       //               channel 8 !!not right for chambers of type III  ! (there 5)
       //       1. Start:                   Signal[Start] > Signal[Start+1]
       //       2. Mid: last channel with   Signal[mid]   < Signal[mid+1]
       //       3. Cut  on minimal rise of signal: 
       //                                   Signal[mid]-Signal[Start] < minDiff
       //       4. End: last channel with   Signal[end]   < Signal[end+1]
       //       5. Local Baseline: take  LocalMeanSize channels before and after 
       //               the pulse (if there is no other signal candidate) to
       //               calculate local Baseline  
       //       6. Calculate Signal: totPulse:Integrated Signal - Local Baseline
       //                            cog: weighted cluster position  
       //                            Clustersize:# channels btw Start and End 
       //                            maxSignal: Signal[mid]-LocalBaseline
       //       7. Apply Cuts: maxSignal > pulseHeight
       //                      maxSignal > CutRMS*rms (see Noise treatment)
       //                      maxSignal/Clusize > relSTEIG 
       //                      last cut is needed to avoid very broad small pulses
       //
       // Noise: count percentage of channels with positive fluctuation 
       //        > pulseHeight (posFluct)
       //        if posFluct > MaxPositiveCh
       //        a MSGC Pulse has to be bigger than CutRMS*rms 
       //        rms is only calculated if CutRMS > 0
       //
       //
       // Cuts:  pulseHeight        Cut on max Pulseheight
       //        relSTEIG           maxSignal/Clusize to avoid broad small pulses
       //                           typically: 3
       //        minDiff            min difference in ADC channels for begin of
       //                           pulse search, typically 4     
       //        LocalMeanSize      # channels before and after the pulse for 
       //                           local baseline, typically 3
       //        MaxPositiveCh      recognition of noise events if fraction of
       //                           channels with large positive signals >
       //                           MaxPositiveCh, typically 0.15
       //        CutRMS             Cut to find large signals in noise events
       //                           Signals have to be larger than  CutRMS*rms
       //                           typically 3
       //
       ************************************************************************ */



    ITR_Marvin_Sevt_t *sevt;

    int start, mid, end, Clusize;

    float cog, localmean, totPulse;	/*localmean is necessary because of charge-reflux */
    /*totPulse is charge */
    int icl, jj, jm, il, iplus, ichip;
    int PulseFound;

    float rms = 0.;		/*rms of all channels > mean                      */
    float mean = 0;		/*mean of all channels (calc for CutRMS>0)        */
    float varX = 0;		/*variation of channels > mean                    */
    float posFluct = 0;		/*percentage of channels with height > pulseHeight */
    float diff = 0;		/*Signal[mid]-Signal[start]                       */
    float diff2 = 0;		/*Signal[end]-Signal[mid] */

    float pulseHeight = chamber->threshold;	/*Cut on max. Pulseheight */
/* float    relSTEIG      = ITR_marvin_iol.relSTEIG.value;  *//*Cut maxSignal/Clusize  */
    float minDiff = ITR_marvin_iol.minDiff.value;	/*Cut min diff for pulse search */
    int LocalMeanSize = (int) (ITR_marvin_iol.LocalMeanSize.value);	/*#channels for local baseline */
    float MaxPositiveCh = ITR_marvin_iol.MaxPositiveCh.value;	/*Cut for noise treatment */
    float CutRMS = ITR_marvin_iol.CutRMS.value;	/*Cut on max Pulseheight, if noise */


    float localstart, maxSignal, SignaloverThresh;
    int begPulse, endPulse;
    int firststrip, laststrip;
    int pwi, pwf;		/*initial and final from pulswidth */

    int NOChamberChannels, ncells = -1 /* ,reliasymm=0 */ ;
    int first, last;
    float **Signals, **v;
    /* int      channelstatus[ITRNCHAN_CHAMBER];  */

    char *routine = "ITR_marvin_getITRClusters";

  /*------------------------------------------------------*/
    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "pulseHeight %f, relSTEIG %f, minDiff %f, LocalMeanSize"
		    "%f, MaxPositiveCh %f, CutRMS %f \n"
		    "Getting clusters for chamber 0x%04x,"
		    "strips %d->%d ", pulseHeight, 
		    ITR_marvin_iol.relSTEIG.value,
		    ITR_marvin_iol.minDiff.value,
		    ITR_marvin_iol.LocalMeanSize.value,
		    ITR_marvin_iol.MaxPositiveCh.value,
		    ITR_marvin_iol.CutRMS.value, chamber->chamberid,
		    chamber->first, chamber->last);


    /* fetch sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no sparse event buffer");
	return GRR_FISHY;
    }
    Signals = chamber->data;
    /* get ncells */
    for (ichip = 0; ichip < chamber->nchip; ichip++)
	if (chamber->chip[ichip]->chipstatus >= 0) {
	    ncells = chamber->chip[ichip]->ncells;
	    break;
	}
    if (ncells == -1)
	return GRR_FISHY;
    v = chamber->dvar + (cell % ncells) * chamber->nchan;

    /*get type of chamber */
    if (chamber->chambertype == 1 || chamber->chambertype == 2) {
	firststrip = 8;
	laststrip = 759;
    } else if (chamber->chambertype == 3) {
	firststrip = 5;
	laststrip = 762;
    } else {
	aframe_mess(routine, GRR_SERROR, "unknown chambertype");
	return GRR_FISHY;
    }
    /*firststrip=8;laststrip=759; */

    /*VB: no read out for the first and last channels !!! */
    first = chamber->first;
    if (first < firststrip)
	first = firststrip;
    last = chamber->last;
    if (last > laststrip)
	last = laststrip;
    NOChamberChannels = last - first + 1;
    if (NOChamberChannels <= 0) {
	aframe_mess(routine, GRR_SWARNING, "last %d < first %d",
		    last, first);
	return GRR_FISHY;
    }
    if (NOChamberChannels > ITRNCHAN_CHAMBER - 2 * firststrip) {
	/*VB*/
	    aframe_mess(routine, GRR_SWARNING,
			"too many channels %d per chamber for clusterfinding",
			NOChamberChannels);
	return GRR_FISHY;
    }

    /* VB: handle channelstatus */
    /* das sollte wohl lieber einzeln in den Algorithmus, nicht so pauschal */
    /*  for (i=0;i<ITRNCHAN_CHAMBER;i++)
   channelstatus[i]=1; *//*1=ok *//*VB lieber 0?? */
    /* if (ITR_marvin_getChannelstatus(chamber->chamberid,channelstatus)!=GRR_OK)
       return GRR_FISHY;   
       for (jj = first+1; jj<= last-1; jj++){
       if (channelstatus[jj]  != 1)
       *(Signals[jj]) =  (*(Signals[jj-1]) + *(Signals[jj+1]))/2;
       } */


    /*remove bad channels */
    /*VB alles richtig???!!!!!!!!!!! */
    /*calculate <dvar> */
    /* for (jj = first+1; jj <= last-1; jj++){
       middvar += (*(chamber->dvar[jj]));
       }
       middvar /= (last-first-1);
       for (jj = first+1; jj <= last-1; jj++){
       if (*(chamber->dvar[jj]) > CutRMS * middvar)
       *(Signals[jj]) = (*(Signals[jj-1]) + *(Signals[jj+1]))/2;
       } */


    /*CutRMS=0: no noise treatement get mean and rms, 
       count channels with positive fluctuations> pulseHeight */

    if (CutRMS != 0) {
	if (ITR_marvin_iol.debug_mode.value)
	    aframe_mess(routine, GRR_SINFO, "noise treatment...");

	/*calculation of mean */
	for (jj = first; jj <= last; jj++) {
	    mean += *(Signals[jj]);
	}
	mean /= NOChamberChannels;


	/* calculation of varX, rms and #posFluct */
	/* iplus = number of signals > mean       */
	iplus = 0;
	for (jj = first; jj <= last; jj++) {
	    if (*(Signals[jj]) > mean) {
		if (*(Signals[jj]) - mean > pulseHeight) {
		    posFluct++;
		}
		iplus++;
		varX += (*(Signals[jj]) - mean) * (*(Signals[jj]) - mean);
	    }
	}
	posFluct /= NOChamberChannels;
	if (iplus > 0)
	    varX /= iplus;
	rms = sqrt(varX);

	if (posFluct < MaxPositiveCh) {
	    CutRMS = 0;
	}
	if (ITR_marvin_iol.debug_mode.value)
	    aframe_mess(routine, GRR_SINFO, "mean %f rms %f posFluct  %f",
			mean, rms, posFluct);
    }


    /*end noise treatment */
    start = mid = end = jj = first;

    /*get mid and start of pulse */
    while (jj < last - 2) {
	totPulse = 0.;
	Clusize = 0;
	PulseFound = 0;
	start = mid = end = jj;
	/* rising edge: */
	while (mid < last - 1 && *(Signals[mid]) >= *(Signals[mid + 1])) {
	    mid++;
	}
	/*get end of pulse */
	end = mid;
	diff = (*(Signals[start]) - *(Signals[mid]));
	if (start == first || diff >= minDiff) {	/*Cut on minimal rise of signal */
	    /* falling edge */
	    while (end < last - 1
		   && *(Signals[end]) <= *(Signals[end + 1])) {
		end++;
	    }
	    diff2 = *(Signals[end]) - *(Signals[mid]);

	    /*calculate first localmean at start and end of cluster */
	    icl = 0;		/*counts the entrys used for localmean */
	    localstart = 0;
	    if (end < last - 1) {
		localstart -= *(Signals[end]);
		icl++;
	    }
	    if (start > first) {
		localstart -= *(Signals[start]);
		icl++;
	    }
	    localmean = localstart;
	    if (icl)
		localstart /= icl;

	    /* calculate local baseline in window LocalMeanSize if no new signal cand */
	    /* after the pulse ... */
	    il = 1;
	    while (il < LocalMeanSize && end + il < last - 1 &&
		   -1 * *(Signals[end + il]) - localstart < 2 * minDiff) {
		localmean -= *(Signals[end + il]);
		icl++;
		il++;
	    }
	    /* ... and in front of the pulse */
	    il = 1;
	    while (il < LocalMeanSize && start - il > first &&
		   -1 * *(Signals[start - il]) - localstart < 2 * minDiff) {
		localmean -= *(Signals[start - il]);
		icl++;
		il++;
	    }
	    if (icl > 0)
		localmean /= icl;


	    /*calculate totSignal, cog and maxSignal */
	    if (end == last - 1 || diff2 >= minDiff) {
		pwi = pwf = 0;	/* --wg-- 5/feb/00 :  line was "pwi = pwf == 0; " */
		totPulse = 0.;
		Clusize = 0;
		cog = 0;
		begPulse = start + 1;
		endPulse = end - 1;
		if (endPulse < begPulse) {
		    begPulse = mid;
		    endPulse = mid;
		}

		/*new "physical clusterwidth" is pwf-pwi */
		for (jm = start; jm <= end; jm++) {
		    SignaloverThresh = -1 * *(Signals[jm]) - localmean;
		    if ((SignaloverThresh >= pulseHeight) && (pwi == 0)) {
			pwi = jm;
			pwf = pwi;
		    }
		    if ((SignaloverThresh < pulseHeight) && (pwi != 0)
			&& (pwf == pwi)) {
			pwf = jm;
		    }
		}
		if (pwf == pwi)
		    pwf++;

		if (start == first) {
		    begPulse = first;
		}
		if (end == last - 1) {
		    endPulse = last - 1;
		}
		for (jm = pwi; jm < pwf; jm++) {
		    totPulse += -1 * *(Signals[jm]) - localmean;
		    cog += (-1 * *(Signals[jm]) - localmean) * jm;
		    Clusize++;
		}
		if (totPulse)
		    cog /= totPulse;

		maxSignal = -1 * *(Signals[mid]) - localmean;

		/*cut on pulse: pulseheight and real width */
		if (maxSignal > pulseHeight &&
		    (pwf - pwi) < ITR_marvin_iol.maxCluSize.value) {

		    /*VB das endPulse + 1 ist wichtig, sonst fehlt manchmal der eigentliche Puls!!! */
		    ITR_marvin_packSevt(chamber->chamberid, pwi, begPulse,
					endPulse + 1, Signals + begPulse,
					v + begPulse, (pwf - pwi), &cog,
					&totPulse, &localmean);

		    PulseFound = 1;

		    if (ITR_marvin_iol.debug_mode.value) {
			aframe_mess(routine, GRR_SINFO,
				    "cog %f totPulse %f clusize %d localmean %f start %d end %d",
				    cog, totPulse, (pwf - pwi), localmean,
				    pwi, pwf);
			if (cog + 0.0001 < pwi)
			    aframe_mess(routine, GRR_SWARNING,
					"cog %f < begin %d", cog, pwi);
		    }
		}
	    }
	}
	if (start >= end || !PulseFound) {
	    jj++;
	} else {
	    jj = end;
	}
    }

    return GRR_OK;
}


/*-----------------------------------------------------------------------------
 * marvin's cluster finding routine
 * ================================
 * Cluster candidates are defined as connected regions of digitizings with
 * the same charge and significance=(signal/noise)^2 larger than SEEDCUT.
 * At most one consecutive dead channel is allowed within one cluster. 
 * The total significance of the candidate, chi2 = sum (signal/noise)^2, 
 * then is used to decide whether to accept the cluster. Note that the 
 * chi2 defined here, because of the SEEDCUT constraint, has different
 * properties than the standard chi2-variable. A simple Monte Carlo 
 * simulation of uncorrelated gaussian random deviates shows that the
 * probability density function for large values of the local chi2 variable
 * is an almost universal exponential function, i.e. for random clusters 
 * it is only weakly dependent on SEEDCUT and the cluster width. 
 * It thus is well suited to be used in a quality cut: 
 * cluster candidates are accepted for chi2 > CHI2CUT.
 *
 * Author: Michael Schmelling / 21-May-1998
 *         last change     MS / 14-Oct-1998
 *---------------------------------------------------------------------------*/

/* Only negative cluster charges are allowed
 * ========================================= */
int ITR_marvin_getNegativeClusters(ITR_Marvin_Chamber_t * chamber,
				   int cell)
{
    ITR_Marvin_Sevt_t *sevt;

    float seedcut, chi2cut;
    float **d, **v;
    int i, gap, start, width, ncells;
    float digi, var, chi2, clchi2, cog = 0;
    float localmean;
    float maxSignal, totPulse, totNoise;
    char *routine = "ITR_marvin_getNegativeClusters";
    int first, last, NOChamberChannels;
    int jm, firststrip, laststrip;
    int pwi, pwf, maxCh;

  /*-------------------------------------------------------------------------*/

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "Getting clusters for chamber 0x%x, strips %d -> %d",
		    chamber->chamberid, chamber->first, chamber->last);

    /* fetch sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no sparse event buffer");
	return GRR_FISHY;
    }

    d = chamber->data;
    /*VB:!!!! */
    ncells = chamber->chip[0]->ncells;
    v = chamber->dvar + (cell % ncells) * chamber->nchan;

    firststrip = 8;
    laststrip = 759;

    /*get type of chamber */
    if (chamber->chambertype == 1 || chamber->chambertype == 2) {
	firststrip = 8;
	laststrip = 759;
    } else if (chamber->chambertype == 3) {
	firststrip = 5;
	laststrip = 762;
    } else {
	aframe_mess(routine, GRR_SERROR, "unknown chambertype");
	return GRR_FISHY;
    }

    first = chamber->first;
    if (first < firststrip)
	first = firststrip;
    last = chamber->last;
    if (last > laststrip)
	last = laststrip;
    NOChamberChannels = last - first + 1;

    if (NOChamberChannels <= 0) {
	aframe_mess(routine, GRR_SWARNING, "last %d < first %d",
		    last, first);
	return GRR_FISHY;
    }
    if (NOChamberChannels > ITRNCHAN_CHAMBER - 2 * firststrip) {
	/*VB*/
	    aframe_mess(routine, GRR_SWARNING,
			"too many channels %d per chamber for clusterfinding",
			NOChamberChannels);
	return GRR_FISHY;
    }

    seedcut = ITR_marvin_iol.seedcut.value;
    chi2cut = ITR_marvin_iol.chi2cut.value;

    /* loop over strips for cluster search */
    width = 0;
    gap = 0;
    clchi2 = 0.;

    for (i = first; i <= last; i++) {

	digi = *(d[i]);
	var = *(v[i]);
	chi2 = digi * digi / var;

	if (chi2 > seedcut && digi < -1.) {
	    width++;
	    clchi2 += (chi2);
	    gap = 0;
	}

	/* no seed found - but we had a cluster til now */
	else {
	    if (width) {

		/* allow gaps of 1 dead channel */

#if 0
		if (digi == 0. && !gap) {
		    gap++;
		    width++;
		} else
#endif
		{		/* cluster ends */
		    if (clchi2 > chi2cut) {

			pwi = pwf = 0;
			start = i - width;
			maxCh = 0;
			maxSignal = 0;
			totPulse = 0;
			totNoise = 0;

			for (jm = start; jm < start + width; jm++) {
			    if ((*(d[jm])) * (*(d[jm])) / (*(v[jm])) >
				seedcut && (pwi == 0)) {
				pwi = jm;
				pwf = pwi;
			    }
			    if ((*(d[jm])) * (*(d[jm])) / (*(v[jm])) <
				seedcut && (pwi != 0) && (pwf == pwi)) {
				pwf = jm;
			    }
			}
			if (pwf == pwi)
			    pwf++;
			if (pwi == 0) {
			    pwi = start;
			    pwf = start + width;
			}

			cog = 0;

			for (jm = start; jm < start + width; jm++) {
			    totPulse += -1 * *(d[jm]);
			    totNoise += *(v[jm]);
			    cog += (-1 * *(d[jm])) * jm;
			}

			if (totPulse)
			    cog /= totPulse;

			totNoise = sqrt(totNoise);

			ITR_marvin_packSevt(chamber->chamberid, start,
					    start, start + width,
					    d + start, v + start, width,
					    &cog, &totPulse, &localmean);
		    }
		    /* if ( clchi2 < chi2cut ) */
		    width = 0;
		    gap = 0;
		    clchi2 = 0.;

		}		/* cluster ends */
	    }			/* if (width) */
	}
    }				/*end for-loop over strips */

    return GRR_OK;
}


int ITR_marvin_getSparseITRClusters(ITR_Marvin_Chamber_t * chamber,
				    int cell)
{

/* ***********************************************************************
//
// Author: Arnd Lange           02-2000
// 
//   Search for MSGC Pulses with special noise treatement
// ITR_marvin_getITRClusters: usage with VDS-Tools 
//
// Pulse search: Remark: Pulse is negativ!
// 
// User parameter: relSTEIG, pulseHeight
//
//       1. Calculate baseline and 'noise' of the chamber
//          1. calculate average(av0) and RMS(RMS0) of all channels
//          2. calculate average (-> baseline) and RMS for channels 
//             lying inside +/- relSTEIG * RMS0 around av0
//
//       2. Detect Cluster
//          if (Baseline - Signal(i)) > pulseHeigt  -> call 'get_cluster_length'
//
//       3. Determine Clusterwidth
//          begin of cluster: first channels with (Baseline - Signal(j)) > relSTEIG * RMS   
//                            with j <= i
//          end   of cluster: last channels with  (Baseline - Signal(k)) > relSTEIG * RMS   
//                            with k >= i
//
//       -  Calculate Signal: totPulse:Integrated Signal - Baseline
//                            cog: weighted cluster position  
//                            Clustersize:# channels btw Start (j) and End (k) 
//                            maxSignal: max(Signal[j]..Signal[k])
//       -  Double Pulse detection: included 
//            remark:
//               result of    get_cluster_length: 
//               0                              if no double pulse detected
//               ch no of begin of 2. pulse     if double pulse detected
//
//
//
// Cuts:  pulseHeight        Cut on max Pulseheight
//        relSTEIG           threshold for determination of pulse start/end  (in RMS of noise)
//
************************************************************************ */
    ITR_Marvin_Sevt_t *sevt;

    int n, i;
    float sum, sum2;

    float cog, localmean, localsig, totPulse;	/*localmean is necessary because of charge-reflux */
    int jm;

    float pulseHeight = ITR_marvin_iol.pulseHeight.value;	/*Cut on max. Pulseheight */
    float relSTEIG = ITR_marvin_iol.relSTEIG.value;	/*Cut maxSignal/Clusize  */
    float rat = ITR_marvin_iol.ClRat.value;	/*cluster ratio: for detection of cluster with width > 1 */

    float maxSignal;
    int begPulse, endPulse, begPulse_2, Clusize;
    int pwi, pwf;
    /*VB: vorlaeufig */
    int NOChamberChannels, ncells = -1;
    int first, last;
    int firststrip, laststrip;
    float **Signals, **v;
    char *routine = "ITR_marvin_getITRClusters";

    float c_level_1, c_level_2;
    float signal, signal2, signal3;

  /*------------------------------------------------------*/
    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO, "pulseHeight %f, relSTEIG %f,
		    LocalMeanSize % f, MaxPositiveCh % f, CutRMS % f ",
		    ITR_marvin_iol.pulseHeight.value,
		    ITR_marvin_iol.relSTEIG.value,
		    ITR_marvin_iol.LocalMeanSize.value,
		    ITR_marvin_iol.MaxPositiveCh.value,
		    ITR_marvin_iol.CutRMS.value);

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "Getting clusters for chamber %d, strips %d -> %d",
		    chamber->chamberid, chamber->first, chamber->last);

    /* fetch sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no sparse event buffer");
	return GRR_FISHY;
    }
    Signals = chamber->data;

    /* v       = chamber->dvar + (cell%ITR_marvin_iol.ncells.value)*chamber->nchan; */
    /*VB:!!!! */
    ncells = chamber->chip[0]->ncells;
    v = chamber->dvar + (cell % ncells) * chamber->nchan;

    /* get chamber type */
    if (chamber->chambertype == 1 || chamber->chambertype == 2) {
	firststrip = 8;
	laststrip = 759;
    } else if (chamber->chambertype == 3) {
	firststrip = 5;
	laststrip = 762;
    } else {
	aframe_mess(routine, GRR_SERROR, "unkown chambertype");
	return GRR_FISHY;
    }

    /*VB: no read out for the first and last 8 channels !!! */

    first = (chamber->first < firststrip) ? firststrip : chamber->first;
    last = (chamber->last > laststrip) ? laststrip : chamber->last;

    NOChamberChannels = last - first + 1;
    if (NOChamberChannels <= 0) {
	aframe_mess(routine, GRR_SWARNING, "last %d < first %d", last,
		    first);
	return GRR_FISHY;
    }
    if (NOChamberChannels > ITRNCHAN_CHAMBER - 2 * firststrip) {
	/*VB*/
	    aframe_mess(routine, GRR_SWARNING,
			"too many channels %d per chamber for clusterfinding",
			NOChamberChannels);
	return GRR_FISHY;
    }

/*--- 1. calculate local mean  (i.e. average over chip, without peaks) */

  /*-- 1.1 average over all channels */
    sum = 0;
    sum2 = 0;

    for (i = first; i < last; i++) {
	sum += *(Signals[i]);
	sum2 += *(Signals[i]) * *(Signals[i]);
    }

    localmean = sum / NOChamberChannels;
    localsig =
	sqrt((sum2 - sum * sum / NOChamberChannels) /
	     (NOChamberChannels - 1.));

  /*--- 1.2 average over channels within values inside +/- 'C_level' from calculated average   */

    sum = 0;
    sum2 = 0;
    n = 0;

    for (i = first; i < last; i++) {
	if (fabs(*(Signals[i]) - localmean) <= relSTEIG * localsig) {
	    sum += *(Signals[i]);
	    sum2 += *(Signals[i]) * *(Signals[i]);
	    n++;
	}
    }

    localmean = (n > 0) ? (float) sum / n : -1;
    localsig = (n > 1) ? sqrt((sum2 - sum * sum / n) / (n - 1.)) : -1;

    c_level_1 = relSTEIG * localsig;
    c_level_2 = (pulseHeight < c_level_1) ? c_level_1 : pulseHeight;

    begPulse = 0;


/*--- 2. detect cluster -----------------------------------------------*/

    pwi = first;
    while (pwi < last) {
	signal = localmean - *(Signals[pwi]);

	signal2 = (pwi < last - 1
		   && rat > 0) ? localmean - *(Signals[pwi + 1]) : 0;
	signal3 = (pwi < last - 2
		   && rat > 0) ? localmean - *(Signals[pwi + 2]) : 0;

	signal3 = (signal3 > 0
		   && (signal > c_level_1 && signal2 > c_level_1
		       && signal3 > c_level_1))
	    ? 2 * (signal + signal2 + signal3) / (3 * rat)
	    : 0;
	signal2 = (signal2 > 0
		   && (signal > c_level_1
		       && signal2 >
		       c_level_1)) ? (signal + signal2) / rat : 0;

	if (signal > c_level_1 && !begPulse)
	    begPulse = pwi;
	else if (signal <= c_level_1)
	    begPulse = 0;

	if (signal > c_level_2 || signal2 > c_level_2
	    || signal3 > c_level_2) {begPulse_2 =
		ITR_marvin_get_cluster_length(pwi, begPulse, &endPulse,
					      &pwf, last, Signals,
					      localmean, c_level_1,
					      c_level_2);

	    totPulse = 0;
	    cog = 0;
	    maxSignal = localmean - *(Signals[pwi]);
	    Clusize = endPulse - begPulse + 1;
	    for (jm = begPulse; jm <= endPulse; jm++) {
		signal = localmean - *(Signals[jm]);
		totPulse += signal;
		cog += (signal * jm);
		if (maxSignal < signal)
		    maxSignal = signal;
	    }
	    cog /= totPulse;

	    if (pwf - pwi < 10)
		ITR_marvin_packSevt(chamber->chamberid, pwi, begPulse,
				    endPulse, Signals + begPulse,
				    v + begPulse, (pwf - pwi + 1), &cog,
				    &totPulse, &localmean);

	    if (ITR_marvin_iol.debug_mode.value)
		aframe_mess(routine, GRR_SINFO,
			    "cog %f totPulse %f clusize %d localmean %f start %d end %d",
			    cog, totPulse, Clusize, localmean, begPulse,
			    endPulse);

	    begPulse = begPulse_2;

	    pwi = (!begPulse_2) ? endPulse + 2	/* +2 because endPulse+1 is checked to be below cut_level_1 */
		: begPulse_2;
	} else
	    pwi++;
    }

    return GRR_OK;
}



/* help functions */

/* helper function for ITR_marvin_getSparseITRClusters                     */
/*-------------------------------------------------------------------------*/
int ITR_marvin_get_cluster_length(int js0, int j0, int *pj1, int *pjs1,
				  int jlast, float **signal,
				  float localmean, float c_level_1,
				  float c_level_2)
/*-------------------------------------------------------------------------*/
{
    int i, mid;
    int dpeak_flag = 1;

    float puresignal;

    *pjs1 = js0;
    *pj1 = js0;

    for (i = js0 + 1; i < jlast; i++) {
	puresignal = localmean - *(signal[i]);

	if (puresignal > c_level_2) {
	    if (dpeak_flag)
		*pjs1 = i;
	    else {
		mid = ((i + *pjs1) / 2.);
		*pj1 = mid;
		return (mid + 1);
	    }
	}

	dpeak_flag = (puresignal > c_level_2);

	if (puresignal > c_level_1)
	    *pj1 = i;
	else
	    return 0;
    }
    return 0;
}


/*======================================================================
 * Cluster finder working on Region of Interest only
 *
 * The algorithm can be tested on whole chambers using the
 * function ITR_marvin_getROIClustersChamber() 
 * (see below)
 * 
 *======================================================================*/


int ITR_marvin_getROIClusters(void *source, void *dest, int maxsize, 
			      ESP_machine into_fmt, SLB_SUBHDR *fed,
			      int first_channel, int size, int *nChan){
    ITR_Marvin_Chip_t *mchip1, *mchip2;

    float localmean, totPulse;	/*localmean is necessary because of charge-reflux */
    /*totPulse is charge */
    int icl, jj, jm, il, iplus, ichip;
    int PulseFound;

    float pulseHeight   = ITR_marvin_iol.pulseHeight.value;	        /*Cut on max. Pulseheight */
    float minDiff       = ITR_marvin_iol.minDiff.value;	                /*Cut min diff for pulse search */
    int LocalMeanSize   = (int) (ITR_marvin_iol.LocalMeanSize.value);	/*#channels for local baseline */
    float MaxPositiveCh = ITR_marvin_iol.MaxPositiveCh.value;	        /*Cut for noise treatment */
    float CutRMS        = ITR_marvin_iol.CutRMS.value;	                /*Cut on max Pulseheight, if noise */
    float diff = 0;		/*Signal[mid]-Signal[start]                       */
    float diff2 = 0;		/*Signal[end]-Signal[mid] */

    int begPulse, endPulse;

    float localstart, SignaloverThresh;
    float seedcut, chi2cut;
    int i, j, k, gap, start, mid, end, width, ncells, Clusize, pwi, pwf;
    float digi, var, chi2, clchi2, cog = 0, sum_q, sum_qi;
    float maxSignal, totNoise;
    char *routine = "ITR_marvin_getROIClusters";
    int first, last, NOChamberChannels;
    int chipid = -1, offset, sl, start1, stop1, start2, stop2;
    int tmpData[768] = { 0 };
    float data[128]  = { 0. };
    float svar[128]  = { 0. };

    /*-------------------------------------------------------------------------*/

    /* find chips corresponding to [FED, first_channel] */
    mchip1 = ITR_marvin_findChip ((fed->fedid  & 0xFF0F));
    mchip2 = ITR_marvin_findChip ((fed->fedid  & 0xFF0F) | 0x0010);

    /* obtain datavalid offset for this FED */
    sl     = ITR_silraw_get_superlayer(fed->fedid);
    offset = ITR_silraw_get_datavalidoffset(sl);

    start1 = offset;
    stop1  = start1 + 128;
    start2 = stop1  + 8;
    stop2  = start2 + 128;

    /* now get raw data and subtract pedestal */
    /* since the buffer map has an entry for each chip,
       we know for sure that data does not span two chips, so: */
    if (start1 <= first_channel  && first_channel + size <= stop1) {      /*    first chip  */
	for (i = 0; i + first_channel < stop1 && i < size; i++) {
	    int icc = first_channel - start1 + i;
	    data[i] = (float) ((unsigned char*) source) [i] - mchip1->pval[icc] ;
	    svar[i] = mchip1->svar[icc];
	}
    }
    else if (start2 <= first_channel && first_channel + size <= stop2) {  /*   second chip  */
	for ( i = 0 ; i + first_channel < stop2 && i < size; i++) {
	    int icc = first_channel - start2 + i;
	    data[i] = (float) ((unsigned char*) source) [i] - mchip2->pval[icc] ;
	    svar[i] = mchip2->svar[icc];
	}
    }


#if defined(DEBUG_PRINT)
    for (i = 0; i < size; ) {
	for (j = 0; j < 24 && i < size; j++, i++) {
	    int icc = first_channel + i - offset;

	    if      (icc < 128) green_fore;
	    else if (136 <= icc && icc < 264) blue_fore;
	    else    plain;
	    printf (" %2x", ((unsigned char*)source)[i]);
	}
	printf ("\n");
    }

    plain;
    printf ("\n");
#endif

#if defined (DEBUG_PRINT_DATA)
    for (i = 0; i < size; ) {
	for (j = 0; j < 24 && i < size; j++, i++) {
	    int icc = first_channel - start1 + i;

	    if      (icc < stop1) green_fore;
	    else if (136 <= icc && icc < 264) blue_fore;
	    else    plain;
	    printf ("  %+4.1f", data[i]);
	}
	printf ("\n");
    }

    plain;
    printf ("\n");
#endif

#if 1


    /* loop over strips for cluster search */
    width  = 0;
    gap    = 0;
    clchi2 = 0.;

    start = 0;
    mid   = 0;
    end   = 0;

    *nChan = 0;

    start = mid = end = jj = 0;

    first = 0;
    last  = size;

    /*get mid and start of pulse */
    while (jj < last - 2) {
	totPulse = 0.;
	Clusize = 0;
	PulseFound = 0;
	start = mid = end = jj;
	/* rising edge: */
	while (mid < last - 1 && data[mid] >= data[mid + 1]) {
	    mid++;
	}
	/*get end of pulse */
	end = mid;
	diff = data[start] - data[mid];
	if (start == first || diff >= minDiff) {	/*Cut on minimal rise of signal */
	    /* falling edge */
	    while (end < last - 1
		   && (data[end]) <= (data[end + 1])) {
		end++;
	    }
	    diff2 = (data[end]) - (data[mid]);

	    /*calculate first localmean at start and end of cluster */
	    icl = 0;		/*counts the entrys used for localmean */
	    localstart = 0;
	    if (end < last - 1) {
		localstart -= (data[end]);
		icl++;
	    }
	    if (start > first) {
		localstart -= (data[start]);
		icl++;
	    }
	    localmean = localstart;
	    if (icl)
		localstart /= icl;

	    /* calculate local baseline in window LocalMeanSize if no new signal cand */
	    /* after the pulse ... */
	    il = 1;
	    while (il < LocalMeanSize && end + il < last - 1 &&
		   -1 * (data[end + il]) - localstart < 2 * minDiff) {
		localmean -= (data[end + il]);
		icl++;
		il++;
	    }
	    /* ... and in front of the pulse */
	    il = 1;
	    while (il < LocalMeanSize && start - il > first &&
		   -1 * (data[start - il]) - localstart < 2 * minDiff) {
		localmean -= (data[start - il]);
		icl++;
		il++;
	    }
	    if (icl > 0)
		localmean /= icl;


	    /*calculate totSignal, cog and maxSignal */
	    if (end == last - 1 || diff2 >= minDiff) {
		pwi = pwf = 0;	/* --wg-- 5/feb/00 :  line was "pwi = pwf == 0; " */
		totPulse = 0.;
		Clusize = 0;
		cog = 0;
		begPulse = start + 1;
		endPulse = end - 1;
		if (endPulse < begPulse) {
		    begPulse = mid;
		    endPulse = mid;
		}

		/*new "physical clusterwidth" is pwf-pwi */
		for (jm = start; jm <= end; jm++) {
		    SignaloverThresh = -1 * (data[jm]) - localmean;
		    if ((SignaloverThresh >= pulseHeight) && (pwi == 0)) {
			pwi = jm;
			pwf = pwi;
		    }
		    if ((SignaloverThresh < pulseHeight) && (pwi != 0)
			&& (pwf == pwi)) {
			pwf = jm;
		    }
		}
		if (pwf == pwi)
		    pwf++;

		if (start == first) {
		    begPulse = first;
		}
		if (end == last - 1) {
		    endPulse = last - 1;
		}
		for (jm = pwi; jm < pwf; jm++) {
		    totPulse += -1 * (data[jm]) - localmean;
		    cog += (-1 * (data[jm]) - localmean) * jm;
		    Clusize++;
		}
		if (totPulse)
		    cog /= totPulse;

		maxSignal = -1 * (data[mid]) - localmean;

		/*cut on pulse: pulseheight and real width */
		if (maxSignal > pulseHeight &&
		    (pwf - pwi) < ITR_marvin_iol.maxCluSize.value) {

		    PulseFound = 1;

		    ((float*)dest) [*nChan] = cog;
		    ++ *nChan ;
		}
	    }
	}
	if (start >= end || !PulseFound) {
	    jj++;
	} else {
	    jj = end;
	}
    }


#if 0
    seedcut = 6; /*ITR_marvin_iol.seedcut.value;*/
    chi2cut = 3; /*ITR_marvin_iol.chi2cut.value;*/

    for (i = 0; i < size-1; i++) {
	digi = data[i+1] - data[i];
	var  = svar[i]   + svar[i+1];
	chi2 = digi * digi / var;

	if (!start && chi2 > seedcut && digi < -1.) {
	    start = i+1;
	    mid   = start;
	}

	if (start) {
	    while (mid < size - 1 && data[mid] >= data[mid + 1]) {
		mid++;
	    }
	
	    end = mid;
	    
	    while (end < size - 1 && fabs(data[end] - data[end+1]) > seedcut)
		end ++;

	    i = end;

/* 	    ITR_marvin_packSevt(chamber->chamberid, start, */
/* 				start, end, */
/* 				d + start, v + start, end - start, */
/* 				&cog, &totPulse, &localmean); */

	    /* compute centre of gravity for this cluster */
	    sum_q  = 0;
	    sum_qi = 0;
	    for (k = start; k <= end; k++) {
		tmpData[k] = 1;
		sum_q  += data[k];
		sum_qi += data[k] * k;
	    }
	    
	    if (fabs (sum_q) >  0.1) {
		((float*)dest) [*nChan] = sum_qi/sum_q;
		++ *nChan ;
	    }
	    else {
		((float*)dest) [*nChan] = 0.5 * (start+end);
		++ *nChan;
	    }
	    
	    start = 0;
	}
    }				/*end for-loop over strips */
#endif

#if 0
    for (i = 0; i < size; i++) {
	for (j = 0; j < 12; j++, i++) {
	    if (tmpData[i]) red_back;
	    else            plain;
	    printf ("   %4.1f", data[i]);
	    plain;
	}
	printf ("\n");
    }
    printf ("======================================================================\n");
    
#endif
#endif

    return GRR_OK;
}



int ITR_marvin_getROIClustersChamber(ITR_Marvin_Chamber_t * chamber, int cell)
{
    ITR_Marvin_Sevt_t *sevt;

    float seedcut, chi2cut;
    float **d, **v;
    int i, j, k, gap, start, mid, end, width, ncells;
    float digi, var, chi2, clchi2, cog = 0;
    float localmean;
    float maxSignal, totPulse, totNoise;
    char *routine = "ITR_marvin_getROIClustersChamber";
    int first, last, NOChamberChannels;
    int jm, firststrip, laststrip;
    int pwi, pwf, maxCh;

    int tmpData[768] = { 0 };


  /*-------------------------------------------------------------------------*/

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "Getting clusters for chamber 0x%x, strips %d -> %d",
		    chamber->chamberid, chamber->first, chamber->last);

    /* fetch sparsified event */
    sevt = ITR_marvin_fetch_sevt();
    if (!sevt) {
	aframe_mess(routine, GRR_SINFO, "no sparse event buffer");
	return GRR_FISHY;
    }

    d = chamber->data;
    /*VB:!!!! */
    ncells = chamber->chip[0]->ncells;
    v = chamber->dvar + (cell % ncells) * chamber->nchan;

    firststrip = 8;
    laststrip = 759;

    /*get type of chamber */
    if (chamber->chambertype == 1 || chamber->chambertype == 2) {
	firststrip = 8;
	laststrip = 759;
    } else if (chamber->chambertype == 3) {
	firststrip = 5;
	laststrip = 762;
    } else {
	aframe_mess(routine, GRR_SERROR, "unknown chambertype");
	return GRR_FISHY;
    }

    first = chamber->first;
    if (first < firststrip)
	first = firststrip;
    last = chamber->last;
    if (last > laststrip)
	last = laststrip;
    NOChamberChannels = last - first + 1;

    if (NOChamberChannels <= 0) {
	aframe_mess(routine, GRR_SWARNING, "last %d < first %d",
		    last, first);
	return GRR_FISHY;
    }
    if (NOChamberChannels > ITRNCHAN_CHAMBER - 2 * firststrip) {
	/*VB*/
	    aframe_mess(routine, GRR_SWARNING,
			"too many channels %d per chamber for clusterfinding",
			NOChamberChannels);
	return GRR_FISHY;
    }

    seedcut = ITR_marvin_iol.seedcut.value;
    chi2cut = ITR_marvin_iol.chi2cut.value;

    /* loop over strips for cluster search */
    width = 0;
    gap = 0;
    clchi2 = 0.;

    start = 0;
    mid   = 0;
    end   = 0;

    for (i = first; i < last; i++) {
	digi = *(d[i+1]) - *(d[i]);
	var = *(v[i]) + *(v[i+1]);
	chi2 = digi * digi / var;

	if (!start && chi2 > seedcut && digi < -1.) {
	    start = i+1;
	    mid   = start;
	}

	if (start) {
	    while (mid < last - 1 && *(d[mid]) >= *(d[mid + 1])) {
		mid++;
	    }
	
	    end = mid;
	    
	    while (end < last - 1 && fabs(*(d[end]) - *(d[end+1])) > seedcut)
		end ++;

	    i = end;

	    ITR_marvin_packSevt(chamber->chamberid, start,
				start, end,
				d + start, v + start, end - start,
				&cog, &totPulse, &localmean);

	    
	    for (k = start; k <= end; k++) 
		tmpData[k] = 1;
	    
	    start = 0;
	}

    }				/*end for-loop over strips */

    return GRR_OK;
}



/* -----------------------------------------------------------------------------
// function to get the channel-information for one chamber
// 
//
// VB: Oct 1999
// 
//-------------------------------------------------------------------------------*/

int ITR_marvin_getChannelstatus(int chamberid, int *channelstatus)
{

    int ichip, i, strip;
    ITR_Marvin_Chamber_t *chamber;
    /* char *routine = "ITR_marvin_getChannelstatus"; */

    chamber = ITR_marvin_findChamber(chamberid);
    if (!chamber)
	return GRR_FISHY;

    /* loop over chips of this chamber and fill channelstatus */
    for (ichip = 0; ichip < chamber->nchip; ichip++) {
	for (i = 0; i < chamber->chip[ichip]->nchan; i++) {
	    strip = chamber->chip[ichip]->strip[i];
	    channelstatus[strip] = chamber->chip[ichip]->status[i];
	}
    }
    return GRR_OK;
}


/*-----------------------------------------------------------------------
 * normalize gain variations, TZ                                                    
 * modified to take advantage of the updated gain measure per chip, WG             
 *   - argument *chamber is obsolete now. 
 *
 *----------------------------------------------------------------------*/
int ITR_marvin_norm_gain(ITR_Marvin_Chip_t * chip,
			 ITR_Marvin_Chamber_t * chamber)
{
    /* char *routine = "ITR_marvin_norm_gain"; */
    register int i;

    if (chip->hcolno_mean < ITR_marvin_iol.gc_MinHeight.value)
	return GRR_OK;

    for (i = 0; i < chip->nchan; i++) {
	chip->data[i] *= (meanGain / chip->hcolno_mean);
    }

    return GRR_OK;
}

/*=======================================================================
 * the following functions are taken from TZ's Marvin 1-3
 *
 *=======================================================================*/

/* set chip->data[dead channel] =localmean  TZ 05/10/99 */
int ITR_marvin_setdeadChannels(ITR_Marvin_Chip_t * chip)
{
    int i, m;
    float meanb = 0.;

    for (i = 0; i < chip->nchan; i++) {
	if (chip->status[i] <= 0) {
	    if (i > 2 && i < (chip->nchan - 3)) {
		for (m = i - 3; m <= i + 3; m++)
		    if (m != i)
			meanb += chip->data[m];
	    }
	    meanb /= 6.;
	    chip->data[i] = meanb;
	}
    }

    return GRR_OK;
}



/*  detection of hot channels 
 * 
 *  note: a hot channel is detected by a raw noise level which
 *        is a factor hotcut higher than the mean raw noise
 *        value of all channels of this chip 
 *        */

int ITR_marvin_detect_hotchannel(ITR_Marvin_Chip_t * chip, float *rvar,
				 int nevmon, int nevtr)
{
    int i;
    float mvar = 0;
    float hotcut;

    hotcut = ITR_marvin_iol.hotcut.value;

    if (!nevmon)
	return GRR_OK;

    if ((nevtr % nevmon) == 0) {
	for (i = 0; i < chip->nchan; i++)
	    mvar += rvar[i];
	mvar /= chip->nchan;

	for (i = 0; i < chip->nchan; i++) {
	    if (rvar[i] > (hotcut * mvar))
		chip->status[i] = 0;
	}
    }
    return GRR_OK;

}

/* setting a threshold for each chamber, this threshold is calculated as 
   ITR_marvin_iol.pulseHeight.value times the sigma of the chamber noise
   14/02/00 TZ
   */

int ITR_marvin_calculate_threshold(ITR_Marvin_Chamber_t * chamber,
				   int nevmon, int nevts)
{
    int i;
    float **dvar;
    float dvar_sigma = 0.;
    int NOChamberChannels;

    if (!nevmon)
	return GRR_OK;

    if (nevts % ITR_marvin_iol.svarUpdateRate.value)
	return GRR_OK;


    if (ITR_marvin_iol.dynamic_threshold.value == 0)
	chamber->threshold = ITR_marvin_iol.pulseHeight.value;
    else {
	dvar = chamber->dvar;
	for (i = chamber->first; i < chamber->last; i++)
	    dvar_sigma += *(dvar[i]);

	NOChamberChannels = chamber->last - chamber->first + 1;
	dvar_sigma /= (NOChamberChannels - 1);

	chamber->threshold = dvar_sigma * ITR_marvin_iol.pulseHeight.value;
    }

    return GRR_OK;
}




int ITR_marvin_cleanSetup (ITR_Marvin_Event_t *marvin_evt) {
  int i;

  for (i = 0; i < marvin_evt->nchip; i++) {
    if (marvin_evt->chip[i]->strip)  { 
      free (marvin_evt->chip[i]->strip); 
      marvin_evt->chip[i]->strip = NULL; 
    }

    if (marvin_evt->chip[i]->data)  { 
      free (marvin_evt->chip[i]->data); 
      marvin_evt->chip[i]->data = NULL; 
    }
    
    if (marvin_evt->chip[i]->status)  { 
      free (marvin_evt->chip[i]->status); 
      marvin_evt->chip[i]->status = NULL; 
    }

    if (marvin_evt->chip[i]->peds)  {   
      free (marvin_evt->chip[i]->peds); 
      marvin_evt->chip[i]->peds = NULL; 
    }

    if (marvin_evt->chip[i]->rvar)  {   
      free (marvin_evt->chip[i]->rvar); 
      marvin_evt->chip[i]->rvar = NULL;   
    }

    if (marvin_evt->chip[i]->svar)  {   
      free (marvin_evt->chip[i]->svar); 
      marvin_evt->chip[i]->svar = NULL;   
    }        

    free (marvin_evt->chip[i]);
  }                                     



  for (i = 0; i < marvin_evt->nchamber; i++) {
    if (marvin_evt->chamber[i]->data) {
      free (marvin_evt->chamber[i]->data);
      marvin_evt->chamber[i]->data = NULL;
    }
    if (marvin_evt->chamber[i]->dvar) {
      free (marvin_evt->chamber[i]->dvar);
      marvin_evt->chamber[i]->dvar = NULL;
    }
    free (marvin_evt->chamber[i]);
  }

  marvin_evt->nchip    = 0;
  marvin_evt->nchamber = 0;

  return GRR_OK;
}



/***********************************************************************/
/* old routines, kept here for historical reasons
 * do not use them!
 ***********************************************************************/

#if 0
/*-----------------------------------------------------------------------------
 *
 * Full processing for one readout chip
 * ====================================
 *   o pedestal subtraction
 *   o pedestal following
 *   o raw noise monitoring
 *   o baseline subtraction
 *   o subtracted noise following
 *
 * Author: Michael Schmelling / 21-May-1998
 *         last change     MS / 28-Jul-1998
 *
 *---------------------------------------------------------------------------*/

int ITR_marvin_processChip_OLD(ITR_Marvin_Chip_t * chip, int cell)
{
    ITR_Marvin_Chamber_t *chamber;

    static int status[128];
    float *peds, *rvar, *svar;
    int *nevtp, *nevtr, *nevts, nevmon;
    int i, j, nchan, nfit;

    char *routine = "ITR_marvin_processChip";


    /* check the obvious */
    if (!chip) {
	aframe_mess(routine, GRR_SWARNING, "No chip given!");
	return GRR_FISHY;
    }

    nchan = chip->nchan;
    if (nchan != NCHAN_CHIP) {
	aframe_mess(routine, GRR_SWARNING,
		    "invalid number of channels <%d> for chip <0x%x> => exit",
		    nchan, chip->chipid);
	return GRR_FISHY;
    }

    /* Is chamber existing and ok ? */
    chamber = (ITR_Marvin_Chamber_t *) chip->chamber;
    if (!chamber) {
	aframe_mess(routine, GRR_SWARNING,
		    "Unknown detector chamber=<0x%x>", chip->chamberid);
	return GRR_FISHY;
    }
    if (chamber->chamberstatus < 0)
	return GRR_OK;

    /* initialize local pointers */
    peds = chip->peds + (cell % chip->ncellp) * nchan;
    nevtp = chip->nevtp + (cell % chip->ncellp);
    rvar = chip->rvar + (cell % chip->ncellr) * nchan;
    nevtr = chip->nevtr + (cell % chip->ncellr);
    svar = chip->svar + (cell % chip->ncells) * nchan;
    nevts = chip->nevts + (cell % chip->ncells);
    nevmon = ITR_marvin_iol.nevmon.value;

    /* pedestal subtraction */
    ITR_marvin_subtractPedestals(chip, peds);

    /* pedestal following */
    if (ITR_marvin_iol.pedUpdateRate.value) {
	(*nevtp)++;
	ITR_marvin_followPedestals(chip, peds, nevmon, *nevtp);
    }

    /* raw noise following */
    if (ITR_marvin_iol.rvarUpdateRate.value) {
	(*nevtr)++;
	ITR_marvin_followRawNoise(chip, rvar, nevmon, *nevtr);
    }

    /* baseline subtraction */


    nfit = ITR_marvin_iol.nbaseline.value;
    if (nfit == 1)
	ITR_marvin_subtractBaseline(chip->status, chip->data, chip->nchan,
				    ITR_marvin_iol.baseline.value);

    else {
	for (j = 0; j < nfit; j++) {
	    for (i = 0; i < chip->nchan; i++) {
		status[i] = chip->status[i];
		if (i % nfit != j)
		    status[i] = 0;
	    }
	    ITR_marvin_subtractBaseline(status, chip->data, chip->nchan,
					ITR_marvin_iol.baseline.value);
	}
    }

    /* ---wg--- 
     * follow height of column number for each chip.
     * use this number to correct for different electronic gain
     */
    if (ITR_marvin_iol.gaincorrection.value)
	ITR_marvin_followColumnNumber(chip, nevmon, *nevtp);

    /*VB: noch zu testen */
    /* for (ll=0;ll<chip->nchan;ll++){
       if (chip->status[ll]==0){
       chip->data[ll]=0;
       }
       } */

    /* Check the smoothness of corrected data:
     *   If average S/N is above smooth_cut, chip is rejected */
    if (ITR_marvin_iol.smooth_cut.value < 100.)
	if (ITR_marvin_smoothCut(chip, svar, nevmon, *nevts))
	    return GRR_OK;

    /* Check the flatness of corrected data:
     *   If average S/N for certain regions is looked at
     * Regions are:  5 -> 24, 54 -> 83, 103 -> 122        */
    if (ITR_marvin_iol.flat_cut.value < 100.)
	if (ITR_marvin_flatCut(chip, svar, nevmon, *nevts))
	    return GRR_OK;

    /* subtracted noise following */
    if (ITR_marvin_iol.svarUpdateRate.value) {
	(*nevts)++;
	ITR_marvin_followSubtractedNoise(chip, svar, nevmon, *nevts);
    }

    /* detction of hot channels */
    ITR_marvin_detect_hotchannel(chip, rvar, nevmon, *nevtr);

    if (ITR_marvin_iol.gaincorrection.value)
	ITR_marvin_norm_gain(chip, chamber);

    if (ITR_marvin_iol.debug_mode.value)
	aframe_mess(routine, GRR_SINFO,
		    "nevmon = %d, *nevtp = %d, *nevts = %d", nevmon,
		    *nevtp, *nevts);

    /* return, when required number of events not yet taken */
    if (*nevtp < nevmon || *nevts < nevmon)
	return GRR_OK;

    /* update the range indices chamber->first/last and the chamberstatus */
    if (chip->first < chamber->first)
	chamber->first = chip->first;
    if (chip->last > chamber->last)
	chamber->last = chip->last;
    chamber->chamberstatus = 1;

    return GRR_OK;

}
									 									  
#endif


/* To make (X)Emacs behave nicely:
 *
 * Local variables:
 *  c-basic-offset: 4
 * End:
 */

