/*======================================================================
 * 
 * Module itrleda.c
 *
 * provides routines for the in- and output of the ITR database tables for
 * setup, monitoring and alignment
 *
 * Author:  Authors of vdsleda
 *          Wolfgang Gradl <gradl@mail.desy.de>
 *
 * Changes: apr 00 (wg): re-introduce the separation of 
 *                         Setup data   (more or less static)
 *                         Monitor data (can change frequently)
 *                         Alignment data (will change frequently)
 *                       into possibly three different data bases.  
 *
 *          nov 00 (wg): if ITR_leda_startprogram() is called with arguments
 *                       -1, -1, do nothing except noting that it was 
 *                       called once.  This is needed by the keytable
 *                       reading function acReadITR().
 *
 *          dec 00 (wg): remove superfluous calls to ITR_Chanstat->book()
 *                       to setup database in ITR_leda_startprogram
 * 
 *          apr 01 (wg): introduce new table ITR_EffRes.  This is optional,
 *                       i.e. if it does not exist, it is created on the fly
 *                       with default values -1.
 *
 *          nov 01 (wg): make reading of DB more robust:  repeat in case of
 *                       failure with GRR_CONTINUE.  Add new steering parameter
 *                       "num retries" (default: 3)
 *          
 *          mar 02 (wg): disentangle saving of ITR_Chanstat from Noise and Pedestal. 
 *                       new function ITR_leda_update_chanstat()
 *
 * $Id: itrleda.c,v 1.1.1.10 2002/10/10 15:50:53 gradl Exp $
 *
 *======================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "aframelib.h"
#include "par.h"
#include "iol.h"
#include "grr.h"

/* include Leda stuff */
#define LEDA_INITIALIZATION
#include "itrleda.h"

/* IOL definitions */
Itrleda_Iol_t itrleda_iol={
  xcvsMODULE_itrleda,
  IOL_NVERBS(Itrleda_Iol_t,0),
  {"version"       , xcvsVERSION_itrleda, IOL_TYPE_STRING},
  {"mon database"  , "/ITR_MONITOR",      IOL_TYPE_STRING},
  {"setup database", "/ITR_Setup",        IOL_TYPE_STRING},
  {"align database", "/ITR_Align",        IOL_TYPE_STRING},
  {"use mon db",     1, " ",              IOL_TYPE_INT},
  {"use setup db",   1, " ",              IOL_TYPE_INT},
  {"use align db",   1, " ",              IOL_TYPE_INT},
  {"date",           0, " ",              IOL_TYPE_INT},
  {"time",           0, " ",              IOL_TYPE_INT},
  {"dst on",         0, " ",              IOL_TYPE_INT},
  {"major",          0, " ",              IOL_TYPE_INT},
  {"minor",          0, " ",              IOL_TYPE_INT},
  {"align major",    0, " ",              IOL_TYPE_INT},
  {"align minor",    0, " ",              IOL_TYPE_INT},
  {"monitor major", -1, " ",              IOL_TYPE_INT},
  {"monitor minor", -1, " ",              IOL_TYPE_INT},
  {"offsets major", -1, " ",              IOL_TYPE_INT},
  {"offsets minor", -1, " ",              IOL_TYPE_INT},
  {"log level",      0, " ",              IOL_TYPE_INT},
  {"chip file",      "none",              IOL_TYPE_STRING},
  {"chamber file",   "none",              IOL_TYPE_STRING},
  {"fed file",       "none",              IOL_TYPE_STRING},
  {"align file",     "none",              IOL_TYPE_STRING},
  {"chip output",    "none",              IOL_TYPE_STRING},
  {"chamber output", "none",              IOL_TYPE_STRING},
  {"fed output",     "none",              IOL_TYPE_STRING},
  {"dump chips",     0, " ",              IOL_TYPE_INT},
  {"dump chambers",  0, " ",              IOL_TYPE_INT},
  {"dump feds",      0, " ",              IOL_TYPE_INT},
  {"call LedaCheckDD", 1, " ",            IOL_TYPE_INT},
  {"debug mode",     0, " ",              IOL_TYPE_INT},
  {"num retries",    3, " ",              IOL_TYPE_INT},
  {"timeout retries", 500000, " ",        IOL_TYPE_INT}
};							    

static int ITR_paraminit_done    = 0; /* Flag that param_init was done */
static int ITR_startprogram_done = 0; /* Flag that startprogram was done */

/* some arguments for aframe (?) */
extern       int    nArg;
extern const char **vArg;

/* fields to store major/minor numbers of table */
static int ITR_Chip_major     = 0;
static int ITR_Chip_minor     = 0;
static int ITR_Fed_major      = 0;
static int ITR_Fed_minor      = 0;
static int ITR_Chamber_major  = 0;
static int ITR_Chamber_minor  = 0;
/* static int ITR_Alignment_major= 0; */
/* static int ITR_Alignment_minor= 0; */

/* ------------------------------------------------------------------------
 *
 *  Initialize itrleda.
 *
 *  To be called before everything else.
 *  Done is the IOL initialization.
 *
 * ------------------------------------------------------------------------ */
int ITR_leda_param_init()
{
  char *routine = "ITR_leda_param_init";

  /* Check if we haven't called it yet */
  if ( ITR_paraminit_done ) {
    aframe_mess(routine,GRR_SWARNING,
		"ITR_leda_param_init already called, skipping...");
    return GRR_OK;
  }
  ITR_paraminit_done = 1;

  /* tell the user that its alive */
  aframe_mess(routine,GRR_SINFO,
	      "ITR_leda version %s started.",xcvsVERSION_itrleda);

  /* initialize iol parameters of itrleda package */
  iol_declare_structure( (IOL_GENERIC *) &itrleda_iol );

  return GRR_OK;
}


/* ------------------------------------------------------------------------
 *
 * Initialization of data base tables
 * 
 * this has to be moved from ITR_leda_startprogram() because
 * LedaCheckDD() may only be called once per executable which leads to 
 * problems if more than one software package (e.g. VDS & ITR) is present 
 * in something like ARTE
 * 
 * (wg 26/aug/99)
 *
 * wg 09/feb/00:  Parameter do_CheckDD:
 *                if 0,  use iol parameter itrleda call LedaCheckDD
 *                if 1,  always call LedaCheckDD 
 *                if -1, never call LedaCheckDD 
 *
 * wg 02/may/00:  use symbolic constants itrleda_CheckDD_NEVER, _DBASE, _ALWAYS
 *
 * ------------------------------------------------------------------------ */

int ITR_leda_initialize(int do_CheckDD) {

  static int ITR_init_done = 0;
  char *routine = "ITR_leda_initialize";

  /* Check if we haven't called it yet */
  if ( ITR_init_done ) {
    aframe_mess(routine,GRR_SINFO,
		"I was already called, skipping...");
    return GRR_OK;
  }
  ITR_init_done = 1;

  /* tables in ITR_Setup */
  ITR_Chip->initialize();
  ITR_Fed->initialize();
  ITR_Chamber->initialize();
  ITR_TFU->initialize();
  ITR_nede->initialize();
  ITR_nesl->initialize();
  ITR_FedSlb->initialize();
  ITR_DVOffset->initialize();

  /* tables in ITR_Monitor */
  ITR_Chanstat->initialize();
  ITR_Noise->initialize();
  ITR_Ped->initialize();
  ITR_NoiseHelix->initialize();
  ITR_PedHelix->initialize();
  ITR_EffRes->initialize();

  /* tables in ITR_Align */
  ITR_Align->initialize();

  /* Check consistency of LEDA declarations */
  if ((do_CheckDD == itrleda_CheckDD_DBASE && itrleda_iol.call_LedaCheckDD.value) ||
       do_CheckDD == itrleda_CheckDD_ALWAYS) {
    aframe_mess (routine, GRR_SINFO,
		 "calling LedaCheckDD()...");
    LedaCheckDD();
  }

  return GRR_OK;
}



/* ------------------------------------------------------------------------
 *
 *  StartProgram actions.
 *
 *  If requested, the ITR databases (setup & monitoring) 
 *  are opened, tables are booked and filled.
 *
 *  Time stamp is evaluated in this order:
 *     1. date/time   from itrleda IOL
 *     2. major/minor from itrleda IOL
 *     3. major/minor from function arguments
 *
 *  If both major and minor number are equal to 0, the actual time is taken.
 *  Otherwise the highest possible database entries <= major/minor
 *  are fetched.
 *
 *  Return value:  GRR_OK    if everything is ok
 *                 GRR_FISHY if some of the database operation fail
 * ------------------------------------------------------------------------ */
int ITR_leda_startprogram( int major, int minor)
{
  char *routine = "ITR_leda_startprogram";
  int rc, retries, num_retries;
  int kSleepBeforeRetry = 500000;


  /* declare databases */
  MizziDatabase( itrdb_setup );
  MizziDatabase( itrdb_mon   );

  /* Check if we haven't called it yet */
  if ( ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SINFO,
		"ITR_leda_startprogram already called, skipping...");
    return GRR_OK;
  }
  ITR_startprogram_done = 1;

  if (major == -1 && minor == -1) 
    return GRR_OK;

  /* Check for explicit date/time requests */
  ITR_leda_get_majorminor( &major, &minor );

  /* Initializations */
  AddStderrStream(stderr,"stderr");
  if ( itrleda_iol.debug_mode.value )
    AddStdoutStream(stdout,"stdout");
  SetLogLevel( itrleda_iol.log_level.value );

  /* for convenience only */
  num_retries = itrleda_iol.num_retries.value;
  kSleepBeforeRetry  = itrleda_iol.timeout_retries.value;

  /* Open Setup Database */
  if ( strcmp( itrleda_iol.setup_database.value, "none" ) ) {
    /* open DB */
    itrdb_setup.open( itrleda_iol.setup_database.value, 0 );

    /* book tables 
     *
     * book() returns 0 on failure and >0 on success 
     */   
    rc = 0;
    
    rc |= ! ITR_Chip    ->book( itrdb_setup );
    rc |= ! ITR_Fed     ->book( itrdb_setup );
    rc |= ! ITR_Chamber ->book( itrdb_setup );
    rc |= ! ITR_nesl    ->book( itrdb_setup );
    rc |= ! ITR_nede    ->book( itrdb_setup );
    rc |= ! ITR_TFU     ->book( itrdb_setup );
    rc |= ! ITR_FedSlb  ->book( itrdb_setup );
    rc |= ! ITR_DVOffset->book( itrdb_setup );

    if (rc) {
      aframe_mess(routine, GRR_SERROR, "Error booking tables in server %s!",
		  itrleda_iol.setup_database.value);
      return (GRR_FISHY);
    }

    /* time stamp */
    aframe_mess(routine,GRR_SINFO, "Reading ITR Setup DB %s for [%d,%d] (%s)",
		itrleda_iol.setup_database.value, major, minor, print_date(major) );
    itrdb_setup.stamp( major, minor);
    
    /* fetch data, set keys, and get major/minor numbers of match */
    /* ---------------------------------------------------------- */

    /* chip */

    if ( !strcmp( itrleda_iol.chip_file.value, "none" ) ) {
      retries = 0;
      do {
	if (retries++) { 
	  aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Chip.  Retrying... ");
	  usleep (kSleepBeforeRetry);
	}
	rc = ITR_Chip->fetch_data( DBeq ); 
      } while (!rc && retries < num_retries);

      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chip!");
        return (GRR_FISHY);
      }

      ITR_Chip   ->set_keys();
      rc = ITR_Chip   ->match( DBeq );
      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Chip!");
        return (GRR_FISHY);
      }
      ITR_Chip_major = TableMajor( itrdb_setup.get_id() );
      ITR_Chip_minor = TableMinor( itrdb_setup.get_id() );
      if ( itrleda_iol.debug_mode.value )
	aframe_mess(routine,GRR_SINFO, "ITR_Chip read for %s, minor = %d",
		    print_date(ITR_Chip_major), ITR_Chip_minor );
    }
    else {
      ITR_Chip_major = major;
      ITR_Chip_minor = minor;
    }      


    /* fed */
    if ( !strcmp( itrleda_iol.fed_file.value, "none" ) ) {
      retries = 0;
      do {
	if (retries++) { 
	  aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Fed.  Retrying... ");
	  usleep (kSleepBeforeRetry);
	}
	rc = ITR_Fed->fetch_data( DBeq ); 
      } while (!rc && retries < num_retries);

      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Fed!");
        return (GRR_FISHY);
      }

      ITR_Fed->set_keys();
      rc = ITR_Fed->match( DBeq );
      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Fed!");
        return (GRR_FISHY);
      }
      ITR_Fed_major = TableMajor( itrdb_setup.get_id() );
      ITR_Fed_minor = TableMinor( itrdb_setup.get_id() );
      if ( itrleda_iol.debug_mode.value )
	aframe_mess(routine,GRR_SINFO, "ITR_Fed read for %s, minor = %d",
		    print_date(ITR_Fed_major), ITR_Fed_minor );
    }
    else {
      ITR_Fed_major = major;
      ITR_Fed_minor = minor;
    }      


    /* chamber */
    if ( !strcmp( itrleda_iol.chamber_file.value, "none" ) ) {
      retries = 0;
      do {
	if (retries++) { 
	  aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Chamber.  Retrying... ");
	  usleep (kSleepBeforeRetry);
	}
	rc = ITR_Chamber->fetch_data( DBeq ); 
      } while (!rc && retries < num_retries);

      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chamber!");
        return (GRR_FISHY);
      }

      ITR_Chamber   ->set_keys();
      rc = ITR_Chamber   ->match( DBeq );
      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Chamber!");
        return (GRR_FISHY);
      }
      ITR_Chamber_major = TableMajor( itrdb_setup.get_id() );
      ITR_Chamber_minor = TableMinor( itrdb_setup.get_id() );
      if ( itrleda_iol.debug_mode.value )
	aframe_mess(routine,GRR_SINFO, "ITR_Chamber read for %s, minor = %d",
		    print_date(ITR_Chamber_major), ITR_Chamber_minor );
    }
    else {
      ITR_Chamber_major = major;
      ITR_Chamber_minor = minor;
    }


    /* nede/nesl */
    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_nesl.  Retrying... ");
	usleep (kSleepBeforeRetry);
      }
      rc = ITR_nesl->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);
    
    if (!rc) {
      aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_nesl!");
      return (GRR_FISHY);
    }

    ITR_nesl -> set_keys   ();
    rc = ITR_nesl -> match      (DBeq);
    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_nesl!");
        return (GRR_FISHY);
    }


    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_nede.  Retrying... ");
	usleep (kSleepBeforeRetry);
      }
      rc = ITR_nede->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);
    
    if (!rc) {
      aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_nede!");
      return (GRR_FISHY);
    }
    ITR_nede -> set_keys   ();
    rc = ITR_nede -> match      (DBeq);
    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_nede!");
        return (GRR_FISHY);
    }  

    /* FedSlb */
    rc = ITR_FedSlb    ->fetch_data (DBeq);
    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_FedSlb!");
        return (GRR_FISHY);
    }
    ITR_FedSlb    ->set_keys ();
  }




  /* Open Monitoring Database */
  if ( strcmp( itrleda_iol.mon_database.value, "none" ) ) {
  
    /* Open DB */
    itrdb_mon.open( itrleda_iol.mon_database.value, 0 );
    
    rc = 0;
    rc |= ! ITR_Chanstat->book( itrdb_mon );
    rc |= ! ITR_Ped     ->book( itrdb_mon );
    rc |= ! ITR_Noise   ->book( itrdb_mon );
    rc |= ! ITR_EffRes  ->book( itrdb_mon );

    if (rc) {
      aframe_mess(routine, GRR_SERROR, "Error booking tables!");
      return (GRR_FISHY);
    }

    /* default:  read same version as Setup database */
    if (itrleda_iol.monitor_major.value < 0 && itrleda_iol.monitor_minor.value < 0) {
      itrleda_iol.monitor_major.value = major;
      itrleda_iol.monitor_minor.value = minor;
    }

    aframe_mess(routine,GRR_SINFO, "Reading ITR Monitoring DB %s for [%d,%d] (%s)",
		itrleda_iol.mon_database.value, 
		itrleda_iol.monitor_major.value, itrleda_iol.monitor_minor.value, 
		print_date(itrleda_iol.monitor_major.value) ); 
    itrdb_mon.stamp(itrleda_iol.monitor_major.value, itrleda_iol.monitor_minor.value); 

    /* Pedestal */
    
    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Ped.  Retrying... ");
	usleep (kSleepBeforeRetry);
      }
      rc = ITR_Ped->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);

    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Ped!");
        return (GRR_FISHY);
    }

    ITR_Ped->set_keys();
    rc = ITR_Ped->match(DBeq);
    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Ped!");
        return (GRR_FISHY);
    }

    /* Noise */
    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Noise.  Retrying... ");
	usleep (kSleepBeforeRetry);
      }
      rc = ITR_Noise->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);

    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Noise!");
        return (GRR_FISHY);
    }

    ITR_Noise->set_keys();
    rc = ITR_Noise->match(DBeq);
    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Noise!");
        return (GRR_FISHY);
    }

    /* Chanstat */
    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_Chanstat.  Retrying... ");
	usleep (kSleepBeforeRetry);
      }
      rc = ITR_Chanstat->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);

    if (!rc) {
      aframe_mess(routine, GRR_SERROR, "Error fetching table ITR_Chanstat!");
      return (GRR_FISHY);
    }

    ITR_Chanstat  ->set_keys  ();
    rc = ITR_Chanstat->match(DBeq);
    if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_Chanstat!");
        return (GRR_FISHY);
    }

    /* EffRes */
    retries = 0;
    do {
      if (retries++) { 
	aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_EffRes.  Retrying... ");
	usleep (kSleepBeforeRetry);
      }
      rc = ITR_EffRes->fetch_data( DBeq ); 
    } while (!rc && retries < num_retries);

    if (!rc) {
        aframe_mess(routine, GRR_SINFO, "Warning: cannot fetch table ITR_EffRes!"); 
        ITR_leda_inventEffRes();
    } 
    else {
      ITR_EffRes->set_keys  ();
      rc = ITR_EffRes->match(DBeq);
      if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error matching table ITR_EffRes!");
        return (GRR_FISHY);
      }
    }

  }      


  /* Finally fetch relations */
  /* ----------------------- */

  /* feds */
  
  if ( strcmp( itrleda_iol.setup_database.value, "none" ) )
    rc = ITR_Fed    ->fetch_relations( DBeq );
  
  else
    aframe_mess(routine,GRR_SWARNING,
		"Neither Database nor ASCII file given for FEDs");
 

  /* chambers */
 
  if ( strcmp( itrleda_iol.setup_database.value, "none" ) )
    rc = ITR_Chamber   ->fetch_relations( DBeq );
  else
    aframe_mess(routine,GRR_SWARNING,
		"Neither Database nor ASCII file given for CHAMBERs");
  
  
  /* chips */  
 
  if ( strcmp( itrleda_iol.setup_database.value, "none" ) )
    rc = ITR_Chip   ->fetch_relations( DBeq );
  else
    aframe_mess(routine,GRR_SWARNING,
		"Neither Database nor ASCII file given for CHIPs");
 
 		

  rc = ITR_Chanstat  -> fetch_relations( DBeq );
  rc = ITR_Ped       -> fetch_relations( DBeq );
  rc = ITR_Noise     -> fetch_relations( DBeq );

  rc = ITR_nede      -> fetch_relations( DBeq );
  rc = ITR_nesl      -> fetch_relations( DBeq );

  if (!rc) {
        aframe_mess(routine, GRR_SERROR, "Error fetching ITR_relations!");
        return (GRR_FISHY);
  }


  rc = ITR_leda_fetch_alignment (itrleda_iol.align_database.value,
				 itrleda_iol.align_major.value,
				 itrleda_iol.align_minor.value);

  if (rc != GRR_OK) {
    return (GRR_FISHY);
  }

  if (itrleda_iol.offsets_major.value == -1) {
    itrleda_iol.offsets_major.value = itrleda_iol.major.value;
    itrleda_iol.offsets_minor.value = itrleda_iol.minor.value;
  }

  aframe_mess(routine,GRR_SINFO, "Reading ITR Setup DB %s for DVOffsets table [%d,%d] (%s)",
	      itrleda_iol.setup_database.value, 
	      itrleda_iol.offsets_major.value, itrleda_iol.offsets_minor.value, 
	      print_date(itrleda_iol.offsets_major.value) ); 
  itrdb_setup.stamp(itrleda_iol.offsets_major.value, itrleda_iol.offsets_minor.value); 

  retries = 0;
  do {
    if (retries++) { 
      aframe_mess(routine, GRR_SINFO, "Failed to fetch table ITR_DVOffset.  Retrying... ");
      usleep (kSleepBeforeRetry);
    }
    rc = ITR_DVOffset->fetch_data( DBeq ); 
  } while (!rc && retries < num_retries);
  
  if (!rc) {
    aframe_mess(routine, GRR_SWARNING, 
		"Error fetching table ITR_DVOffset! Using default values from steering.");
    ITR_leda_createDVOffset ();
  } else {
    ITR_DVOffset -> set_keys   ();
    rc = ITR_DVOffset -> match      (DBeq);
    if (!rc) {
      aframe_mess(routine, GRR_SERROR, "Error matching table ITR_DVOffset!");
      return (GRR_FISHY);
    }
  }
 
  /* close setup database */
  if ( strcmp( itrleda_iol.setup_database.value, "none" ) ) {
    ITR_Chip    ->unbook();
    ITR_Fed     ->unbook();
    ITR_Chamber ->unbook();
    ITR_nede    ->unbook();
    ITR_nesl    ->unbook();
    ITR_TFU     ->unbook();
    ITR_FedSlb  ->unbook();
    ITR_DVOffset->unbook();
    itrdb_setup.close();
  }      
  if ( strcmp( itrleda_iol.mon_database.value, "none" ) ) {
    ITR_Ped     ->unbook();
    ITR_Noise   ->unbook();    
    ITR_Chanstat->unbook();
    ITR_EffRes  ->unbook();
    itrdb_mon.close();
  }

  return GRR_OK;
}

/* ================================================================

   ROUTINES FOR RETRIEVING MAJOR AND MINOR NUMBER OF TABLE

   Returning major/minor number of the table.
   If no database is given, 0 is returned.

   ITR_leda_startprogram( major, minor ) has to be called somewhere
   before.

   ================================================================ */
int ITR_leda_chip_major()     { return ITR_Chip_major;     }
int ITR_leda_chip_minor()     { return ITR_Chip_minor;     }
int ITR_leda_fed_major()      { return ITR_Fed_major;      }
int ITR_leda_fed_minor()      { return ITR_Fed_minor;      }
int ITR_leda_chamber_major()  { return ITR_Chamber_major;  }
int ITR_leda_chamber_minor()  { return ITR_Chamber_minor;  }


/* ================================================================

   Get major and minor number from
   1. date, time, dst-flag  (if given in iol-file)
   2. major, minor  (if given in iol-file)
   3. major, minor  from function arguments

   If major = minor = 0, the actual time-of-the-day is taken. 

   ================================================================ */

void ITR_leda_get_majorminor( int *major, int *minor ) {

  /* Check for explicit date/time requests */
  if ( itrleda_iol.date.value ) {
    datetime_to_utime( itrleda_iol.date.value, itrleda_iol.time.value,
		       itrleda_iol.dst_on.value, major );
    *minor = itrleda_iol.minor.value;
  }
  else if ( itrleda_iol.major.value || itrleda_iol.minor.value ) {
    *major = itrleda_iol.major.value;
    *minor = itrleda_iol.minor.value;
  }

  /* check major, minor for being 0 */
  if ( *major == 0 && *minor == 0 )
    *major = time( NULL );

  return;
}

/* ================================================================

   ROUTINES FOR STORING AND UPDATING TABLES

   Storing:  Tables are stored at the exact major/minor # given.
             If already existent, it will be overwritten.

   Updating: A match TABLEle is performed with the major/minor # given.
             The Table is stored at the matched major/minor+1
             (i.e. just the minor # is increased by one).

   In any case ITR_leda_startprogram(...) must have been called before!

   If major = minor = 0, the current time is taken.

   Location of stored tables:
   --------------------------
   
   File:     iol-parameter 
             "<output file>  itrleda write ..."
             has to be specified.
	     (E.g.:  "chamber_new.map   itrleda write chambers")

   Database: If database is set & enabled in iol-file
             AND if NO output file has been specified.

   ================================================================ */




/* routine to update Pedestal and Noise tables in
   the DB 
   called from ITR_marvin from endrun
*/

int ITR_leda_update_calib( int major, int minor, char *pass )
{
  int oldmajor, oldminor, rc;

  char *routine = "ITR_leda_update_calib";

  MizziDatabase( itrdb_monitor );

  /* check for ITR_leda_startprogram call */
  if ( !ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SERROR,
		"No call to ITR_leda_startprogram() before!" );
    return GRR_FISHY;
  }

  /* Check for ITR database (has to be accessible here) */
  if ( !strcmp( itrleda_iol.mon_database.value, "none" ) ) {
    aframe_mess(routine,GRR_SERROR,
		"No ITR Monitor database specified!" );
    return GRR_FISHY;
  }

  /* password given? */
  if (pass == NULL) {
    aframe_mess (routine, GRR_SERROR,
		 "No password given for writing to database.  Go away." );
    return GRR_FISHY;
  }

  /* check major, minor */
  if ( major == 0 && minor == 0 )
    major = time( NULL );

  itrdb_monitor.open( itrleda_iol.mon_database.value,  getintpasswd(pass));
  ITR_Ped  ->book( itrdb_monitor );
  ITR_Noise->book( itrdb_monitor );
  
  itrdb_monitor.stamp( major, minor );
  ITR_Ped  ->match( DBeq );
  ITR_Noise->match( DBeq );

  itrdb_monitor.stamp( major, TABLEmaxrev - 1);
  rc = ITR_Ped->match( DBle );

  oldmajor = TableMajor( itrdb_monitor.get_id() );
  oldminor = TableMinor( itrdb_monitor.get_id() );

  if (oldmajor != major) 
    rc = 0;

  if (minor == 0 && rc) {
    oldminor ++;
  }
  else if (minor == 0 && !rc) {
    oldmajor = major;
    oldminor = 1;
  }
  else if (minor != 0) {
    oldmajor = major;
    oldminor = minor;
  }

  itrdb_monitor.stamp( oldmajor, oldminor );
  ITR_Ped->store_data( DBeq );
  ITR_Ped->store_relations( DBeq );
  ITR_Noise->store_data( DBeq );
  ITR_Noise->store_relations( DBeq );
  
  aframe_mess(routine,GRR_SINFO, "Monitoring tables stored in %s, revision [%d,%d]",
	      itrleda_iol.mon_database.value, oldmajor, oldminor);
  
  itrdb_monitor.close();

  return GRR_OK;
}



/* routine to update Chanstat table in
   the DB 
   should only be called by dedicated process.
*/

int ITR_leda_update_chanstat( int major, int minor, char *pass )
{
  int oldmajor, oldminor, rc;

  char *routine = "ITR_leda_update_calib";

  MizziDatabase( itrdb_monitor );

  /* check for ITR_leda_startprogram call */
  if ( !ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SERROR,
		"No call to ITR_leda_startprogram() before!" );
    return GRR_FISHY;
  }

  /* Check for ITR database (has to be accessible here) */
  if ( !strcmp( itrleda_iol.mon_database.value, "none" ) ) {
    aframe_mess(routine,GRR_SERROR,
		"No ITR Monitor database specified!" );
    return GRR_FISHY;
  }

  /* password given? */
  if (pass == NULL) {
    aframe_mess (routine, GRR_SERROR,
		 "No password given for writing to database.  Go away." );
    return GRR_FISHY;
  }

  /* check major, minor */
  if ( major == 0 && minor == 0 )
    major = time( NULL );

  itrdb_monitor.open( itrleda_iol.mon_database.value,  getintpasswd(pass));
  ITR_Chanstat->book( itrdb_monitor );
  
  itrdb_monitor.stamp( major, minor );
  ITR_Chanstat->match( DBeq );

  itrdb_monitor.stamp( major, TABLEmaxrev - 1);
  rc = ITR_Chanstat->match( DBle );

  oldmajor = TableMajor( itrdb_monitor.get_id() );
  oldminor = TableMinor( itrdb_monitor.get_id() );

  if (oldmajor != major) 
    rc = 0;

  if (minor == 0 && rc) {
    oldminor ++;
  }
  else if (minor == 0 && !rc) {
    oldmajor = major;
    oldminor = 1;
  }
  else if (minor != 0) {
    oldmajor = major;
    oldminor = minor;
  }

  itrdb_monitor.stamp( oldmajor, oldminor );
  ITR_Chanstat->store_data ( DBeq );
  ITR_Chanstat->store_relations( DBeq );
  
  aframe_mess(routine,GRR_SINFO, "Chanstat table stored in %s, revision [%d,%d]",
	      itrleda_iol.mon_database.value, oldmajor, oldminor);
  
  itrdb_monitor.close();

  return GRR_OK;
}



/* ----------------------------------------------------------------------
 * 
 * routine to update ITR_Align table 
 *
 * should only be called by a dedicated alignment process
 *
 * ----------------------------------------------------------------------*/

int ITR_leda_update_align( int major, int minor, char *pass )
{
  int oldmajor, oldminor;

  char *routine = "ITR_leda_update_align";

  MizziDatabase( itrdb_align );

  /* check for ITR_leda_startprogram call */
  if ( !ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SERROR,
		"No call to ITR_leda_startprogram() before!" );
    return GRR_FISHY;
  }

  /* Check for ITR database (has to be accessible here) */
  if ( !strcmp( itrleda_iol.align_database.value, "none" ) ) {
    aframe_mess(routine,GRR_SERROR,
		"No ITR Alignment database specified!" );
    return GRR_FISHY;
  }

  /* password given? */
  if (pass == NULL) {
    aframe_mess (routine, GRR_SERROR,
		 "No password given for writing to database.  Go away." );
    return GRR_FISHY;
  }

  /* check major, minor */
  if ( major == 0 && minor == 0 )
    major = time( NULL );

  itrdb_align.open( itrleda_iol.align_database.value,  getintpasswd(pass));
  ITR_Align->book( itrdb_align );

  itrdb_align.stamp( major, minor );
  ITR_Align->match( DBeq );

  oldmajor = TableMajor( itrdb_align.get_id() );
  oldminor = TableMinor( itrdb_align.get_id() );
  
  itrdb_align.stamp( oldmajor, oldminor+1 );
  ITR_Align->store_data( DBeq );
  ITR_Align->store_relations( DBeq );
  
  aframe_mess(routine,GRR_SINFO, "Alignment tables for %s (%d) stored in %s",
	      print_date(oldmajor), oldminor+1, itrleda_iol.align_database.value );
  
  itrdb_align.close();

  return GRR_OK;
}



/* ----------------------------------------------------------------------
 * 
 * routine to update ITR_EffRes table 
 *
 * should only be called by a dedicated alignment process
 *
 * ----------------------------------------------------------------------*/

int ITR_leda_update_effres( int major, int minor, char *pass )
{
  int oldmajor, oldminor, rc;

  char *routine = "ITR_leda_update_effres";

  MizziDatabase( itrdb_mon );

  /* check for ITR_leda_startprogram call */
  if ( !ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SERROR,
		"No call to ITR_leda_startprogram() before!" );
    return GRR_FISHY;
  }

  /* Check for ITR database (has to be accessible here) */
  if ( !strcmp( itrleda_iol.mon_database.value, "none" ) ) {
    aframe_mess(routine,GRR_SERROR,
		"No ITR Monitoring database specified!" );
    return GRR_FISHY;
  }

  /* password given? */
  if (pass == NULL) {
    aframe_mess (routine, GRR_SERROR,
		 "No password given for writing to database.  Go away." );
    return GRR_FISHY;
  }

  
  itrdb_mon.open( itrleda_iol.mon_database.value,  getintpasswd(pass));
  ITR_EffRes->book( itrdb_mon );

  itrdb_mon.stamp( major, TABLEmaxrev - 1);
  rc = ITR_EffRes->match( DBle );
  oldmajor = TableMajor( itrdb_mon.get_id() );
  oldminor = TableMinor( itrdb_mon.get_id() );

  if (oldmajor != major) 
    rc = 0;

  if (minor == 0 && rc) {
    oldminor ++;
  }
  else if (minor == 0 && !rc) {
    oldmajor = major;
    oldminor = 1;
  }
  else if (minor != 0) {
    oldmajor = major;
    oldminor = minor;
  }

  itrdb_mon.stamp( oldmajor, oldminor );
  ITR_EffRes->store_data( DBeq );
  ITR_EffRes->store_relations( DBeq );
  
  aframe_mess(routine,GRR_SINFO, "EffRes tables for [%d,%d] (%s) stored in server/file %s",
	      oldmajor, oldminor, print_date(oldmajor),  itrleda_iol.mon_database.value );
  
  itrdb_mon.close();

  return GRR_OK;
}



/* ----------------------------------------------------------------------
 * 
 * routine to update ITR_EffRes table 
 *
 * should only be called by a dedicated alignment process
 *
 * ----------------------------------------------------------------------*/

int ITR_leda_update_DVOffset( int major, int minor, char *pass )
{
  int oldmajor, oldminor, rc;

  char *routine = "ITR_leda_update_DVOffset";

  MizziDatabase( itrdb_setup );

  /* check for ITR_leda_startprogram call */
  if ( !ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SERROR,
		"No call to ITR_leda_startprogram() before!" );
    return GRR_FISHY;
  }

  /* Check for ITR database (has to be accessible here) */
  if ( !strcmp( itrleda_iol.setup_database.value, "none" ) ) {
    aframe_mess(routine,GRR_SERROR,
		"No ITR Setup database specified!" );
    return GRR_FISHY;
  }

  /* password given? */
  if (pass == NULL) {
    aframe_mess (routine, GRR_SERROR,
		 "No password given for writing to database.  Go away." );
    return GRR_FISHY;
  }

  
  itrdb_setup.open( itrleda_iol.setup_database.value,  getintpasswd(pass));
  ITR_DVOffset->book( itrdb_setup );

  itrdb_setup.stamp( major, TABLEmaxrev - 1);
  rc = ITR_DVOffset->match( DBle );
  oldmajor = TableMajor( itrdb_setup.get_id() );
  oldminor = TableMinor( itrdb_setup.get_id() );

  if (oldmajor != major) 
    rc = 0;

  if (minor == 0 && rc) {
    oldminor ++;
  }
  else if (minor == 0 && !rc) {
    oldmajor = major;
    oldminor = 1;
  }
  else if (minor != 0) {
    oldmajor = major;
    oldminor = minor;
  }

  itrdb_setup.stamp( oldmajor, oldminor );
  ITR_DVOffset->store_data( DBeq );
  ITR_DVOffset->store_relations( DBeq );
  
  aframe_mess(routine,GRR_SINFO, "DVOffset tables for [%d,%d] (%s) stored in server/file %s",
	      oldmajor, oldminor, print_date(oldmajor),  itrleda_iol.setup_database.value );
  
  itrdb_setup.close();

  return GRR_OK;
}




char* ITR_leda_getChamberDescr (int id) {
  int alpha;
  char *routine = "ITR_leda_getChamberDescr";
  static char theDescr[256];
  LedaTableRow(ITR_Chamber) *chamber = ITR_Chamber->get->key(id);
  LedaRelationIterator (chamber->tonede, chamber_nede);

  chamber_nede.begin();

  sprintf (theDescr, "Invalid/%04x", id);

  if ( !ITR_startprogram_done ) {
    aframe_mess(routine,GRR_SERROR,
		"No call to ITR_leda_startprogram() before!" );
    return theDescr;
  }

  if (chamber_nede.valid()) {
    LedaTableRow (ITR_nede) *itr_nede = chamber_nede.get();
    LedaRelationIterator (itr_nede->tonesl, nede_nesl);

    nede_nesl.begin();

    if (nede_nesl.valid()) {
      LedaTableRow (ITR_nesl) *itr_nesl = nede_nesl.get();

      alpha = (itr_nede->rot_db[3] > 0) ? 5 : -5;
      if (itr_nede->rot_db[3] == 0) alpha = 0;

      sprintf (theDescr, "M%2s%c%c%1d%+1d",
	                 &itr_nesl->name[2],
	                 (itr_nede->sec_db[0] % 2) ? '-' : '+',
	                 (itr_nede->sec_db[0] == 2 || itr_nede->sec_db[0] == 7) ? '+' : '-',
	                 itr_nede->lay_db,
	                 alpha
	       );
    }
  }
  return (theDescr);
}



int ITR_leda_inventEffRes(void) {
  LedaTableIterator(ITR_Chamber, chamberIt);
  char *routine = __PRETTY_FUNCTION__;

  aframe_mess (routine, GRR_SINFO, "Creating a dummy version of table ITR_EffRes...");

  for (chamberIt.begin();  chamberIt.valid(); chamberIt.next()) {
    LedaTableRow(ITR_Chamber) *theChamber = chamberIt.get();
    LedaTableRow(ITR_EffRes)  *effres;

    unsigned short chId = theChamber->id;

    effres = ITR_EffRes->add->key (chId);
    effres->res = -1.0;
    effres->eff = -1.0;
  }
  
  return (0);
}


extern int ITR_silraw_get_datavalidoffset(int);

int ITR_leda_createDVOffset(void) {
  LedaTableIterator(ITR_Chamber, chamberIt);
  char *routine = __PRETTY_FUNCTION__;

  aframe_mess (routine, GRR_SINFO, "Creating default version of table ITR_DVOffset ...");

  for (chamberIt.begin();  chamberIt.valid(); chamberIt.next()) {
    LedaTableRow(ITR_Chamber) *theChamber = chamberIt.get();
    LedaTableRow(ITR_nede)    *theNede;
    LedaRelationIterator(theChamber->tonede, cha_nede);
    LedaTableRow(ITR_DVOffset)*dvoffset;
    unsigned short chId = theChamber->id;
    int sl;

    cha_nede.begin();
    theNede = cha_nede.get();
    sl = theNede->superl_db;

    dvoffset = ITR_DVOffset->add->key (chId);
    dvoffset->offset = ITR_silraw_get_datavalidoffset (sl);
  }
  
  return (0);
}


int ITR_leda_getDVOffset (int chamber) {
  LedaTableRow(ITR_DVOffset) *theOffset = ITR_DVOffset->get->key(chamber);

  if (theOffset)
    return (theOffset->offset);
  else 
    return -1;
}

int ITR_leda_FED_getDVOffset (int fed) {
  LedaTableRow(ITR_Chip)     *theChip   = ITR_Chip->get->key ((fed & 0xFF0F));
  LedaTableRow(ITR_Chamber)  *theChamber= LedaRelation (theChip->chamber);
  LedaTableRow(ITR_DVOffset) *theOffset = ITR_DVOffset->get->key(theChamber->id);

  if (theOffset)
    return (theOffset->offset);
  else 
    return -1;
}
