/*----------------------------------------------------------------
  ITR sparsification and related stuff running on SLT

  authors    : Arnd Lange (AL), 
               Wolfgang Gradl (WG) <gradl@mail.desy.de>

  ................................................................
  
  Changes:  20/apr/00 (WG):  give up idea of doing sparsification 
                             and hit finding separately.  Call usual 
                             ITR_Marvin - procedure (maybe with looser
                             cuts?) 

            20/may/00 (WG):  read configuration from data base 
                             /ITR_Setup:ITR/Steering.  Default (hardcoded)
                             is major/minor 11/1, can be changed by command 
                             line parameter -ITRSteer:[maj]/[min]

                             Parameters from database can still be overwritten
                             by reading in an .iol file, but this is 
                             strongly discouraged!

            30/jun/00 (WG):  incorporate CNA capabilities (link this 
                             together with itr_cifclient.c).
                             
                             Steering parameter:  CNA itrslt

            28/jul/00 (WG):  Change startup behaviour to make DAQ booting faster:
                             steering database is _only_ read if command line 
                             parameter -ITRsteer[:<maj>/<min>] is present.  
                             maj/min now defaults to 10/2 (used for SLT nodes)
                             .iol file is read in any case.
                             
	    19/apr/02 (WG):  Activate region-of-interest cluster-finder

  $Id: itrslt.c,v 1.1.1.8 2002/04/19 09:28:28 gradl Exp $

  ----------------------------------------------------------------*/

/* include files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>
#include <time.h>

#include "table.h"
#include "component.h"
#include "event.h"
#include "etb.h"
#include "esp.h"
#include "grr.h"
#include "cif.h"

#include "aframelib.h"
#include "itrleda.h"
#include "ITR_silraw.h"
#include "ITR_marvin.h"


#include "itrslt.h"
#include "itr_cif.h"

#include "rhp.h"

/*------------------------------
    global macros
 ------------------------------- */
#define nint(A)    ((A>0) ? (int) (A+0.5) : (int) (A-0.5))


/*------------------------------
    global variables --- local for this compilation unit
 ------------------------------- */
static int const charsize = sizeof (char);
static int const intsize = sizeof (int);
static int const floatsize = sizeof (float);
static int const headersize = sizeof (ITR_Marvin_Header_t);

static const char *ITRsteer = "-ITRsteer:";

static int errflg = 0;
static int itrslt_paraminit_done = 0;

extern const char **vArg;
extern const int nArg;


/* if code is commited to xcvs repository: */ 
#include "CVStag.itrslt" 

#ifndef xcvsVERSION_itrslt
#define xcvsVERSION_itrslt "/dev"
#endif

#ifndef xcvsMODULE_itrslt
#define xcvsMODULE_itrslt  "itrslt"
#endif

int ITR_marvin_getROIClusters(void *source, void *dest, int maxsize, 
			      ESP_machine into_fmt, SLB_SUBHDR *fed,
			      int first_channel, int size, int *nChan);

Itrslt_Iol_t itrslt_iol = {
    xcvsMODULE_itrslt, 
    IOL_NVERBS (Itrslt_Iol_t, 0),
    {"version", xcvsVERSION_itrslt, IOL_TYPE_STRING},
    {"debug mode", 0, " ", IOL_TYPE_INT},
    {"CNA", 0, " ", IOL_TYPE_INT},
};

/*-----------------------------------------------------------------------------
 *
 * Initialize itrslt_iol
 *
 *---------------------------------------------------------------------------*/
int itrslt_param_init ()
{
    char *routine = "itrslt_param_init";

    /* Check if we haven't called it yet */
    if (itrslt_paraminit_done) {
	aframe_mess (routine, GRR_SWARNING,
		     "itrslt_param_init already called, skipping...");
	return GRR_OK;
    }
    itrslt_paraminit_done = 1;

    /* tell the user that its alive */
    aframe_mess (routine, GRR_SINFO, "itrslt version %s started.", 
		 xcvsVERSION_itrslt);

    /* declare itrrhp_iol to IOL */
    iol_declare_structure ((IOL_GENERIC *) & itrslt_iol);

    return GRR_OK;
}


/*=======================================================*/
int CIF_ITR_init ()
/*=======================================================*/
{
    int major = 0;
    int minor = 0;
    int steer_major = 10;
    int steer_minor = 2;

    CIF_handlers hdlrs;
    int rc, cnt;

    char *routine = "CIF_ITR_init";

    /* ---------------------------------------- */
    /* define handlers for ITR event processing */
    /* ---------------------------------------- */

    hdlrs.fed_handler = 0;	/* no handler for single FED processing */
    hdlrs.roi_handler = &CIF_ITR_sparse_roi;	/* &CIF_ITR_RoI_handler; */
    hdlrs.evt_handler = &CIF_ITR_sparse_event;
    hdlrs.flt_generator = 0;

    rc = CIF_install_component (COMP_N_ITR, &hdlrs);
    if (rc) {
	aframe_mess (routine, GRR_SFATAL, "Could not install handlers");
	return rc;
    }
    else {
	aframe_mess (routine, GRR_SINFO,  
		     "Successfully installed handlers for RoI and Event Sparsification");
    }

    /* ---------------------------------------- */
    /* ITR initialisation                       */
    /* ---------------------------------------- */

    /* Initialize some control words */
    aframe_store_program_name (vArg[0]);

    aframe_param_init ();
    itrslt_param_init ();
    ITR_leda_param_init ();
    ITR_silraw_param_init ();
    ITR_marvin_param_init ();

    for (cnt = 1; cnt < nArg; cnt++) {
	if (!strncmp (vArg[cnt], ITRsteer, strlen (ITRsteer))) {
	    rc = sscanf (vArg[cnt] + strlen (ITRsteer), "%d/%d", &major, &minor);
	    if (rc >= 1)
		steer_major = major;
	    if (rc == 2)
		steer_minor = minor;

	    /* only read data base if asked for by command line parameter! */
	    iol_scan_preferences_from_db ("/ITR_Setup:ITR/Steering", steer_major, steer_minor);
	}
    }



    /* preferences are loaded here */
    if (aframe_process_args (nArg, vArg, AFRAME_CLI) != AFRAME_OK) {
	aframe_mess (routine, GRR_SWARNING, "Error processing arguments.");
	/*return GRR_FISHY; */
    }

    if (itrslt_iol.CNA.value == 0) {
	aframe_mess (routine, GRR_SINFO, "Starting in non-CNA mode!");
	if ((rc = ITR_leda_startprogram (0, 0)) != GRR_OK) {
	    aframe_mess (routine, GRR_SERROR,
			 "Could not succeed with ITR_leda_startprogram");
	    return rc;
	}
	if ((rc = ITR_silraw_startprogram ()) != GRR_OK) {
	    aframe_mess (routine, GRR_SERROR,
			 "Could not succeed with ITR_silraw_startprogram");
	    return rc;
	}
	ITR_marvin_init ();

    } else {
	aframe_mess (routine, GRR_SINFO, "Starting in CNA mode!");
	itr_client_init (nArg, vArg);
    }


    return GRR_OK;
}



int CIF_ITR_sparse_event (SLB_HDR ** slbs, int n_slbs, void *dest,
			  int maxsize, ESP_machine into_fmt)
{
    SLB_HDR slbout;		/* New SLB header for sparsified data     */
    SLB_SUBHDR slbsub;		/* New SLB sub header for sparsified data */
    SLB_TAIL slbtail;		/* New SLB tail for sparsified data       */
    int slb_recid = RID_SPARSE_SLB0;
    int fed_recid = RID_NORMAL_SPARSE;

    int rc, nchip, i;
    int slb_empty, islb, datalen;

    ITR_Marvin_Sevt_t *sevt;	/* sparsified event from marvin */
    ITR_Silraw_Event_t *silraw;

    char *routine = "CIF_ITR_sparse_event";

    /* the SLB sub block we write out must not be larger than a signed 
       short.  This obviously includes some headers/trailers.  Therefore: */
    const int kMaxSparseSize = 0x7FFF - 0x2F ;  

    /* --------------------------------------------------------- */

    /* check for maxsize being smaller than half a short (so it
       fits into slb_subhdr.length.... */
    if (maxsize - sizeof (SLB_HDR) - sizeof (SLB_TAIL) > kMaxSparseSize)
	maxsize = kMaxSparseSize + sizeof (SLB_HDR) + sizeof (SLB_TAIL);

    /* loop over SLB blocks until a non-empty one is found */
    slb_empty = 1;
    islb = 0;
    while (slb_empty && islb < n_slbs) {

	/* Get first SLB header and use them it template later on */
	ESP_tran_from_slb_hdr (slbs[islb], &slbout, into_fmt);

	/* Get first SLB sub header, if SLB is non-empty  */
	if (slbout.length > sizeof (SLB_HDR) + sizeof (SLB_TAIL)) {
	    ESP_tran_from_slb_subhdr ((SLB_SUBHDR *)
				      ((char *) slbs[islb] + sizeof (SLB_HDR)), &slbsub, into_fmt);
	    slb_empty = 0;
	}

	islb++;
    }

    /* check for non-empty event first */
    if (!slb_empty) {

	/* fill SLB blocks into SILRAW structure */
	rc = ITR_silraw_get_CIF_event (slbs, n_slbs, into_fmt);
	if (rc != GRR_OK)
	    return 0;

	/* get ITR_silraw event */
	silraw = ITR_silraw_fetch_event ();

	/* get pipeline lentgh for all chips from database */
	ITR_silraw_get_nchip (&nchip);
	for (i = 0; i < nchip; i++) {
	    LedaTableRow (ITR_Chip) * ledachip =
		ITR_Chip->get->key (silraw->chip[i].chipid);

	    if (ledachip == NULL)
		continue;

	    silraw->chip[i].pipelen = ledachip->pipelen;
	}

	/* Call Marvin (and initialize at Start of Run) */
	sevt = ITR_marvin_fetch_sevt ();
	if (sevt == NULL) {
	    aframe_mess (routine, GRR_SERROR,
			 "no ITR_Marvin_Sevt_t available (marvin not initialized?)");
	    return 0;
	}
	sevt->data = dest + sizeof (SLB_HDR) + sizeof (SLB_SUBHDR);
	sevt->size = maxsize - sizeof (SLB_HDR) - sizeof (SLB_SUBHDR)
	    - sizeof (SLB_TAIL);
	sevt->next = sevt->data;
	ITR_marvin_event ();

	/* translate marvin's output - just a list of int's so far  */
	/* hopefully its allowed to operate on the same location... */
	/*  ESP_put_int( (int*) sevt->data, (int*) sevt->data,
	   (int) (sevt->next - sevt->data) );
	 */


	/* ensure that data block ends on a 4-byte boundary: */
	datalen = sevt->next - sevt->data;

	if (datalen % 4 != 0)
	    sevt->next += 4 - (datalen % 4);

	/* update SLB sub header (old one is used as template) */
	slbsub.length = (int) (sevt->next - sevt->data) + sizeof (SLB_SUBHDR);

	slbsub.recid = fed_recid;
	slbsub.fedid = 0;
	ESP_tran_to_slb_subhdr (&slbsub, (SLB_SUBHDR *) (dest + sizeof (SLB_HDR)),
				into_fmt);

	/* update output SLB length (not necessary in slb_empty case) */
	slbout.length = (int) (sevt->next - sevt->data)
	    + sizeof (SLB_HDR) + sizeof (SLB_SUBHDR) + sizeof (SLB_TAIL);

    }

    /* update SLB header (old one is used as template) */
    slbout.recid = slb_recid;
    ESP_tran_to_slb_hdr (&slbout, (SLB_HDR *) dest, into_fmt);

    /* write new SLB tail */
    slbtail.recid = slb_recid;
    slbtail.fltevt = slbout.fltevt;
    ESP_tran_to_slb_tail (&slbtail,
			  (SLB_TAIL *) ((char *) dest + slbout.length
					- sizeof (SLB_TAIL)), into_fmt);

    return slbout.length;
}


/**********************************************************************/
 
int CIF_ITR_sparse_roi(void *source, void *dest, int maxsize, 
		       ESP_machine into_fmt, SLB_SUBHDR *fed,
		       int first_channel, int size, int *nChan) {

    static int firsttime = 1;
    ITR_Marvin_Chip_t *mChip;
    ITR_Marvin_Chamber_t * mChamber;
    int cnt;

    if (firsttime) {
	rhp_hbook1 (1111, "size of ROI request", 288, 0, 288);
	rhp_hbook1 (1112, "hits found in 0x1000", 768, 0, 768);
	firsttime = 0;
    }

    printf (" => %s: called for %3d channels of FED %04x starting at %4d\n",
	    __PRETTY_FUNCTION__, size, fed->fedid, first_channel);

    rhp_hfill (1111, (float) size, 0., 1.);
    
    ITR_marvin_getROIClusters(source, dest, maxsize, into_fmt, fed,
			      first_channel, size, nChan);

    /*&&&&&&& DEBUGGING ONLY &&&&&&&&&& */
    mChip    = ITR_marvin_findChip ((fed->fedid  & 0xFF0F));
    mChamber = mChip->chamber;
    if (mChamber->chamberid == 0x1000) {
	for (cnt = 0; cnt < *nChan; cnt ++) {
	    rhp_hfill (1112,  ((float*)dest)[cnt], 0., 1.);
	}
    }

    return (0);  /* nothing done yet ... */
}




/*VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV./
 * Remains of Arnds sparsification.
 * This code is never actually called.
 * It is kept here for the time being for the case we decide to use
 * something in this line after all.
 *
 * these are Arnd's comments on this:
 
  format of sparsification:
    - does not change structure:  
      SLB header| SUBSLB header | data | SUBSLB header | data | SLB tail 
    - leaves SLB header as they are
    - in SUBSLB headers only the recid is changed, all other parameters are unchanged
    - sparsifies the data and writes pedestal / baseline corrected 1-byte data 
      (+ 4 bytes add. information)
    - leaves SLB tails as they are


 *VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV*/

/*=======================================================*/
ITR_Silraw_Event_t *ITR_silraw_get_CIF (SLB_HDR ** slb, int nslb, int ext_format)
/*=======================================================*/
/*
 * Fill ITR_Silraw_Event_t from list of SLB blocks
 *
 * Inputs :  SLB_HDR **slb     : Ptr to list of ptr's to slb blocks
 *           int     nslb      : number of SLB blocks in list
 *           int     ext_format: format of SLB blocks
 *
 * ---------------------------------------------------------------- */
{
    int islb, nslbfed;
    int nfed;
    static ETB_fed etb_feds[DIM_NFED];
    int nchip;
    int rc, i;

    ITR_Silraw_Event_t *silraw;

    char *routine = "ITR_silraw_get_CIF";

    /* ---------------------------------------------------------- */

    /* fetch ITR_Silraw_Event_t */
    if (!(silraw = ITR_silraw_fetch_event ())) {
	aframe_mess (routine, GRR_SINFO, "no ITR-Silraw_Event_t found");
	return NULL;
    }
    /* loop over SLBs to get all FEDs (=SLB sub-blocks) */
    nfed = 0;
    for (islb = 0; islb < nslb; islb++) {

	rc =
	    ETB_get_feds_from_slb (slb[islb], ext_format, DIM_NFED - nfed,
				   &(etb_feds[nfed]), &nslbfed);
	if (rc != ETB_ok || nslbfed < 0) {
	    if (!(errflg & 4)) {
		aframe_mess (routine, GRR_SERROR,
			     "Could not find FED blocks, nfed=%d, islb=%d, ETB rc=%d",
			     nslbfed, islb, rc);
		errflg |= 4;
	    }
	    return NULL;
	}
	nfed += nslbfed;
    }

    /* process list of FEDs */
    if ((rc = ITR_silraw_fetch_feds_from_etb_list (etb_feds, nfed)) != GRR_OK)
	return NULL;

    /* fill ITR_Silraw_header_t */
    silraw->header.fltevt = (int) etb_feds[0].slb->fltevt;
    silraw->header.sltevt = 0;
    silraw->header.bxid = (int) etb_feds[0].hdr.bxid;

    /* No daq and target event blocks known */
    silraw->daq = 0;

    /*fill the variable pipelen from database */
    if ((rc = ITR_silraw_get_nchip (&nchip)) != GRR_OK)
	return NULL;
    if (nchip == 0)
	return NULL;

    for (i = 0; i < nchip; i++) {
	LedaTableRow (ITR_Chip) * ledachip = ITR_Chip->get->key (silraw->chip[i].chipid);
	int tmp;

	if (ledachip != NULL) {
	    tmp = (int) (ledachip->pipelen);
	    silraw->chip[i].pipelen = tmp;
	}
    }

    return silraw;
}

/*=======================================================*/
void add_channels (int *pflag, int n)
/*=======================================================*/
{
    int i;

    if (n < 0)
	for (i = n; i < 0; i++)
	    pflag[i] = 1;
    else
	for (i = 1; i <= n; i++)
	    pflag[i] = 1;
}

/*=======================================================*/
int ITR_chip_sparsedata (ITR_Marvin_Chip_t * chip)
/*=======================================================*/
{
    const int Min_value = 4;	/* minimum number of different FED values for the chip 
				   which is needed to be selecetd as 'chip with data' state */
    float A_level = 2;		/* ITR_marvin_iol.aLevel.value;  / level for average caluculation in 2. step: 
				   only those channels for taken to calculate the average of the 
				   FED value which are +/- A_level * sigma inside the (1. step-) average  */
    float C_level = 1.5;	/*ITR_marvin_iol.cLevel.value;  / level for setting values to average:
				   all those channels are set to average which are +/- C_level * sigma 
				   inside the average  */
    int C_addcha = 0;		/*ITR_marvin_iol.cAddCha.value; / no. additional of channels left and
				   right to the selected which should be stored also */
    float sum, sum2;
    int n, status1, status2, i, j;
    int values[Min_value + 2];

    struct {
	int n0, n, val;
	float av0, sig0, av, sig;
    } fed_data;

    int flag[128];


/*--- 1. calculate no. of different FED values from digits ------------------------------
         average value of data of all channels       */

    sum = 0.;
    sum2 = 0.;

    n = 0;
    values[n] = chip->digi[0];
    for (i = 0; i < chip->nchan; i++) {
	sum += chip->data[i];
	sum2 += chip->data[i] * chip->data[i];

	if (n <= Min_value) {
	    j = 0;
	    while (j <= n) {
		if (chip->digi[i] == values[j])
		    break;
		j++;
	    }

	    if (j > n) {
		n++;
		values[n] = chip->digi[i];
	    }
	}
    }

    fed_data.val = n + 1;
    fed_data.n0 = chip->nchan;
    fed_data.av0 = sum / (float) fed_data.n0;
    fed_data.sig0 = sqrt ((sum2 - sum * sum / fed_data.n0) / (float) (fed_data.n0 - 1.));

/*--- 2. calculate average FED value for channels with on signal ---------------------
      i.e. only channels with values inside +/- 'A_level' from calculated average   */

    if (fed_data.val > Min_value) {
	/* skip this step to save about 10 msec time --> */

	sum = 0;
	sum2 = 0;
	n = 0;

	for (i = 0; i < chip->nchan; i++) {
	    if (fabs (chip->data[i] - fed_data.av0) <= A_level * fed_data.sig0) {
		sum += chip->data[i];
		sum2 += chip->data[i] * chip->data[i];

		n++;
	    }
	}

	fed_data.n = n;
	fed_data.av = (n > 0) ? (float) sum / fed_data.n : -2;
	fed_data.sig =
	    (n > 1) ? sqrt ((sum2 - sum * sum / fed_data.n) / (fed_data.n - 1.)) : -2;
    }
    /*  up to here -->  */
    /* then use values for average and RMS of step one instead ./
       { fed_data.n   = fed_data.n0;
       fed_data.av  = fed_data.av0;
       fed_data.sig = fed_data.sig0;
       }
       /.  up to here -->  */

    else {
	fed_data.n = -1;
	fed_data.av = -1;
	fed_data.sig = -1;
    }



/*--- 3. prepare for sparsification ---------------------------------------*/
    for (i = 0; i < 128; i++)
	flag[i] = 0;

    if (fed_data.val > Min_value) {
	for (i = 0; i < chip->nchan; i++)
	    flag[i] = (fabs (chip->data[i] - fed_data.av) > C_level * fed_data.sig);

	/* add additional channels to left (dependend on c_AddCha) */
	if (C_addcha > 0) {
	    status1 = flag[0];

	    for (i = 1; i < chip->nchan; i++) {
		if ((status2 = flag[i]) && !status1)
		    add_channels (flag + i, -C_addcha);
		status1 = status2;
	    }

	    /* add additional channels to right (dependend on c_AddCha) */
	    status1 = flag[chip->nchan - 1];
	    for (i = chip->nchan - 2; i >= 0; i--) {
		if ((status2 = flag[i]) && !status1)
		    add_channels (flag + i, C_addcha);
		status1 = status2;
	    }
	}
    }
/*--- 4. store data ------------------------------------------------------- */

    if (fed_data.val > Min_value) {
	for (i = 0; i < chip->nchan; i++) {
	    if (chip->status[i]) {
		if (flag[i])
		    chip->status[i] = 2;	/* flagged for data value storage     */
		else
		    chip->status[i] = 1;	/* flagged for data value suppression */
	    }
	}
	chip->last = nint (fed_data.av);
    } else {
	for (i = 0; i < chip->nchan; i++)
	    if (chip->status[i])
		chip->status[i] = 3;	/* no real data in FED                */

	chip->last = 0;
    }

    return 1;
}

/*=======================================================*/
int ITR_sparse_event (SLB_HDR ** slbs, int n_slbs, int into_fmt)
/*=======================================================*/
{
    ITR_Silraw_Event_t *silraw;
    ITR_Marvin_Event_t *marvin_evt;
    ITR_Marvin_Chip_t *chip;

    /* local variables */
    int i, nchip, chipid, pipelen;
    int cellshort, celllong;

    char *routine = "ITR_sparse_event";

/*---------------------------------------------------------------------------*/

    /* access information and return in case of error */

    if (!(silraw = ITR_silraw_get_CIF (slbs, n_slbs, into_fmt))) {
	aframe_mess (routine, GRR_SINFO, "no ITR_Silraw_Event_t found");
	return GRR_FISHY;
    }
    if (!(marvin_evt = ITR_marvin_fetch_event ())) {
	aframe_mess (routine, GRR_SINFO, "no Marvin_Event_t found");
	return GRR_FISHY;
    }
    /* reset all chips */
    for (i = 0; i < marvin_evt->nchip; i++)
	marvin_evt->chip[i]->chipstatus = 0;

    /* access basic event information */
    cellshort = silraw->header.cellshort;
    celllong = silraw->header.celllong;
    nchip = silraw->nchip;

    /*
     *  start processing of data 
     *  =====================================
     */

    /* Loop over all ITR_silraw-chips to process them */
    if (nchip > 0) {
	for (i = 0; i < nchip; i++) {
	    /* Don't do anything, if ITR_silraw-chip-flag is set or chipstatus < 0 */
	    if (silraw->chip[i].flag >= 1)
		continue;	/* silraw->chip[i].flag=2 for false pipeline */

	    chipid = silraw->chip[i].chipid;
	    chip = ITR_marvin_findChip (chipid);
	    if (!chip)
		continue;
	    if (chip->chipstatus < 0)
		continue;

	    chip->digi = silraw->chip[i].data;

	    ITR_marvin_setdeadChannels (chip);

	    /*VB: */
	    pipelen = silraw->chip[i].pipelen;

	    if (pipelen == NCELLHELIXSHORT)
		ITR_marvin_processChip (chip, cellshort);
	    else if (pipelen == NCELLHELIXLONG)
		ITR_marvin_processChip (chip, celllong);
	    else {
		if (!(errflg & 1))
		    aframe_mess (routine, GRR_SWARNING, "pipelinelength %d not valid",
				 pipelen);
		errflg |= 1;
	    }

			/*------------------------------------------------------------------ 
			  sparsification
			  ------------------------------------------------------------------*/
	    ITR_chip_sparsedata (chip);

	}
    }				/* end-if Nchip > 0 ? */
    return GRR_OK;
}

/*=======================================================*/
char pack (float f)
/*=======================================================*/
{
    short result;

    /* char *routine = "pack"; */

    result = nint (f);

    if (result > 127)
	result = 127;		/* aframe_mess(routine,GRR_SINFO, "value set to max"); */
    if (result < -128)
	result = -128;		/* aframe_mess(routine,GRR_SINFO, "value set to min"); */

    return (char) result;
}

/*=======================================================*/
int pack_subslb_data (int cid, void *dest, ESP_machine into_fmt)
/*=======================================================*/
{
    ITR_Marvin_Chip_t *chip[2];
    int mean[2];
    char *data = dest;
    int i, j, k, byte4bound, ndata = 2;

    char *routine = "pack_subslb_data";

    data[0] = data[1] = -2;
    data[2] = data[3] = -128;

    for (k = 0; k <= 1; k++) {
	if (k == 1)
	    cid = cid | 0x0010;

	if (!(chip[k] = ITR_marvin_findChip (cid))) {
	    if (!(errflg & 2)) {
		aframe_mess (routine, GRR_SERROR, "Chip (ID: 0x%04x) not found", cid);
		errflg |= 2;
	    }
	}
	if (!chip[k] || chip[k]->chipstatus < 0)
	    data[k] = -1;	/* no valid data in chip 1 */
	else {
	    i = 0;
	    while (i < chip[k]->nchan) {
		if (chip[k]->status[i]) {
		    if (chip[k]->status[i] == 3) {
			data[k] = -1;	/* no valid data in chip   */
			break;
		    } else if (chip[k]->status[i] == 1) {
			data[k] = 1;
			mean[k] = (chip[k]->last > 127) ? 127 :
			    ((chip[k]->last < -128) ? -128 : chip[k]->last);
			break;
		    }
		}
		i++;
	    }
	}
    }

    if (data[0] == 1) {
	data[ndata] = (char) mean[0];
	ndata++;
    }
    if (data[1] == 1) {
	data[ndata] = (char) mean[1];
	ndata++;
    }
    for (k = 0; k <= 1; k++) {
	if (data[k] == 1) {	/* sparsification */
	    i = 0;
	    while (i < chip[k]->nchan) {
		j = 0;
		while (i + j < chip[k]->nchan && chip[k]->status[i + j] != 2)
		    j++;	/* skip 'not flagged' channels */

		if (i + j < chip[k]->nchan) {
		    i += j;
		    data[ndata] = i;
		} else
		    break;

		j = 0;
		if (
		    (i + 1 < chip[k]->nchan && chip[k]->status[i + 1] != 2
		     && i > 0) || i + 1 == chip[k]->nchan) {
		    data[ndata] = -i;
		    data[ndata + 1] = pack (chip[k]->data[i]);
		} else {
		    while (i + j < chip[k]->nchan && chip[k]->status[i + j] == 2) {
			data[ndata + 2 + j] = pack (chip[k]->data[i + j]);
			j++;
		    }

		    data[ndata + 1] = j;
		}

		ndata += j + 2;
		i += j + 1;
	    }
	} else if (data[k] == -2)	/* flag not properly set  */
	    aframe_mess (routine, GRR_SWARNING,
			 "Chip (ID: 0x%04x): sparsification mode not set properly", cid);
	else if (data[k] == -1)	/* not valid data in FED */
	    ;

	if (data[k] == 1)
	    data[k] = (k == 0) ? ndata - 4 : ndata - data[0] - 4;
    }

    /* take care of the 4-byte boundary */
    if ((byte4bound = (ndata % 4)))
	ndata += (4 - byte4bound);

    return ndata;
}

/*=======================================================*/
int CIF_ITR_sparse_event_manySLBs (SLB_HDR ** slbs, int n_slbs, void *dest,
				   int maxsize, ESP_machine into_fmt)
/*=======================================================*/
{
    SLB_HDR slbout;		/* New SLB header for sparsified data     */
    SLB_SUBHDR slbsub, slbsub1;	/* New SLB sub header for sparsified data */
    SLB_TAIL slbtail;		/* New SLB tail for sparsified data       */
    int fed_recid = RID_NORMAL_SPARSE;
    int islb, byte4bound;
    int dest_slb_offs = 0, dest_subslb0_offs, dest_subslb1_offs, slb_data_length;

    char *routine = "CIF_ITR_sparse_event";

/*-----  call silraw & Marvinevent processing */
    errflg = 0;
    ITR_sparse_event (slbs, n_slbs, into_fmt);

/*----- loop over SLB blocks */
    for (islb = 0; islb < n_slbs; islb++) {
	ESP_tran_from_slb_hdr (slbs[islb], &slbout, into_fmt);

	slbsub.length = slbsub1.length = 0;

	if (slbout.length > sizeof (SLB_HDR) + sizeof (SLB_TAIL)) {	/* Get SLB sub headers, if SLB is non-empty */
	    ESP_tran_from_slb_subhdr ((SLB_SUBHDR *)
				      ((char *) slbs[islb] + sizeof (SLB_HDR)), &slbsub, into_fmt);
	    ESP_tran_from_slb_subhdr ((SLB_SUBHDR *)
				      ((char *) slbs[islb] +
				       sizeof (SLB_HDR) + slbsub.length), &slbsub1,
				      into_fmt);

	    dest_subslb0_offs = dest_slb_offs + sizeof (SLB_HDR);
	    slb_data_length =
		pack_subslb_data (slbsub.fedid & 0xff0f,
				  ((char *) dest) + dest_subslb0_offs +
				  sizeof (SLB_SUBHDR), into_fmt);
	    slbsub.length = sizeof (SLB_SUBHDR) + slb_data_length;
	    slbsub.recid = fed_recid;

	    dest_subslb1_offs = dest_subslb0_offs + slbsub.length;
	    slb_data_length =
		pack_subslb_data (slbsub1.fedid & 0xff0f,
				  ((char *) dest) + dest_subslb1_offs +
				  sizeof (SLB_SUBHDR), into_fmt);
	    slbsub1.length = sizeof (SLB_SUBHDR) + slb_data_length;
	    slbsub1.recid = fed_recid;

	    ESP_tran_to_slb_subhdr (&slbsub,
				    (SLB_SUBHDR *) (((char *) dest) + dest_subslb0_offs),
				    into_fmt);
	    ESP_tran_to_slb_subhdr (&slbsub1,
				    (SLB_SUBHDR *) (((char *) dest) + dest_subslb1_offs),
				    into_fmt);
	}
	slbout.length =
	    sizeof (SLB_HDR) + slbsub.length + slbsub1.length + sizeof (SLB_TAIL);

	/* take care that next SLB start on next 4-byte boundary */
	if ((byte4bound = (slbout.length % 4)))
	    slbout.length += (4 - byte4bound);

	ESP_tran_to_slb_hdr (&slbout, (SLB_HDR *) (((char *) dest) + dest_slb_offs),
			     into_fmt);

	dest_slb_offs += slbout.length;

	/* write new SLB tail */
	slbtail.recid = slbout.recid;
	slbtail.fltevt = slbout.fltevt;
	ESP_tran_to_slb_tail (&slbtail,
			      (SLB_TAIL *) (((char *) dest) +
					    dest_slb_offs - sizeof (SLB_TAIL)), into_fmt);

	if (dest_slb_offs + slbout.length > maxsize)
	    aframe_mess (routine, GRR_SWARNING,
			 "Event size nearly at maximum %d (max %d)", dest_slb_offs,
			 maxsize);

	if (dest_slb_offs > maxsize) {
	    aframe_mess (routine, GRR_SERROR,
			 "Event size overflow %d (max %d)", dest_slb_offs, maxsize);
	    return dest_slb_offs;
	}
    }

    return dest_slb_offs;
}


/* To make (X)Emacs behave nicely:
 *
 * Local variables:
 *  c-basic-offset: 4
 * End:
 */
